(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     86008,       2429]
NotebookOptionsPosition[     42192,       1392]
NotebookOutlinePosition[     79589,       2208]
CellTagsIndexPosition[     79546,       2205]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["The Gauss Application Example", "Title"],

Cell["\<\
Copyright \[Copyright] MathCore AB 1999-2005\
\>", "Copyright"],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell["\<\
The Gauss application example uses a text-book Gauss-elimination algorithm to \
solve a linear equation system. Two versions of the Gauss elimination \
algorithm are presented:\
\>", "Text",
 CellChangeTimes->{{3.5003717914378843`*^9, 3.5003717924378653`*^9}}],

Cell[TextData[{
 StyleBox["GaussSolveArraySlice",
  FontWeight->"Bold"],
 ". The algorithm is largely coded using the Matlab-like array section \
operations."
}], "Itemize"],

Cell[TextData[{
 StyleBox["GaussSolveForLoops",
  FontWeight->"Bold"],
 ". The algorithm is coded using traditional for-loops."
}], "Itemize"],

Cell[TextData[{
 "Both versions of the algorithm are compiled to C++ code. The performance is \
measured, both including and without MathLink overhead. Complete evaluation \
of this notebook takes 3-5 minutes on a 1.5 GHz computer. See ",
 StyleBox["factor",
  FontWeight->"Bold"],
 " variable below to adjust the speed."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[CellGroupData[{

Cell["Load the MathCode package", "Subsection"],

Cell["\<\
First put a Needs statement, to make sure that the MathCode application is \
loaded:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<MathCode`\>\"", "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"MathCode C++ 1.4.4\"\>", "\[InvisibleSpace]", "\<\" for \"\>", 
   "\[InvisibleSpace]", "\<\"vc100\"\>", 
   "\[InvisibleSpace]", "\<\" loaded from \"\>", 
   "\[InvisibleSpace]", "\<\"C:\\\\b\\\\omath\\\\MathCode\"\>"}],
  SequenceForm[
  "MathCode C++ 1.4.4", " for ", "vc100", " loaded from ", 
   "C:\\b\\omath\\MathCode"],
  Editable->False]], "Print",
 CellChangeTimes->{3.508659609669181*^9, 3.5086596831213655`*^9, 
  3.5086597522142315`*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Check current directory", "Subsection"],

Cell["\<\
You might want to set the directory to some place where all generated files \
are put, e.g. the one below, or some other directory.\
\>", "Text"],

Cell["\<\
SetDirectory[$MCRoot<>$PathnameSeparator<>\"Demos\"<>$PathnameSeparator<>\"\
Gauss\"];\
\>", "Input"],

Cell["Check whether this file is in the current directory.", "Text"],

Cell["ExistFileQ[\"Gauss.nb\"];", "Input"],

Cell["Off[General::spell1]", "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Start the Gauss Package", "Section"],

Cell["BeginPackage[\"Gauss`\",{MathCodeContexts}];", "Input"],

Cell["Define exported symbols", "Text"],

Cell["\<\
Begin[\"Gauss`\"];

GaussSolveArraySlice;
GaussSolveForLoops;
GaussSolveForLoopsC;
End[];\
\>", "Input"],

Cell[CellGroupData[{

Cell["Set compilation options", "Subsection"],

Cell["\<\
Possibly set compilation options. In this case no extra options are needed.\
\>", "Text"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["The Body of the Gauss Package", "Section"],

Cell["Begin[\"`Private`\"];", "Input"],

Cell[CellGroupData[{

Cell["The GaussSolveArraySlice function", "Subsection"],

Cell["\<\
GaussSolveArraySlice[Real[n_,n_] ain_, Real[n_,m_] bin_ ,
 Integer iterations_ ] -> Real[n,m] :=
  Module [ {
    Real[n]     dumc,
    Real[n,n] \ta,
    Real[n,m] \tb,
    Integer[n] \t{ipiv, indxr, indxc},
    Integer \t{i,k,l,irow,icol},
    Real   \t\t{pivinv, amax, tmp},
    Integer \t{beficol, afticol,  count}
   },
    
For[count=1,count<=iterations,count=count+1,(
   a=ain;  (* The arguments are always generated as const references
              and therefore cannot be changed.
              Actuall we get waste of space and time here when they are \
copied.   *)
   b=bin;
   For [k=1,k<=n, k=k+1,ipiv[[k]]=0];
   
   For [i=1,i<=n, i=i+1,
 
     (* Algorithm first finds absolutely largest matrix element *)
     amax=0.0;
      For [k=1,k<=n, k=k+1,
       If [ ipiv[[k]]==0 ,
        For [l=1,l<=n, l=l+1,
          If [ ipiv[[l]]==0 ,
            If [ Abs[a[[k,l]]] > amax,
                    amax= Abs[a[[k,l]]];
                    irow=k;
                    icol=l
              ]]]]];
     (* Abs is generated as LIGHTABS , a macro in lightmat.h *)


     ipiv[[icol]]=ipiv[[icol]]+1;
     If[ipiv[[icol]]>1,
        \"*** Gauss2 input data error ***\">>\"\"; 
        Break];
     If[ irow!=icol ,
        For [k=1,k<=n, k=k+1,
               tmp=a[[irow,k]] ;          
               a[[irow,k]]=a[[icol,k]]; 
               a[[icol,k]]=tmp];
        For [k=1,k<=m, k=k+1,
               tmp=b[[irow,k]] ;          
               b[[irow,k]]=b[[icol,k]]; 
               b[[icol,k]]=tmp]
        ];
      indxr[[i]]=irow;
      indxc[[i]]=icol;
      If [ a[[icol,icol]]==0,
            Print[\"*** Gauss2 input data error 2 ***\"];
            Break];
      
      pivinv=1.0 / a[[icol,icol]];
      a[[icol,icol]]=1.0;

      a[[icol,_]]=a[[icol,_]]*pivinv;
      b[[icol,_]]=b[[icol,_]]*pivinv;
      dumc=a[[_,icol]];
      For [k=1,k<=n, k=k+1, a[[k,icol]]=0]; (* This cannot be simplified... *) 

      a[[icol,icol]]= pivinv;

      For [k=1,k<=n, k=k+1,  
       If [ k != icol,
              a[[k,_]]= a[[k,_]]- dumc[[k]]*a[[icol,_]];
              b[[k,_]]= b[[k,_]]- dumc[[k]]*b[[icol,_]]
          ]]
   ];
   (* Extra semicolons above are harmful *)

\tFor [l=n,l>=1, l=l-1,
    \tFor [k=1,k<=n, k=k+1,
     \ttmp= a[[k,indxr[[l]]]];
     \ta[[k,indxr[[l]]]]=a[[k,indxc[[l]]]];
     \ta[[k,indxc[[l]]]]=tmp
       \t]]
   )];
  b
]; 

\
\>", "Input",
 CellChangeTimes->{{3.5003714121170425`*^9, 3.500371413585764*^9}, {
  3.50848144214148*^9, 3.5084814667192907`*^9}}]
}, Closed]],

Cell[CellGroupData[{

Cell["The GaussSolveForLoops function ", "Subsection"],

Cell["\<\
GaussSolveForLoops[Real[n_,n_] ain_, Real[n_,m_] bin_ ,
 Integer iterations_ ] -> Real[n,m] :=
 Module [ {
    Real[n] \tdumc,
    Real[n,n] \ta,
    Real[n,m] \tb,
    Integer[n] {ipiv, indxr, indxc},
    Integer \t{i, k, l, irow, icol},
    Real\t\t{pivinv, amax, tmp},
    Integer \t{beficol, afticol, count}
  },
For[count=1,count<=iterations,count=count+1,(
   a=ain;  (* The arguments are always generated as const references
              and therefore cannot be changed.
              Actuall we get waste of space and time here when they are \
copied.   *)
   b=bin;
   For [k=1,k<=n, k=k+1,ipiv[[k]]=0];
   
   For [i=1,i<=n, i=i+1,
 
     (* Algorithm first finds absolutely largest matrix element *)
     amax=0.0;
      For [k=1,k<=n, k=k+1,
       If [ ipiv[[k]]==0 ,
        For [l=1,l<=n, l=l+1,
          If [ ipiv[[l]]==0 ,
            If [ Abs[a[[k,l]]] > amax,
                    amax= Abs[a[[k,l]]];
                    irow=k;
                    icol=l
              ]]]]];


     ipiv[[icol]]=ipiv[[icol]]+1;
     If[ipiv[[icol]]>1,
        \"*** Gauss2 input data error ***\">>\"\"; 
        Break];
     If[ irow!=icol ,
        For [k=1,k<=n, k=k+1,
               tmp=a[[irow,k]] ;     
               a[[irow,k]]=a[[icol,k]]; 
               a[[icol,k]]=tmp];
        For [k=1,k<=m, k=k+1,
               tmp=b[[irow,k]] ;          
               b[[irow,k]]=b[[icol,k]]; 
               b[[icol,k]]=tmp]
        ];
      indxr[[i]]=irow;
      indxc[[i]]=icol;
      If [ a[[icol,icol]]==0,
            Print[\"*** Gauss2 input data error 2 ***\"];
            Break];
      
      pivinv=1.0 / a[[icol,icol]];
      a[[icol,icol]]=1.0;

      For [k=1,k<=n, k=k+1,  
          a[[icol,k]]=a[[icol,k]]*pivinv];
      For [k=1,k<=m, k=k+1,  
          b[[icol,k]]=b[[icol,k]]*pivinv];
      For [k=1,k<=n, k=k+1,  
          dumc[[k]]=a[[k,icol]];a[[k,icol]]=0];

      
      a[[icol,icol]]= pivinv;

      For [k=1,k<=n, k=k+1,  
       If [ k != icol,
         For [l=1,l<=n, l=l+1,
              a[[k,l]]= a[[k,l]]- dumc[[k]]* a[[icol,l]]];  
             For [l=1,l<=m, l=l+1,
              b[[k,l]]= b[[k,l]]- dumc[[k]]* b[[icol,l]]] 
          ]]
   ];

For [l=n,l>=1, l=l-1,
    For [k=1,k<=n, k=k+1,
     tmp= a[[k,indxr[[l]]]];
     a[[k,indxr[[l]]]]=a[[k,indxc[[l]]]];
     a[[k,indxc[[l]]]]=tmp
        ]]
   )]; 
  b
]; \
\>", "Input",
 CellChangeTimes->{{3.5003714193512783`*^9, 3.5003714205387554`*^9}}]
}, Closed]],

Cell[CellGroupData[{

Cell["The Compiled GaussSolveForLoops function, using Compile[] ", \
"Subsection"],

Cell["\<\
GaussSolveForLoopsC=
Compile[{{ain,_Real,2},{bin,_Real,2},{iterations,_Integer}},
Module[{
\tn=Dimensions[ain][[1]],
\tm=Dimensions[bin][[2]],
\tdumc=Table[0.0,{n}],
\ta=Table[0.0,{n},{n}],
\tb=Table[0.0,{n},{m}],
\tipiv=Table[0,{n}], 
    indxr=Table[0,{n}], 
    indxc=Table[0,{n}],
    i=0, k=0, l=0, irow=0, icol=0,
    pivinv=0.0, amax=0.0, tmp=0.0,
    beficol=0, afticol=0, count=0
  },
  
For[count=1,count<=iterations,count=count+1,(
   a=ain;  (* The arguments are always generated as const references
              and therefore cannot be changed.
              Actuall we get waste of space and time here when they are \
copied.   *)
   b=bin;
   For [k=1,k<=n, k=k+1,ipiv[[k]]=0];
   
   For [i=1,i<=n, i=i+1,
 
     (* Algorithm first finds absolutely largest matrix element *)
     amax=0.0;
      For [k=1,k<=n, k=k+1,
       If [ ipiv[[k]]==0 ,
        For [l=1,l<=n, l=l+1,
          If [ ipiv[[l]]==0 ,
            If [ Abs[a[[k,l]]] > amax,
                    amax= Abs[a[[k,l]]];
                    irow=k;
                    icol=l
              ]]]]];


     ipiv[[icol]]=ipiv[[icol]]+1;
     If[ipiv[[icol]]>1,
        \"*** Gauss2 input data error ***\">>\"\"; 
        Break];
     If[ irow!=icol ,
        For [k=1,k<=n, k=k+1,
               tmp=a[[irow,k]] ;     
               a[[irow,k]]=a[[icol,k]]; 
               a[[icol,k]]=tmp];
        For [k=1,k<=m, k=k+1,
               tmp=b[[irow,k]] ;          
               b[[irow,k]]=b[[icol,k]]; 
               b[[icol,k]]=tmp]
        ];
      indxr[[i]]=irow;
      indxc[[i]]=icol;
      If [ a[[icol,icol]]==0,
            Print[\"*** Gauss2 input data error 2 ***\"];
            Break];
      
      pivinv=1.0 / a[[icol,icol]];
      a[[icol,icol]]=1.0;

      For [k=1,k<=n, k=k+1,  
          a[[icol,k]]=a[[icol,k]]*pivinv];
      For [k=1,k<=m, k=k+1,  
          b[[icol,k]]=b[[icol,k]]*pivinv];
      For [k=1,k<=n, k=k+1,  
          dumc[[k]]=a[[k,icol]];a[[k,icol]]=0];

      
      a[[icol,icol]]= pivinv;

      For [k=1,k<=n, k=k+1,  
       If [ k != icol,
         For [l=1,l<=n, l=l+1,
              a[[k,l]]= a[[k,l]]- dumc[[k]]* a[[icol,l]]];  
             For [l=1,l<=m, l=l+1,
              b[[k,l]]= b[[k,l]]- dumc[[k]]* b[[icol,l]]] 
          ]]
   ];

For [l=n,l>=1, l=l-1,
    For [k=1,k<=n, k=k+1,
     tmp= a[[k,indxr[[l]]]];
     a[[k,indxr[[l]]]]=a[[k,indxc[[l]]]];
     a[[k,indxc[[l]]]]=tmp
        ]]
   )]; 
  b]
  (*  ,
      CompilationTarget->\"WVM\"
  *)
  ] ;\
\>", "Input",
 CellChangeTimes->{
  3.5086593740315723`*^9, {3.5086594065467806`*^9, 3.508659484248911*^9}, {
   3.508659527857728*^9, 3.508659538826338*^9}, {3.5086597278239183`*^9, 
   3.508659729011403*^9}, {3.5086598410255947`*^9, 3.508659875447029*^9}}],

Cell["\<\
End[];
EndPackage[];\
\>", "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Executing the interpreted version in Mathematica", "Subsection"],

Cell["Create two random matrices.", "Text"],

Cell["\<\
at=Table[Random[],{10},{10}];
bt=Table[Random[],{10},{2}];\
\>", "Input"],

Cell["\<\
SetAttributes[AbsTime,HoldFirst];
AbsTime[x_] := Module[{start,res},
\t\tstart = AbsoluteTime[];
\t\tres=x;
\t\t{AbsoluteTime[]-start,res}
\t];\
\>", "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Run the ArraySlice version", "Subsection"],

Cell[TextData[{
 "Note: Too short loop time may be spent for reliable measurements, only 3- 4 \
seconds combined with a resolution of 1 second. Relevant ",
 StyleBox["factor",
  FontWeight->"Bold"],
 " for 1.5 GHz computer is 10."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"factor", "=", "10"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell["\<\
loops=1*factor;
s=(ct=GaussSolveArraySlice[at,bt,loops];)//AbsTime;
meval=s[[1]]/loops;
Print[\"TIMING FOR NON-COMPILED VERSION= \",meval];
Dot[at,ct] - bt // MatrixForm\
\>", "Input",
 CellChangeTimes->{{3.39140444265625*^9, 3.391404447875*^9}}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"TIMING FOR NON-COMPILED VERSION= \"\>", "\[InvisibleSpace]", 
   "0.11874848`7.526173052787929"}],
  SequenceForm[
  "TIMING FOR NON-COMPILED VERSION= ", 0.11874848`7.526173052787929],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086587626487727`*^9, 3.5086593836251993`*^9, 
  3.508659612231648*^9, 3.5086596861838264`*^9, 3.508659753636088*^9}],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "3.1086244689504383`*^-15"}], "2.1050869380978554`*^-15"},
     {"8.881784197001252`*^-16", "5.551115123125783`*^-16"},
     {"4.218847493575595`*^-15", 
      RowBox[{"-", "1.887379141862766`*^-15"}]},
     {
      RowBox[{"-", "8.881784197001252`*^-16"}], "1.1102230246251565`*^-15"},
     {"4.440892098500626`*^-16", "3.3306690738754696`*^-16"},
     {"7.771561172376096`*^-16", "3.885780586188048`*^-16"},
     {
      RowBox[{"-", "1.7763568394002505`*^-15"}], "1.9984014443252818`*^-15"},
     {"1.4988010832439613`*^-15", 
      RowBox[{"-", "6.661338147750939`*^-16"}]},
     {"3.1086244689504383`*^-15", 
      RowBox[{"-", "1.3322676295501878`*^-15"}]},
     {"3.219646771412954`*^-15", 
      RowBox[{"-", "3.3306690738754696`*^-16"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086587626800222`*^9, 3.5086593836251993`*^9, 
  3.508659612231648*^9, 3.5086596861838264`*^9, 3.508659753651713*^9}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Run the For-loop version", "Subsection"],

Cell[CellGroupData[{

Cell["\<\
loops=1*factor;
s=(ct=GaussSolveForLoops[at,bt,loops];)//AbsTime;
mevalFor=s[[1]]/loops;
Print[\"TIMING FOR NON-COMPILED VERSION (FOR-LOOPS)=\",mevalFor];
Dot[at,ct] - bt // MatrixForm\
\>", "Input",
 CellChangeTimes->{{3.391404459546875*^9, 3.39140446315625*^9}}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"TIMING FOR NON-COMPILED VERSION (FOR-LOOPS)=\"\>", 
   "\[InvisibleSpace]", "0.05781176`7.213561184574137"}],
  SequenceForm[
  "TIMING FOR NON-COMPILED VERSION (FOR-LOOPS)=", 
   0.05781176`7.213561184574137],
  Editable->False]], "Print",
 CellChangeTimes->{3.508658763320639*^9, 3.5086593842345667`*^9, 
  3.5086596129035144`*^9, 3.508659686871318*^9, 3.508659754276705*^9}],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "3.1086244689504383`*^-15"}], "2.1050869380978554`*^-15"},
     {"8.881784197001252`*^-16", "5.551115123125783`*^-16"},
     {"4.218847493575595`*^-15", 
      RowBox[{"-", "1.887379141862766`*^-15"}]},
     {
      RowBox[{"-", "8.881784197001252`*^-16"}], "1.1102230246251565`*^-15"},
     {"4.440892098500626`*^-16", "3.3306690738754696`*^-16"},
     {"7.771561172376096`*^-16", "3.885780586188048`*^-16"},
     {
      RowBox[{"-", "1.7763568394002505`*^-15"}], "1.9984014443252818`*^-15"},
     {"1.4988010832439613`*^-15", 
      RowBox[{"-", "6.661338147750939`*^-16"}]},
     {"3.1086244689504383`*^-15", 
      RowBox[{"-", "1.3322676295501878`*^-15"}]},
     {"3.219646771412954`*^-15", 
      RowBox[{"-", "3.3306690738754696`*^-16"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.508658763320639*^9, 3.5086593842345667`*^9, 
  3.508659612919139*^9, 3.508659686871318*^9, 3.508659754276705*^9}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Run the Compiled version", "Subsection"],

Cell[CellGroupData[{

Cell["\<\
loops=200*factor;
s=(ct=GaussSolveForLoopsC[at,bt,loops];)//AbsTime;

mevalForC=s[[1]]/loops;
Print[\"TIMING FOR VERSION (FOR-LOOPS) with Compile[]=\",mevalForC];
Dot[at,ct] - bt // MatrixForm\
\>", "Input",
 PageWidth->Infinity,
 CellChangeTimes->{{3.391404467171875*^9, 3.39140447678125*^9}},
 ShowSpecialCharacters->False],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"TIMING FOR VERSION (FOR-LOOPS) with Compile[]=\"\>", 
   "\[InvisibleSpace]", "0.00028905879999999999999137947`7.213561184574135"}],
  SequenceForm[
  "TIMING FOR VERSION (FOR-LOOPS) with Compile[]=", 
   0.00028905879999999999999137947`7.213561184574135],
  Editable->False]], "Print",
 CellChangeTimes->{
  3.508658763883132*^9, 3.508659384781434*^9, {3.5086595907006736`*^9, 
   3.508659613466007*^9}, 3.508659687152564*^9, 3.508659754886072*^9}],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "3.1086244689504383`*^-15"}], "2.1050869380978554`*^-15"},
     {"8.881784197001252`*^-16", "5.551115123125783`*^-16"},
     {"4.218847493575595`*^-15", 
      RowBox[{"-", "1.887379141862766`*^-15"}]},
     {
      RowBox[{"-", "8.881784197001252`*^-16"}], "1.1102230246251565`*^-15"},
     {"4.440892098500626`*^-16", "3.3306690738754696`*^-16"},
     {"7.771561172376096`*^-16", "3.885780586188048`*^-16"},
     {
      RowBox[{"-", "1.7763568394002505`*^-15"}], "1.9984014443252818`*^-15"},
     {"1.4988010832439613`*^-15", 
      RowBox[{"-", "6.661338147750939`*^-16"}]},
     {"3.1086244689504383`*^-15", 
      RowBox[{"-", "1.3322676295501878`*^-15"}]},
     {"3.219646771412954`*^-15", 
      RowBox[{"-", "3.3306690738754696`*^-16"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{
  3.508658763883132*^9, 3.508659384781434*^9, {3.5086595907006736`*^9, 
   3.508659613466007*^9}, 3.508659687168189*^9, 3.508659754886072*^9}]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Generation of C++ Code", "Section"],

Cell["\<\
The command CompilePackage translates the package to C++ code:\
\>", "Text"],

Cell[CellGroupData[{

Cell["CompilePackage[\"Gauss\"]", "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 2 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5086587666330967`*^9, 3.5086593874064007`*^9}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Building the Executable", "Section"],

Cell["\<\
Call MakeBinary to compile the generated C++ code and build the executable(s).\
\>", "Text"],

Cell["MakeBinary[]", "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Installing Compiled Code", "Section"],

Cell["\<\
Interpreted versions are removed, and compiled ones are used instead:\
\>", "Text"],

Cell[CellGroupData[{

Cell["InstallCode[\"Gauss\"]", "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Gauss\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Gauss", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.508658771867405*^9, 3.508659391046979*^9}],

Cell[BoxData[
 RowBox[{"LinkObject", "[", 
  RowBox[{"\<\"\\\"C:\\\\b\\\\omath\\\\MathCode\\\\Demos\\\\Gauss\\\\Gaussml.\
exe\\\"\"\>", ",", "63119", ",", "8"}], "]"}]], "Output",
 CellChangeTimes->{3.508658771867405*^9, 3.508659391046979*^9}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Test", "Section"],

Cell[BoxData[
 RowBox[{
  RowBox[{"ct", "=", 
   RowBox[{"GaussSolveArraySlice", "[", 
    RowBox[{"at", ",", "bt", ",", "1"}], "]"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Dot", "[", 
    RowBox[{"at", ",", "ct"}], "]"}], "-", "bt"}], "//", 
  "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0.`", 
      RowBox[{"-", "1.1102230246251565`*^-16"}]},
     {"2.498001805406602`*^-16", "2.220446049250313`*^-16"},
     {"2.220446049250313`*^-16", 
      RowBox[{"-", "4.440892098500626`*^-16"}]},
     {"2.7755575615628914`*^-16", "2.220446049250313`*^-16"},
     {"2.7755575615628914`*^-16", "1.1102230246251565`*^-16"},
     {"2.220446049250313`*^-16", "1.6653345369377348`*^-16"},
     {"3.3306690738754696`*^-16", "9.71445146547012`*^-17"},
     {"1.1102230246251565`*^-16", "1.3877787807814457`*^-16"},
     {"1.6653345369377348`*^-16", 
      RowBox[{"-", "1.1102230246251565`*^-16"}]},
     {"0.`", 
      RowBox[{"-", "2.220446049250313`*^-16"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086587718986545`*^9, 3.508659391062604*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"ct", "=", 
   RowBox[{"GaussSolveForLoops", "[", 
    RowBox[{"at", ",", "bt", ",", "1"}], "]"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Dot", "[", 
    RowBox[{"at", ",", "ct"}], "]"}], "-", "bt"}], "//", 
  "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0.`", 
      RowBox[{"-", "1.1102230246251565`*^-16"}]},
     {"2.498001805406602`*^-16", "2.220446049250313`*^-16"},
     {"2.220446049250313`*^-16", 
      RowBox[{"-", "4.440892098500626`*^-16"}]},
     {"2.7755575615628914`*^-16", "2.220446049250313`*^-16"},
     {"2.7755575615628914`*^-16", "1.1102230246251565`*^-16"},
     {"2.220446049250313`*^-16", "1.6653345369377348`*^-16"},
     {"3.3306690738754696`*^-16", "9.71445146547012`*^-17"},
     {"1.1102230246251565`*^-16", "1.3877787807814457`*^-16"},
     {"1.6653345369377348`*^-16", 
      RowBox[{"-", "1.1102230246251565`*^-16"}]},
     {"0.`", 
      RowBox[{"-", "2.220446049250313`*^-16"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.508658771914279*^9, 3.508659391078229*^9}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["External Execution", "Section"],

Cell[CellGroupData[{

Cell["Preperations", "Subsection"],

Cell[CellGroupData[{

Cell["\<\
loops=8000*factor;
externaleval=((ct=GaussSolveArraySlice[at,bt,loops];  )//AbsTime)[[1]]/loops\
\>", "Input"],

Cell[BoxData["0.0001964818599999999999998544`8.647957441227048"], "Output",
 CellChangeTimes->{3.5086587876640778`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=500*factor;
externalevalPass=((Do[ct=GaussSolveArraySlice[at,bt,1],{loops}];)//AbsTime)[[\
1]]/loops\
\>", "Input"],

Cell[BoxData["0.0002999961599999999999933975`7.627630693546707"], "Output",
 CellChangeTimes->{3.5086587892421827`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=5000*factor;
externalevalFor=((ct=GaussSolveForLoops[at,bt,loops];  )//AbsTime)[[1]]/loops\
\>", "Input"],

Cell[BoxData["0.00007093659199999999999961333`8.001385317700263"], "Output",
 CellChangeTimes->{3.5086587928358865`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=500*factor;
externalevalPassFor=((Do[ct=GaussSolveForLoops[at,bt,1],{loops}];)//AbsTime)[[\
1]]/loops\
\>", "Input"],

Cell[BoxData["0.00016562288000000000000267938`7.369635330107926"], "Output",
 CellChangeTimes->{3.5086587937108755`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=10000*factor;
internalEval=((Do[ct=LinearSolve[at,bt];at[[1,1]]=at[[1,1]]+0.001,{loops}];)//\
AbsTime)[[1]]/loops\
\>", "Input"],

Cell[BoxData["0.00005078059999999999999992142`8.157242821486014"], "Output",
 CellChangeTimes->{3.5086587988514347`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Dot[a,c] - b // MatrixForm", "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{
   RowBox[{"-", "b"}], "+", 
   RowBox[{"a", ".", "c"}]}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086587988983088`*^9}]
}, Open  ]],

Cell["UninstallCode[\"Gauss\"];", "Input"],

Cell[CellGroupData[{

Cell["\<\
Map[DeleteFile,FileNames[\"*.o\"]];Map[DeleteFile,FileNames[\"*.obj\"]];
Map[DeleteFile,FileNames[\"*.exe\"]];\
\>", "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"DeleteFile", "::", "privv"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"Privilege violation during \
\[NoBreak]\\!\\(DeleteFile[\\\"Gaussml.exe\\\"]\\)\[NoBreak]. \
\\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\\\", \
ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/General/privv\\\", \
ButtonNote -> \\\"DeleteFile::privv\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.508658799023307*^9}]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Inline C++ Compilation", "Section"],

Cell["\<\
Inline compilation will make matrix operations inline. This will usually \
improve the performance of computations. The drawbacks are increased \
compilation time and size of the binary code.\
\>", "Text"],

Cell["MakeBinary[InlineFlag -> True]", "Input",
 CellChangeTimes->{{3.500371431866663*^9, 3.5003714333978834`*^9}}],

Cell[CellGroupData[{

Cell["InstallCode[\"Gauss\"];", "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Gauss\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Gauss", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.508658815663719*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=8000*factor;
ILexternaleval=((ct=GaussSolveArraySlice[at,bt,loops];  )//AbsTime)[[1]]/loops\
\>", "Input"],

Cell[BoxData["0.00019120849000000000000013987`8.636142152310278"], "Output",
 CellChangeTimes->{3.5086588309916477`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=750*factor;
ILexternalevalPass=((Do[ct=GaussSolveArraySlice[at,bt,1],{loops}];)//AbsTime)[\
[1]]/loops\
\>", "Input"],

Cell[BoxData["0.00029166293333333333333655159`7.791487496185378"], "Output",
 CellChangeTimes->{3.508658833257244*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=5000*factor;
ILexternalevalFor=((ct=GaussSolveForLoops[at,bt,loops];  \
)//AbsTime)[[1]]/loops\
\>", "Input"],

Cell[BoxData["0.00006937411200000000000009948`7.991712434957778"], "Output",
 CellChangeTimes->{3.5086588367728243`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=500*factor;
ILexternalevalPassFor=((Do[ct=GaussSolveForLoops[at,bt,1],{loops}];)//AbsTime)\
[[1]]/loops\
\>", "Input"],

Cell[BoxData["0.00016249792000000000000365166`7.361362804141938"], "Output",
 CellChangeTimes->{3.508658837632188*^9}]
}, Open  ]],

Cell["UninstallCode[\"Gauss\"];", "Input"],

Cell["\<\
Map[DeleteFile,FileNames[\"*.o\"]];Map[DeleteFile,FileNames[\"*.obj\"]];
Map[DeleteFile,FileNames[\"*.exe\"]];\
\>", "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Inline, without Range Check C++ Compilation", "Section"],

Cell["\<\
Disabling range check improves performance even further. The drawback is that \
the program will crash or produce erroneous results instead of reporting an \
error.\
\>", "Text"],

Cell["MakeBinary[InlineFlag -> True, RangeCheckFlag -> False]", "Input",
 CellChangeTimes->{{3.500371622331756*^9, 3.5003716283316407`*^9}}],

Cell[CellGroupData[{

Cell["InstallCode[\"Gauss\"];", "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Gauss\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Gauss", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086588531944885`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=10000*factor;
RLexternaleval=((ct=GaussSolveArraySlice[at,bt,loops];  )//AbsTime)[[1]]/loops\
\>", "Input"],

Cell[BoxData["0.0001620291760000000000000028`8.661138216896182"], "Output",
 CellChangeTimes->{3.508658869413031*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=400*factor;
RLexternalevalPass=((Do[ct=GaussSolveArraySlice[at,bt,1],{loops}];)//AbsTime)[\
[1]]/loops\
\>", "Input"],

Cell[BoxData["0.00025780920000000000000110225`7.464903396049008"], "Output",
 CellChangeTimes->{3.5086588704911423`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=10000*factor;
RLexternalevalFor=((ct=GaussSolveForLoops[at,bt,loops];  \
)//AbsTime)[[1]]/loops\
\>", "Input"],

Cell[BoxData["0.00005296807199999999999978292`8.175559158710223"], "Output",
 CellChangeTimes->{3.508658875834824*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
loops=500*factor;
RLexternalevalPassFor=((Do[ct=GaussSolveForLoops[at,bt,1],{loops}];)//AbsTime)\
[[1]]/loops\
\>", "Input"],

Cell[BoxData["0.0001437481599999999999986433`7.308117292188713"], "Output",
 CellChangeTimes->{3.508658876600439*^9}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Performance Comparison", "Section"],

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"table", "=", GridBox[{
      {"\"\<\>\"", "\"\<Over-\\nhead\>\"", "\"\<Slice\\nTime\\n(sec)\>\"", 
       "\"\<Slice\\nRelative\\nto best C++\>\"", 
       "\"\<Loop\\nTime\\n(sec)\>\"", "\"\<Loop\\nRelative\\nto best C++\>\""},
      {"\"\<Mathematica\>\"", "\"\<-\>\"", "meval", 
       RowBox[{"meval", "/", "RLexternaleval"}], "mevalFor", 
       RowBox[{"mevalFor", "/", "RLexternalevalFor"}]},
      {"\"\<Compile[]\>\"", "\"\<-\>\"", "\"\<-\>\"", "\"\<-\>\"", 
       "mevalForC", 
       RowBox[{"mevalForC", "/", "RLexternalevalFor"}]},
      {"\"\<C++\>\"", "\"\<Yes\>\"", "externalevalPass", 
       RowBox[{"externalevalPass", "/", "RLexternaleval"}], 
       "externalevalPassFor", 
       RowBox[{"externalevalPassFor", "/", "RLexternalevalFor"}]},
      {"\"\<C++\>\"", "\"\<No\>\"", "externaleval", 
       RowBox[{"externaleval", "/", "RLexternaleval"}], "externalevalFor", 
       RowBox[{"externalevalFor", "/", "RLexternalevalFor"}]},
      {"\"\<C++, inline\>\"", "\"\<Yes\>\"", "ILexternalevalPass", 
       RowBox[{"ILexternalevalPass", "/", "RLexternaleval"}], 
       "ILexternalevalPassFor", 
       RowBox[{"ILexternalevalPassFor", "/", "RLexternalevalFor"}]},
      {"\"\<C++, inline\>\"", "\"\<No\>\"", "ILexternaleval", 
       RowBox[{"ILexternaleval", "/", "RLexternaleval"}], "ILexternalevalFor", 
       RowBox[{"ILexternalevalFor", "/", "RLexternalevalFor"}]},
      {"\"\<C++, norange\>\"", "\"\<Yes\>\"", "RLexternalevalPass", 
       RowBox[{"RLexternalevalPass", "/", "RLexternaleval"}], 
       "RLexternalevalPassFor", 
       RowBox[{"RLexternalevalPassFor", "/", "RLexternalevalFor"}]},
      {"\"\<C++, norange\>\"", "\"\<No\>\"", "RLexternaleval", 
       RowBox[{"RLexternaleval", "/", "RLexternaleval"}], "RLexternalevalFor", 
       RowBox[{"RLexternalevalFor", "/", "RLexternalevalFor"}]},
      {"\"\<LinearSolve[]\>\"", "\"\<-\>\"", "\"\<-\>\"", "\"\<-\>\"", 
       "internalEval", 
       RowBox[{"internalEval", "/", "RLexternalevalFor"}]}
     }]}], ";"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"FrameBox", "[", 
   RowBox[{"GridBox", "[", 
    RowBox[{
     RowBox[{"Map", "[", 
      RowBox[{"ToString", ",", "table", ",", 
       RowBox[{"{", "2", "}"}]}], "]"}], ",", 
     RowBox[{"RowLines", "\[Rule]", "True"}], ",", 
     RowBox[{"ColumnLines", "\[Rule]", "True"}]}], "]"}], "]"}], "//", 
  "DisplayForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  FrameBox[GridBox[{
     {"", 
      RowBox[{"Over", "-", "\n", "head"}], 
      RowBox[{"Slice", "\n", "Time", "\n", 
       RowBox[{"(", "sec", ")"}]}], 
      RowBox[{"Slice", "\n", "Relative", "\n", 
       RowBox[{"to", " ", "best", " ", 
        RowBox[{"C", "++"}]}]}], 
      RowBox[{"Loop", "\n", "Time", "\n", 
       RowBox[{"(", "sec", ")"}]}], 
      RowBox[{"Loop", "\n", "Relative", "\n", 
       RowBox[{"to", " ", "best", " ", 
        RowBox[{"C", "++"}]}]}]},
     {"Mathematica", "-", "0.11718600", "723.2401", "0.05937424", "1120.9440"},
     {
      RowBox[{"Compile", "[", "]"}], "-", "-", "-", "0.0002734340", 
      "5.162242"},
     {
      RowBox[{"C", "++"}], "Yes", "0.00029999616", "1.8514947", 
      "0.00016562288", "3.126844"},
     {
      RowBox[{"C", "++"}], "No", "0.000196481860", "1.21263259", 
      "0.000070936592", "1.3392330"},
     {
      RowBox[{
       RowBox[{"C", "++"}], ",", " ", "inline"}], "Yes", "0.00029166293", 
      "1.8000643", "0.00016249792", "3.067847"},
     {
      RowBox[{
       RowBox[{"C", "++"}], ",", " ", "inline"}], "No", "0.000191208490", 
      "1.18008679", "0.000069374112", "1.3097345"},
     {
      RowBox[{
       RowBox[{"C", "++"}], ",", " ", "norange"}], "Yes", "0.00025780920", 
      "1.5911283", "0.00014374816", "2.713864"},
     {
      RowBox[{
       RowBox[{"C", "++"}], ",", " ", "norange"}], "No", "0.000162029176", 
      "1.00000000", "0.000052968072", "1.0000000"},
     {
      RowBox[{"LinearSolve", "[", "]"}], "-", "-", "-", "0.000050780600", 
      "0.9587021"}
    },
    GridBoxDividers->{
     "Columns" -> {False, {True}, False}, "ColumnsIndexed" -> {}, 
      "Rows" -> {False, {True}, False}, "RowsIndexed" -> {}}]],
  DisplayForm]], "Output",
 CellChangeTimes->{3.5086588766473136`*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  TagBox[
   RowBox[{
    RowBox[{
     RowBox[{
     "These", " ", "values", " ", "are", " ", "obtained", " ", "on", " ", "a",
       "  ", "2.6", " ", "GHz", " ", "computer", " ", "with", 
      "       ", "\<\"MathCode 1.4.4\"\>"}], " ", ",", 
     "\"\<  Visual C++ 10.0\>\"", ",", "      ", "\<\"Mathematica 8.0\"\>"}], 
    " ", "\[IndentingNewLine]", 
    FrameBox[GridBox[{
       {"", 
        RowBox[{"Over", "-", "\n", "head"}], 
        RowBox[{"Slice", "\n", "Time", "\n", 
         RowBox[{"(", "sec", ")"}]}], 
        RowBox[{"Slice", "\n", "Relative", "\n", 
         RowBox[{"to", " ", "best", " ", 
          RowBox[{"C", "++"}]}]}], 
        RowBox[{"Loop", "\n", "Time", "\n", 
         RowBox[{"(", "sec", ")"}]}], 
        RowBox[{"Loop", "\n", "Relative", "\n", 
         RowBox[{"to", " ", "best", " ", 
          RowBox[{"C", "++"}]}]}]},
       {"Mathematica", "-", "0.11718600", "723.2401", "0.05937424", 
        "1120.9440"},
       {
        RowBox[{"Compile", "[", "]"}], "-", "-", "-", "0.0002734340", 
        "5.162242"},
       {
        RowBox[{"C", "++"}], "Yes", "0.00029999616", "1.8514947", 
        "0.00016562288", "3.126844"},
       {
        RowBox[{"C", "++"}], "No", "0.000196481860", "1.21263259", 
        "0.000070936592", "1.3392330"},
       {
        RowBox[{
         RowBox[{"C", "++"}], ",", " ", "inline"}], "Yes", "0.00029166293", 
        "1.8000643", "0.00016249792", "3.067847"},
       {
        RowBox[{
         RowBox[{"C", "++"}], ",", " ", "inline"}], "No", "0.000191208490", 
        "1.18008679", "0.000069374112", "1.3097345"},
       {
        RowBox[{
         RowBox[{"C", "++"}], ",", " ", "norange"}], "Yes", "0.00025780920", 
        "1.5911283", "0.00014374816", "2.713864"},
       {
        RowBox[{
         RowBox[{"C", "++"}], ",", " ", "norange"}], "No", "0.000162029176", 
        "1.00000000", "0.000052968072", "1.0000000"},
       {
        RowBox[{"LinearSolve", "[", "]"}], "-", "-", "-", "0.000050780600", 
        "0.9587021"}
      },
      GridBoxDividers->{
       "Columns" -> {False, {True}, False}, "ColumnsIndexed" -> {}, 
        "Rows" -> {False, {True}, False}, "RowsIndexed" -> {}}]]}],
   DisplayForm], "\[IndentingNewLine]"}]], "Text",
 CellChangeTimes->{{3.5086589579118986`*^9, 3.5086590208485928`*^9}, 
   3.508659059191852*^9}],

Cell[BoxData[
 TagBox[
  RowBox[{
   RowBox[{
    RowBox[{
    "These", " ", "values", " ", "are", " ", "obtained", "  ", "on", " ", "a",
      " ", "2.6", " ", "GHz", " ", "computer", " ", "with", 
     "       ", "\<\"MathCode 1.4.4\"\>"}], " ", 
    ",", "\<\"  Mingw gcc 4.5.0\"\>", ",", 
    "      ", "\<\"Mathematica 6.0\"\>"}], " ", "\[IndentingNewLine]", 
   FrameBox[GridBox[{
      {"", 
       RowBox[{"Over", "-", "\n", "head"}], 
       RowBox[{"Slice", "\n", "Time", "\n", 
        RowBox[{"(", "sec", ")"}]}], 
       RowBox[{"Slice", "\n", "Relative", "\n", 
        RowBox[{"to", " ", "best", " ", 
         RowBox[{"C", "++"}]}]}], 
       RowBox[{"Loop", "\n", "Time", "\n", 
        RowBox[{"(", "sec", ")"}]}], 
       RowBox[{"Loop", "\n", "Relative", "\n", 
        RowBox[{"to", " ", "best", " ", 
         RowBox[{"C", "++"}]}]}]},
      {"Mathematica", "-", "0.08906079", "4222.222", "0.05156151", "5238.095"},
      {
       RowBox[{"Compile", "[", "]"}], "-", "-", "-", "0.0005312398", 
       "53.96825"},
      {
       RowBox[{"C", "++"}], "Yes", "0.00024687026", "11.703704", 
       "0.00012499760", "12.69841"},
      {
       RowBox[{"C", "++"}], "No", "0.000156051691", "7.398148", 
       "0.000031249400", "3.174603"},
      {
       RowBox[{
        RowBox[{"C", "++"}], ",", " ", "inline"}], "Yes", "0.00013124748", 
       "6.222222", "0.00011562278", "11.74603"},
      {
       RowBox[{
        RowBox[{"C", "++"}], ",", " ", "inline"}], "No", "0.000038866441", 
       "1.8425926", "0.000022499568", "2.285714"},
      {
       RowBox[{
        RowBox[{"C", "++"}], ",", " ", "norange"}], "Yes", "0.00011327908", 
       "5.370370", "0.00009999808", "10.15873"},
      {
       RowBox[{
        RowBox[{"C", "++"}], ",", " ", "norange"}], "No", "0.000021093345", 
       "1.000000", "0.000009843", "1.0000000"},
      {
       RowBox[{"LinearSolve", "[", "]"}], "-", "-", "-", "0.000068279939", 
       "6.936508"}
     },
     GridBoxDividers->{
      "Columns" -> {False, {True}, False}, "ColumnsIndexed" -> {}, 
       "Rows" -> {False, {True}, False}, "RowsIndexed" -> {}}]]}],
  DisplayForm]], "Text",
 CellChangeTimes->{
  3.5003715163806653`*^9, 3.5003723061936255`*^9, {3.500372482596489*^9, 
   3.5003725591106443`*^9}, {3.5003726504838905`*^9, 
   3.5003726793427114`*^9}, {3.5086591912370367`*^9, 3.5086592275021973`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
  "These", " ", "values", " ", "were", " ", "obtained", " ", "on", " ", "a", 
   " ", "2.8", " ", "GHz", " ", "computer", " ", "with", " ", "MathCode", " ",
    "1.2", ".2"}], ",", " ", 
  RowBox[{
   RowBox[{"VisualC", "++"}], " ", "6.0"}], " ", ",", " ", 
  RowBox[{"Mathematica", " ", "5.1", " ", 
   RowBox[{"installed", ".", "\[IndentingNewLine]", " ", 
    TagBox[
     FrameBox[GridBox[{
        {"", 
         RowBox[{"Over", "-", "\n", "head"}], 
         RowBox[{"Slice", "\n", "Time", "\n", 
          RowBox[{"(", "sec", ")"}]}], 
         RowBox[{"Slice", "\n", "Relative", "\n", 
          RowBox[{"to", " ", "best", " ", 
           RowBox[{"C", "++"}]}]}], 
         RowBox[{"Loop", "\n", "Time", "\n", 
          RowBox[{"(", "sec", ")"}]}], 
         RowBox[{"Loop", "\n", "Relative", "\n", 
          RowBox[{"to", " ", "best", " ", 
           RowBox[{"C", "++"}]}]}]},
        {"Mathematica", "-", "0.1523", "305.114", "0.0951", "486.952"},
        {
         RowBox[{"Compile", "[", "]"}], "-", "-", "-", "0.0009265", "4.74407"},
        {
         RowBox[{"C", "++"}], "Yes", "0.00166853232", "3.3426966", 
         "0.00149226480", "7.6410256"},
        {
         RowBox[{"C", "++"}], "No", "0.000309720060", "0.62048555", 
         "0.000124388784", "0.63692308"},
        {
         RowBox[{
          RowBox[{"C", "++"}], ",", " ", "inline"}], "Yes", "0.00168789504", 
         "3.3814874", "0.00254185776", "13.0153846"},
        {
         RowBox[{
          RowBox[{"C", "++"}], ",", " ", "inline"}], "No", "0.000314602470", 
         "0.63026685", "0.000127393344", "0.65230769"},
        {
         RowBox[{
          RowBox[{"C", "++"}], ",", " ", "norange"}], "Yes", "0.00389340900", 
         "7.7999599", "0.00336310416", "17.2205128"},
        {
         RowBox[{
          RowBox[{"C", "++"}], ",", " ", "norange"}], "No", "0.000499157568", 
         "1.00000000", "0.000195296400", "1.00000000"},
        {
         RowBox[{"LinearSolve", "[", "]"}], "-", "-", "-", "0.000073611720", 
         "0.37692308"}
       }]],
     DisplayForm]}]}]}]], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Clean Up", "Section"],

Cell["\<\
UninstallCode[\"Gauss\"];
CleanMathCodeFiles[Confirm \[Rule] False, CleanAllBut \[Rule] {}]\
\>", "Input",
 CellChangeTimes->{3.508659935883755*^9}]
}, Open  ]]
},
ScreenStyleEnvironment->"Working",
WindowToolbars->{},
WindowSize->{1272, 922},
WindowMargins->{{21, Automatic}, {Automatic, -3}},
ShowSelection->True,
FrontEndVersion->"6.0 for Microsoft Windows (32-bit) (April 20, 2007)",
StyleDefinitions->Notebook[{
   Cell[
    CellGroupData[{
      Cell["Style Definitions", "Subtitle"], 
      Cell[
      "Modify the definitions below to change the default appearance of all \
cells in a given style.  Make modifications to any definition using commands \
in the Format menu.", "Text"], 
      Cell[
       CellGroupData[{
         Cell["Style Environment Names", "Section"], 
         Cell[
          StyleData[All, "Working"], PageWidth -> WindowWidth, ScriptMinSize -> 
          9], 
         Cell[
          StyleData[All, "Presentation"], PageWidth -> WindowWidth, 
          ScriptMinSize -> 12, FontSize -> 16], 
         Cell[
          StyleData[All, "Condensed"], PageWidth -> WindowWidth, 
          CellBracketOptions -> {"Margins" -> {1, 1}, "Widths" -> {0, 5}}, 
          ScriptMinSize -> 8, FontSize -> 11], 
         Cell[
          StyleData[All, "Printout"], PageWidth -> PaperWidth, ScriptMinSize -> 
          7, FontSize -> 10, 
          PrivateFontOptions -> {"FontType" -> "Outline"}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Notebook Options", "Section"], 
         Cell[
         "The options defined for the style below will be used at the \
Notebook level.", "Text"], 
         Cell[
          StyleData["Notebook"], PageHeaders -> {{
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"], None, 
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"]}, {
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"], None, 
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"]}}, 
          PageHeaderLines -> {True, True}, 
          PrintingOptions -> {
           "FacingPages" -> True, "FirstPageHeader" -> False}, 
          CellFrameLabelMargins -> 6, StyleMenuListing -> None]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headings", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 0}, CounterIncrements -> "Title", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subtitle", 0}, {"Subsubtitle", 0}}, FontSize -> 
             34, FontColor -> GrayLevel[1], Background -> RGBColor[0, 0, 1]], 
            
            Cell[
             StyleData["Title", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 0}, FontSize -> 
             44], 
            Cell[
             StyleData["Title", "Condensed"], CellMargins -> {{0, 0}, {0, 0}},
              FontSize -> 20], 
            Cell[
             StyleData["Title", "Printout"], CellMargins -> {{0, 0}, {0, 0}}, 
             FontSize -> 24, FontTracking -> "Plain", Background -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 3}, ParagraphIndent -> -96, CounterIncrements -> 
             "Subtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subsubtitle", 0}}, FontFamily -> "Helvetica", 
             FontSize -> 18, FontColor -> GrayLevel[1], Background -> 
             RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Subtitle", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 6}, 
             ParagraphIndent -> -157, FontSize -> 30], 
            Cell[
             StyleData["Subtitle", "Condensed"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -78, 
             FontSize -> 14], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -85, 
             FontSize -> 16, Background -> GrayLevel[0.6]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {30, 10}}, PageBreakBelow -> False, 
             CounterIncrements -> "Subsubtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}}, FontFamily -> "Helvetica", FontSize -> 14, 
             FontSlant -> "Italic"], 
            Cell[
             StyleData["Subsubtitle", "Presentation"], 
             CellMargins -> {{8, 10}, {40, 20}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Subsubtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 8}}, FontSize -> 12], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{9, 10}, {50, 10}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], CellFrame -> {{6, 0}, {0, 1}}, CellDingbat -> 
             None, CellMargins -> {{12, Inherited}, {4, 24}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameMargins -> 6, CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             FontFamily -> "Helvetica", FontSize -> 16, FontWeight -> "Bold", 
             FontColor -> RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Section", "Presentation"], 
             CellMargins -> {{10, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["Section", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Section", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, FontSize -> 14, FontTracking -> 
             "Plain", FontColor -> GrayLevel[0]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], 
             CellMargins -> {{12, Inherited}, {8, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsection", "Presentation"], 
             CellMargins -> {{11, 10}, {8, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 22], 
            Cell[
             StyleData["Subsection", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Subsection", "Printout"], 
             CellMargins -> {{9, 0}, {4, 40}}, FontSize -> 12]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> "\[FilledSquare]", 
             CellMargins -> {{25, Inherited}, {8, 12}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsubsection", "Presentation"], 
             CellMargins -> {{29, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Subsubsection", "Condensed"], 
             CellMargins -> {{22, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Subsubsection", "Printout"], 
             CellMargins -> {{21, 0}, {4, 20}}, FontSize -> 11]}, Open]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Body Text", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Text"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Text", FontFamily -> "Times"], 
            Cell[
             StyleData["Text", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Text", "Condensed"], CellMargins -> {{8, 10}, {4, 4}},
              LineSpacing -> {1, 1}, ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Text", "Printout"], CellMargins -> {{9, 0}, {4, 4}}, 
             ParagraphSpacing -> {0, 6}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SmallText"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 6}, 
             CounterIncrements -> "SmallText", FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["SmallText", "Condensed"], 
             CellMargins -> {{8, 10}, {2, 2}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Printout"], 
             CellMargins -> {{9, 0}, {4, 4}}, FontSize -> 7]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Input/Output", "Section"], 
         Cell[
         "The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Input"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             CounterIncrements -> "Input", FontWeight -> "Bold", Background -> 
             RGBColor[1, 1, 0]], 
            Cell[
             StyleData["Input", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Input", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["Input", "Printout"], CellMargins -> {{44, 0}, {6, 6}},
              Background -> GrayLevel[0.8]]}, Closed]], 
         Cell[
          StyleData["InlineInput"], Evaluatable -> True, CellGroupingRules -> 
          "InputGrouping", CellHorizontalScrolling -> True, PageBreakWithin -> 
          False, GroupPageBreakWithin -> False, DefaultFormatType -> 
          DefaultInputFormatType, AutoItalicWords -> {}, FormatType -> 
          InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
          CounterIncrements -> "Input", FontWeight -> "Bold"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Output"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, CellEditDuplicate -> True, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, 
             CellLabelMargins -> {{3, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Output", Background -> 
             RGBColor[1, 0.900008, 0.900008]], 
            Cell[
             StyleData["Output", "Presentation"], 
             CellMargins -> {{62, Inherited}, {12, 5}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Output", "Condensed"], 
             CellMargins -> {{40, Inherited}, {4, 1}}], 
            Cell[
             StyleData["Output", "Printout"], 
             CellMargins -> {{44, 0}, {6, 2}}, Background -> 
             GrayLevel[0.900008]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Message"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Message", StyleMenuListing -> 
             None, FontColor -> RGBColor[1, 0, 0]], 
            Cell[
             StyleData["Message", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Message", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Message", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Print"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, ShowCellLabel -> 
             False, DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Print", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Print", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Print", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Print", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Graphics"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "GraphicsGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Graphics", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Graphics", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Graphics", "Condensed"], 
             CellMargins -> {{52, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {175, 175}], 
            Cell[
             StyleData["Graphics", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {250, 250}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["CellLabel"], StyleMenuListing -> None, FontFamily -> 
             "Helvetica", FontSize -> 11, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["CellLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["CellLabel", "Condensed"], FontSize -> 8], 
            Cell[
             StyleData["CellLabel", "Printout"], FontSize -> 8, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Unique Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {2, 10}}, FontSize -> 16, FontSlant -> 
             "Italic"], 
            Cell[
             StyleData["Author", "Presentation"], 
             CellMargins -> {{12, 10}, {2, 12}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}, FontSize -> 20], 
            Cell[
             StyleData["Author", "Condensed"], 
             CellMargins -> {{8, 10}, {1, 4}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}, FontSize -> 12], 
            Cell[
             StyleData["Author", "Printout"], 
             CellMargins -> {{9, 0}, {4, 12}}, ParagraphSpacing -> {0, 6}, 
             FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Copyright"], ShowCellBracket -> False, 
             CellMargins -> {{10, 10}, {40, 2}}, FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Presentation"], 
             CellMargins -> {{12, 10}, {50, 2}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["Copyright", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 1}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Printout"], 
             CellMargins -> {{9, 0}, {72, 4}}, FontSize -> 7]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Example"], CellMargins -> {{12, 10}, {5, 12}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Example", FontFamily -> "Times", 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["Example", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 20}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Example", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 8}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Example", "Printout"], 
             CellMargins -> {{9, 0}, {4, 10}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Definition"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Definition", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Definition", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Definition", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Definition", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.6]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Theorem"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Theorem", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Theorem", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Theorem", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Theorem", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.4]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseMain"], CellFrame -> {{6, 0}, {0, 1}}, 
             CellMargins -> {{12, Inherited}, {4, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameColor -> RGBColor[0.571389, 0.19675, 0.570504], 
             CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 16, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["ExerciseMain", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["ExerciseMain", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["ExerciseMain", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, CellFrameColor -> 
             GrayLevel[0.500008], FontSize -> 14, FontTracking -> "Plain", 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Exercise"], CellDingbat -> "\[FilledDownTriangle]", 
             CellMargins -> {{23, Inherited}, {4, 18}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Exercise", "Presentation"], 
             CellMargins -> {{33, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Exercise", "Condensed"], 
             CellMargins -> {{17, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Exercise", "Printout"], 
             CellFrame -> {{0, 0}, {0.5, 0}}, CellDingbat -> None, 
             CellMargins -> {{9, 0}, {6, 20}}, FontSize -> 11, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseText"], CellMargins -> {{62, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 8}, 
             CounterIncrements -> "ExcersiceText", FontFamily -> "Times"], 
            Cell[
             StyleData["ExerciseText", "Presentation"], 
             CellMargins -> {{74, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["ExerciseText", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 2}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["ExerciseText", "Printout"], 
             CellMargins -> {{54, 0}, {4, 4}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Itemize"], CellDingbat -> "\[FilledSmallCircle]", 
             CellMargins -> {{65, Inherited}, {8, 8}}, FontFamily -> "Times", 
             FontSize -> 13, FontWeight -> "Plain", FontSlant -> "Plain", 
             FontTracking -> "Plain", 
             FontVariations -> {
              "Outline" -> False, "Shadow" -> False, "Underline" -> False}], 
            Cell[
             StyleData["Itemize", "Presentation"], 
             CellMargins -> {{Inherited, 10}, {8, 12}}, LineSpacing -> {1, 0},
              FontSize -> 18], 
            Cell[
             StyleData["Itemize", "Condensed"], 
             CellMargins -> {{Inherited, Inherited}, {2, 12}}, FontSize -> 
             10], 
            Cell[
             StyleData["Itemize", "Printout"], 
             CellMargins -> {{Inherited, 0}, {4, 20}}, FontSize -> 11]}, 
           Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayCode"], CellFrame -> 2, 
             CellMargins -> {{52, 10}, {8, 8}}, CellHorizontalScrolling -> 
             True, PageBreakWithin -> True, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             FontWeight -> "Bold", Background -> RGBColor[1, 1, 0]], 
            Cell[
             StyleData["DisplayCode", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["DisplayCode", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["DisplayCode", "Printout"], 
             CellMargins -> {{44, 0}, {6, 6}}, Background -> GrayLevel[0.8]]},
            Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Formulas and Programming", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayFormula"], CellMargins -> {{62, 10}, {2, 10}}, 
             CellHorizontalScrolling -> True, ScriptLevel -> 0, 
             SingleLetterItalics -> True, 
             UnderoverscriptBoxOptions -> {LimitsPositioning -> True}], 
            Cell[
             StyleData["DisplayFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 18], 
            Cell[
             StyleData["DisplayFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["DisplayFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ChemicalFormula"], CellMargins -> {{62, 10}, {2, 10}},
              AutoSpacing -> False, ScriptBaselineShifts -> {0.6, Automatic}, 
             ScriptLevel -> 1, SingleLetterItalics -> False, ZeroWidthTimes -> 
             True], 
            Cell[
             StyleData["ChemicalFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Program"], CellMargins -> {{12, 10}, {Inherited, 6}}, 
             FontFamily -> "Courier"], 
            Cell[
             StyleData["Program", "Presentation"], 
             CellMargins -> {{13, 30}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Condensed"], 
             CellMargins -> {{8, 10}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Printout"], 
             CellMargins -> {{9, 0}, {Inherited, 4}}, FontSize -> 9.5]}, 
           Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Automatic Numbering", "Section"], 
         Cell[
         "The following styles are useful for numbered equations, figures, \
etc.  They automatically give the cell a FrameLabel containing a reference to \
a particular counter, and also increment that counter.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedEquation"], 
             CellMargins -> {{62, 10}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, 
                Cell[
                 TextData[{"(", 
                   CounterBox["NumberedEquation"], ")"}]]}, {None, None}}, 
             DefaultFormatType -> DefaultInputFormatType, CounterIncrements -> 
             "NumberedEquation", FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedEquation", "Presentation"], 
             CellMargins -> {{74, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Condensed"], 
             CellMargins -> {{52, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Printout"], 
             CellMargins -> {{54, 0}, {Inherited, Inherited}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedFigure"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Figure ", 
                   CounterBox["NumberedFigure"]}]], None}}, CounterIncrements -> 
             "NumberedFigure", 
             ImageMargins -> {{43, Inherited}, {Inherited, 0}}, 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedFigure", "Presentation"]], 
            Cell[
             StyleData["NumberedFigure", "Condensed"]], 
            Cell[
             StyleData["NumberedFigure", "Printout"]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedTable"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Table ", 
                   CounterBox["NumberedTable"]}]], None}}, TextAlignment -> 
             Center, CounterIncrements -> "NumberedTable", 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedTable", "Presentation"]], 
            Cell[
             StyleData["NumberedTable", "Condensed"]], 
            Cell[
             StyleData["NumberedTable", "Printout"]]}, Open]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headers and Footers", "Section"], 
         Cell[
          StyleData["Header"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["Footer"], CellMargins -> {{0, 0}, {0, 4}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 6],
          
         Cell[
          StyleData["PageNumber"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontWeight -> "Bold"]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Hyperlink Styles", "Section"], 
         Cell[
         "The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Hyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, 
             FontVariations -> {"Underline" -> True}, FontColor -> 
             RGBColor[0, 0, 1], Background -> None, 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonNote -> 
               ButtonData}], 
            Cell[
             StyleData["Hyperlink", "Presentation"]], 
            Cell[
             StyleData["Hyperlink", "Condensed"]], 
            Cell[
             StyleData["Hyperlink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Open]], 
         Cell[
         "The following styles are for linking automatically to the on-line \
help system.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MainBookLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MainBook", #]}]& )}], 
            Cell[
             StyleData["MainBookLink", "Presentation"]], 
            Cell[
             StyleData["MainBookLink", "Condensed"]], 
            Cell[
             StyleData["MainBookLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["AddOnsLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["AddOns", #]}]& )}], 
            Cell[
             StyleData["AddOnsLink", "Presentation"]], 
            Cell[
             StyleData["AddOnsLink", "Condensed"]], 
            Cell[
             StyleData["AddOnLink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuideLink", #]}]& )}], 
            Cell[
             StyleData["RefGuideLink", "Presentation"]], 
            Cell[
             StyleData["RefGuideLink", "Condensed"]], 
            Cell[
             StyleData["RefGuideLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["GettingStartedLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["GettingStarted", #]}]& )}], 
            Cell[
             StyleData["GettingStartedLink", "Presentation"]], 
            Cell[
             StyleData["GettingStartedLink", "Condensed"]], 
            Cell[
             StyleData["GettingStartedLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["OtherInformationLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["OtherInformation", #]}]& )}], 
            Cell[
             StyleData["OtherInformationLink", "Presentation"]], 
            Cell[
             StyleData["OtherInformationLink", "Condensed"]], 
            Cell[
             StyleData["OtherInformationLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]]}, Closed]]},
      Open]]}, Visible -> False, FrontEndVersion -> 
  "6.0 for Microsoft Windows (32-bit) (April 20, 2007)", StyleDefinitions -> 
  "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[580, 21, 46, 0, 55, "Title"],
Cell[629, 23, 73, 2, 54, "Copyright"],
Cell[CellGroupData[{
Cell[727, 29, 31, 0, 54, "Section"],
Cell[761, 31, 270, 5, 25, "Text"],
Cell[1034, 38, 173, 5, 31, "Itemize"],
Cell[1210, 45, 142, 4, 31, "Itemize"],
Cell[1355, 51, 333, 7, 43, "Text"]
}, Closed]],
Cell[CellGroupData[{
Cell[1725, 63, 33, 0, 34, "Section"],
Cell[CellGroupData[{
Cell[1783, 67, 47, 0, 44, "Subsection"],
Cell[1833, 69, 108, 3, 25, "Text"],
Cell[CellGroupData[{
Cell[1966, 76, 74, 1, 48, "Input"],
Cell[2043, 79, 499, 11, 70, "Print"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2591, 96, 45, 0, 44, "Subsection"],
Cell[2639, 98, 155, 3, 25, "Text"],
Cell[2797, 103, 111, 3, 48, "Input"],
Cell[2911, 108, 68, 0, 25, "Text"],
Cell[2982, 110, 42, 0, 48, "Input"],
Cell[3027, 112, 37, 0, 48, "Input"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3113, 118, 42, 0, 34, "Section"],
Cell[3158, 120, 61, 0, 48, "Input"],
Cell[3222, 122, 39, 0, 25, "Text"],
Cell[3264, 124, 114, 7, 133, "Input"],
Cell[CellGroupData[{
Cell[3403, 135, 45, 0, 44, "Subsection"],
Cell[3451, 137, 99, 2, 25, "Text"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3599, 145, 48, 0, 34, "Section"],
Cell[3650, 147, 38, 0, 48, "Input"],
Cell[CellGroupData[{
Cell[3713, 151, 55, 0, 44, "Subsection"],
Cell[3771, 153, 2521, 88, 1459, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[6329, 246, 54, 0, 28, "Subsection"],
Cell[6386, 248, 2459, 87, 1459, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[8882, 340, 82, 1, 28, "Subsection"],
Cell[8967, 343, 2763, 98, 1595, "Input"],
Cell[11733, 443, 45, 3, 65, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[11815, 451, 70, 0, 44, "Subsection"],
Cell[11888, 453, 43, 0, 25, "Text"],
Cell[11934, 455, 83, 3, 65, "Input"],
Cell[12020, 460, 168, 7, 133, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[12225, 472, 48, 0, 28, "Subsection"],
Cell[12276, 474, 242, 6, 25, "Text"],
Cell[12521, 482, 73, 2, 48, "Input"],
Cell[CellGroupData[{
Cell[12619, 488, 256, 7, 116, "Input"],
Cell[12878, 497, 402, 8, 70, "Print"],
Cell[13283, 507, 1464, 34, 70, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[14796, 547, 46, 0, 28, "Subsection"],
Cell[CellGroupData[{
Cell[14867, 551, 274, 7, 116, "Input"],
Cell[15144, 560, 426, 9, 70, "Print"],
Cell[15573, 571, 1460, 34, 70, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[17082, 611, 46, 0, 28, "Subsection"],
Cell[CellGroupData[{
Cell[17153, 615, 335, 10, 133, "Input"],
Cell[17491, 627, 498, 10, 23, "Print"],
Cell[17992, 639, 1488, 35, 225, "Output"]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[19541, 681, 41, 0, 34, "Section"],
Cell[19585, 683, 86, 2, 25, "Text"],
Cell[CellGroupData[{
Cell[19696, 689, 42, 0, 48, "Input"],
Cell[19741, 691, 146, 1, 23, "Print"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[19936, 698, 42, 0, 34, "Section"],
Cell[19981, 700, 102, 2, 25, "Text"],
Cell[20086, 704, 29, 0, 48, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[20152, 709, 43, 0, 34, "Section"],
Cell[20198, 711, 93, 2, 25, "Text"],
Cell[CellGroupData[{
Cell[20316, 717, 39, 0, 48, "Input"],
Cell[20358, 719, 246, 5, 23, "Print"],
Cell[20607, 726, 243, 4, 47, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[20899, 736, 23, 0, 34, "Section"],
Cell[20925, 738, 153, 4, 48, "Input"],
Cell[CellGroupData[{
Cell[21103, 746, 144, 5, 48, "Input"],
Cell[21250, 753, 1284, 30, 225, "Output"]
}, Open  ]],
Cell[22549, 786, 151, 4, 48, "Input"],
Cell[CellGroupData[{
Cell[22725, 794, 144, 5, 48, "Input"],
Cell[22872, 801, 1282, 30, 225, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[24203, 837, 37, 0, 34, "Section"],
Cell[CellGroupData[{
Cell[24265, 841, 34, 0, 44, "Subsection"],
Cell[CellGroupData[{
Cell[24324, 845, 120, 3, 65, "Input"],
Cell[24447, 850, 119, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[24603, 856, 131, 4, 65, "Input"],
Cell[24737, 862, 119, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[24893, 868, 121, 3, 65, "Input"],
Cell[25017, 873, 120, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[25174, 879, 132, 4, 65, "Input"],
Cell[25309, 885, 120, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[25466, 891, 144, 4, 65, "Input"],
Cell[25613, 897, 120, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[25770, 903, 43, 0, 48, "Input"],
Cell[25816, 905, 200, 7, 63, "Output"]
}, Open  ]],
Cell[26031, 915, 42, 0, 48, "Input"],
Cell[CellGroupData[{
Cell[26098, 919, 135, 3, 65, "Input"],
Cell[26236, 924, 479, 10, 21, "Message"]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[26776, 941, 41, 0, 34, "Section"],
Cell[26820, 943, 215, 4, 25, "Text"],
Cell[27038, 949, 115, 1, 48, "Input"],
Cell[CellGroupData[{
Cell[27178, 954, 40, 0, 48, "Input"],
Cell[27221, 956, 224, 5, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[27482, 966, 122, 3, 65, "Input"],
Cell[27607, 971, 120, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[27764, 977, 133, 4, 65, "Input"],
Cell[27900, 983, 118, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[28055, 989, 125, 4, 65, "Input"],
Cell[28183, 995, 120, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[28340, 1001, 134, 4, 65, "Input"],
Cell[28477, 1007, 118, 1, 47, "Output"]
}, Open  ]],
Cell[28610, 1011, 42, 0, 48, "Input"],
Cell[28655, 1013, 135, 3, 65, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[28827, 1021, 62, 0, 34, "Section"],
Cell[28892, 1023, 188, 4, 25, "Text"],
Cell[29083, 1029, 140, 1, 48, "Input"],
Cell[CellGroupData[{
Cell[29248, 1034, 40, 0, 48, "Input"],
Cell[29291, 1036, 226, 5, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[29554, 1046, 123, 3, 65, "Input"],
Cell[29680, 1051, 117, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[29834, 1057, 133, 4, 65, "Input"],
Cell[29970, 1063, 120, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[30127, 1069, 126, 4, 65, "Input"],
Cell[30256, 1075, 118, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[30411, 1081, 134, 4, 65, "Input"],
Cell[30548, 1087, 117, 1, 47, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[30714, 1094, 41, 0, 34, "Section"],
Cell[30758, 1096, 2093, 37, 212, "Input"],
Cell[CellGroupData[{
Cell[32876, 1137, 364, 10, 48, "Input"],
Cell[33243, 1149, 1821, 49, 259, "Output"]
}, Open  ]],
Cell[35079, 1201, 2360, 60, 284, "Text"],
Cell[37442, 1263, 2378, 60, 252, "Text"],
Cell[39823, 1325, 2128, 53, 242, "Text"]
}, Closed]],
Cell[CellGroupData[{
Cell[41988, 1383, 27, 0, 34, "Section"],
Cell[42018, 1385, 158, 4, 65, "Input"]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
