(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    314328,       7703]
NotebookOptionsPosition[    261275,       6393]
NotebookOutlinePosition[    298798,       7217]
CellTagsIndexPosition[    298711,       7212]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[TextData[StyleBox["MathCode C++",
 FontWeight->"Bold",
 FontSlant->"Italic"]], "Title"],

Cell["- an overview and introduction", "Subtitle"],

Cell["\<\
Copyright \[Copyright] MathCore Engineering AB 1999-2007\
\>", "Copyright"],

Cell[CellGroupData[{

Cell["Abstract", "Section"],

Cell[TextData[{
 "This notebook gives a basic introduction to the third-party application \
package ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " that generates optimized C++ code from ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " programs. We demonstrate the basic commands for generating, compiling, and \
installing C++ code into ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ". Furthermore, the subset of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " commands that can be translated to C++ will be described. The objective \
with this notebook is to give a brief introduction to ",
 StyleBox["MathCode C+",
  FontSlant->"Italic"],
 "+ and to describe what can be done and what cannot be done with ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 "."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "What is ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 "?"
}], "Section"],

Cell[TextData[{
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " is an application package for ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " that generates optimized C++ code for ",
 StyleBox["numerical computations",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 StyleBox[".",
  FontSlant->"Italic"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Compiled code can easily be called from within ",
 StyleBox["Mathematica.",
  FontSlant->"Italic"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Optionally, stand-alone code can be generated (i.e. code that can be \
executed independent of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ")."
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell["\<\
Increase in performance up to 1000 times for some computations.\
\>", "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "External libraries of C, C++, and Fortran77 code can easily be connected \
and called from within ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell["Extended functionality for extraction of parts of matrices.", "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " turns ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " into a powerful environment for prototyping of advanced numerical \
algorithms for production code development."
}], "Text",
 CellDingbat->None,
 CellMargins->{{54, Inherited}, {Inherited, Inherited}},
 TextAlignment->Left,
 TextJustification->0,
 Background->GrayLevel[0.666667]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Who can benefit from ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 "?"
}], "Section"],

Cell[TextData[{
 "Users that are using ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " as an environment for ",
 StyleBox["developing",
  FontSlant->"Italic"],
 " applications and then want to automatically generate efficient C++ code \
for usage in large software systems: generation of",
 StyleBox[" embedded code",
  FontWeight->"Bold"],
 ". "
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Users that would like to use ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " as an ",
 StyleBox["interface",
  FontWeight->"Bold"],
 " to old function libraries or single programs written in Fortran, C, or C++."
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Users that would like to write ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " packages for numerical computations and then compile them to executables \
for ",
 StyleBox["proprietary reasons",
  FontWeight->"Bold"],
 " and/or for ",
 StyleBox["easy distribution",
  FontWeight->"Bold"],
 ".  "
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Users that would like to take advantage of the excellent ",
 StyleBox["prototyping environment",
  FontWeight->"Bold"],
 " in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " but also have easy access to special purpose routines already written in \
C, C++, or Fortran."
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Users that have ",
 StyleBox["performance",
  FontWeight->"Bold"],
 " ",
 StyleBox["problems",
  FontWeight->"Bold"],
 " e.g. due to very large data files or functions that are computationally \
expensive to compute."
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
}, Closed]],

Cell[CellGroupData[{

Cell["Basic Usage", "Section"],

Cell["Load the package", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<MathCode`\>\"", "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"MathCode C++ 1.4.4\"\>", "\[InvisibleSpace]", "\<\" for \"\>", 
   "\[InvisibleSpace]", "\<\"vc100\"\>", 
   "\[InvisibleSpace]", "\<\" loaded from \"\>", 
   "\[InvisibleSpace]", "\<\"C:\\\\b\\\\omath\\\\MathCode\"\>"}],
  SequenceForm[
  "MathCode C++ 1.4.4", " for ", "vc100", " loaded from ", 
   "C:\\b\\omath\\MathCode"],
  Editable->False]], "Print",
 CellChangeTimes->{3.508662307665896*^9}]
}, Open  ]],

Cell[TextData[{
 "Since a number of temporary files will be generated by ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " we create a temporary directory and set this to the current working \
directory."
}], "Text"],

Cell["SetDirectory[$MCRoot <> \"/Demos/Overview\"];", "Input",
 CellChangeTimes->{{3.508662435758006*^9, 3.5086624366173706`*^9}}],

Cell["\<\
Define a simple function but add type information about argument and return \
value (blue color)\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f", "[", 
    RowBox[{
     StyleBox["Real",
      FontColor->RGBColor[0, 0, 1]], " ", "x_"}], "]"}], 
   StyleBox["\[Rule]",
    FontColor->RGBColor[0, 0, 1]], 
   StyleBox["Real",
    FontColor->RGBColor[0, 0, 1]]}], ":=", 
  FractionBox[
   RowBox[{"Sin", "[", "x", "]"}], 
   RowBox[{"1", "+", 
    SuperscriptBox["x", "2"]}]]}]], "Input"],

Cell["The function can be used a usual", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"f", "[", "x", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwtmXk8VP/3x1VkCCFbKqRCqKRNkjNlyVakQipbhSRKUZlUSnYVSUIL2Su0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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  PlotRange->
   NCache[{{-Pi, Pi}, {-0.43741414645553045`, 
     0.4374140671591159}}, {{-3.141592653589793, 
    3.141592653589793}, {-0.43741414645553045`, 0.4374140671591159}}],
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.5086623077752695`*^9}]
}, Open  ]],

Cell["Generate C++ code and compile it to an executable file", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BuildCode", "[", "]"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 1 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5086623101971135`*^9}]
}, Open  ]],

Cell[TextData[{
 "The generated executable is connected to ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086623138845663`*^9}]
}, Open  ]],

Cell["Now, the executable file is used in computations instead", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"f", "[", "x", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "\[Pi]"}], ",", "\[Pi]"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwtmXk8VP/3x1VkCCFbKqSSopISSc6UJVuRCklEhSRK0TKplOwt5IPQQvaE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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  PlotRange->
   NCache[{{-Pi, Pi}, {-0.43741414645553045`, 
     0.4374140671591159}}, {{-3.141592653589793, 
    3.141592653589793}, {-0.43741414645553045`, 0.4374140671591159}}],
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.508662313931441*^9}]
}, Open  ]],

Cell["We take a look at the generated C++ file", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<Global.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"Global.h\"

#include \"Global.icc\"

#include <math.h>
double Global_Tf ( const double &x)
{
    return sin(x)/(1+(x*x));
}

void Global_TGlobalInit ()
{
; 
}
\
\>", "Print",
 CellChangeTimes->{3.5086623139783154`*^9}]
}, Open  ]],

Cell[TextData[{
 "Uninstall the executable, remove the temporary files, and remove the \
function ",
 StyleBox["f",
  FontFamily->"Courier",
  FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"CleanMathCodeFiles", "[", 
   RowBox[{
    RowBox[{"Confirm", "\[Rule]", "False"}], ",", 
    RowBox[{"CleanAllBut", "\[Rule]", 
     RowBox[{"{", "}"}]}]}], "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"Remove", "[", "f", "]"}], ";"}]}], "Input"],

Cell["\<\
Hence, at the very basic level the MathCode user only have to\
\>", "Text",
 CellDingbat->None],

Cell["Add type information (function argument and return value)", "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Build the C++ code using ",
 StyleBox["BuildCode[]",
  FontFamily->"Courier",
  FontWeight->"Bold"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Install the executable file using ",
 StyleBox["InstallCode[]",
  FontFamily->"Courier",
  FontWeight->"Bold"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Type Declarations", "Section"],

Cell[TextData[{
 "To be able to generate efficient C++ code the types of function arguments \
and return values has to be specified. The basic types used by ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " are"
}], "Text"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"Real", ",", "Integer", ",", "Null"}], "}"}]], "Input",
 Evaluatable->False],

Cell["but also vector- and matrix types of different formats", "Text"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Real", "[", "5", "]"}], ",", 
   RowBox[{"Real", "[", 
    RowBox[{"3", ",", "4"}], "]"}], ",", " ", 
   RowBox[{"Real", "[", "_", "]"}], ",", 
   RowBox[{"Integer", "[", 
    RowBox[{"m", ",", "n"}], "]"}], ",", " ", 
   RowBox[{"Integer", "[", 
    RowBox[{"2", ",", "3", ",", "n_"}], "]"}]}], "}"}]], "Input",
 Evaluatable->False],

Cell["\<\
Type declarations can be given in essentially two different ways:\
\>", "Text"],

Cell["Directly in the function definition:", "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f", "[", 
    RowBox[{
     StyleBox["Real",
      FontColor->RGBColor[0, 0, 1]], " ", "x_"}], "]"}], 
   StyleBox["\[Rule]",
    FontColor->RGBColor[0, 0, 1]], 
   StyleBox["Real",
    FontColor->RGBColor[0, 0, 1]]}], ":=", 
  SuperscriptBox["x", "2"]}]], "Input",
 Evaluatable->False],

Cell["In a separate command:", "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"g", "[", "x_", "]"}], ":=", 
  RowBox[{"Sin", " ", "[", "x", "]"}]}]], "Input",
 Evaluatable->False],

Cell[BoxData[
 RowBox[{
  StyleBox["Declare",
   FontColor->RGBColor[0, 0, 1]], "[", 
  RowBox[{
   RowBox[{"g", "[", 
    RowBox[{
     StyleBox["Real",
      FontColor->RGBColor[0, 0, 1]], 
     StyleBox[" ",
      FontColor->RGBColor[0, 0, 1]], "x_"}], "]"}], 
   StyleBox["\[Rule]",
    FontColor->RGBColor[0, 0, 1]], 
   StyleBox["Real",
    FontColor->RGBColor[0, 0, 1]]}], "]"}]], "Input",
 Evaluatable->False],

Cell[TextData[{
 "The latter construction can be useful if one wants to ",
 StyleBox["separate",
  FontSlant->"Italic"],
 " already existing ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " code with the type information needed to be able to generate C++ code \
using ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 "."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["The Compilable Subset", "Section"],

Cell[TextData[{
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " generates efficient C++ code for a ",
 StyleBox["subset",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 " of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "."
}], "Text",
 TextAlignment->Center,
 TextJustification->0,
 Background->GrayLevel[0.666667]],

Cell[TextData[{
 "A necessary condition for using ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " to generate code for a ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " function is that the input and output are numerical entities such as \
vectors or matrices of integers or real numbers. "
}], "Text"],

Cell[TextData[{
 "Code for intermediate symbolic computations can be generated if ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " can evaluate these computations at compile time.  "
}], "Text"],

Cell[TextData[{
 "The subset of the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " language that MathCode C++ can handle is denotet the \"compilable \
subset\". The following list gives a brief overview and characterization of \
this subset"
}], "Text"],

Cell["\<\
Statically typed functions where the types of function arguments and return \
values are given by the above types.\
\>", "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 StyleBox["Scoping constructs:",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"],
 StyleBox["\n\tBeginPackage[], EndPackage[],\n\tModule[], Block[], With[]",
  FontFamily->"Courier",
  FontWeight->"Bold"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}},
 FontFamily->"Courier",
 FontWeight->"Bold"],

Cell[TextData[{
 StyleBox["Procedural constructs:",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"],
 "\n\tFor[], While[], If[], Which[], Do[]"
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}},
 FontFamily->"Courier",
 FontWeight->"Bold"],

Cell[TextData[{
 StyleBox["Lists and tables:",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"],
 "\n\tList[], Table[], Array[], Range[], IdentityMatrix[]"
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}},
 FontFamily->"Courier",
 FontWeight->"Bold"],

Cell[TextData[{
 StyleBox["Size functions:",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"],
 "\n\tDimensions[], Length[]"
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}},
 FontFamily->"Courier",
 FontWeight->"Bold"],

Cell[TextData[{
 "Arithmetic and logic expressions e.g.:\t\n\t",
 StyleBox["+, -, *, /, ==, !=, >, !, &&, ||, ...",
  FontFamily->"Courier",
  FontWeight->"Bold"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 StyleBox["Elementary",
  FontFamily->"Times New Roman"],
 " functions e.g.:",
 StyleBox["\n\tSin[], Exp[], ArcSin[], Sqrt[], ...\n",
  FontFamily->"Courier",
  FontWeight->"Bold"],
 StyleBox["and others e.g.:",
  FontFamily->"Times New Roman"],
 StyleBox["\n\tRound[], Max[], Cross[], Transpose[], Dot[], ... ",
  FontFamily->"Courier",
  FontWeight->"Bold"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 StyleBox["Constants:",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"],
 StyleBox["\n\t",
  FontFamily->"Times New Roman"],
 "True, False, E, Pi"
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}},
 FontFamily->"Courier",
 FontWeight->"Bold"],

Cell[TextData[{
 StyleBox["Assignments:",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"],
 "\n:=, ="
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}},
 FontFamily->"Courier",
 FontWeight->"Bold"],

Cell[TextData[{
 StyleBox["Functional commands:",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"],
 "\n\tMap[], Apply[]"
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}},
 FontFamily->"Courier",
 FontWeight->"Bold"],

Cell[TextData[{
 StyleBox["Some special commands:",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"],
 "\n\tSum[], Product[]"
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}},
 FontFamily->"Courier",
 FontWeight->"Bold"],

Cell[TextData[{
 "A complete listing of the compilable subset can be found in the ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " manual or at our homepage ",
 StyleBox[ButtonBox["www.mathcore.com",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    URL["http://www.mathcore.com"], None}],
  FontVariations->{"Underline"->True},
  FontColor->RGBColor[0, 0, 1],
  Background->None],
 "."
}], "Text"],

Cell[TextData[{
 "Functions not in the compilable subset can be used in external code by \
call-backs to ",
 StyleBox["Mathematica,",
  FontSlant->"Italic"],
 " see below for an example."
}], "Text"],

Cell["\<\
Examples of functions that are not a part of the compilable subset:\
\>", "Text"],

Cell["\<\
\tIntegrate[], Solve[], FindRoot[], LinearSolve[], Expand[], Factor[]\
\>", "Text",
 FontFamily->"Courier",
 FontWeight->"Bold"],

Cell[TextData[{
 "These function can be used if ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " can evaluate them at compile time to expressions that belong to the \
\"compilable subset\"."
}], "Text"],

Cell[CellGroupData[{

Cell["Enlargement of the compilable subset", "Subsection"],

Cell[CellGroupData[{

Cell["Symbolic expansion of function bodies", "Subsubsection"],

Cell[TextData[{
 "Functions not entirely written using ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " code in the compilable subset, but whose definitions can be evaluated \
symbolically to expressions that belongs to the compilable subset, can be \
handled by ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 "."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f", "[", 
    RowBox[{
     RowBox[{"Real", " ", "a_"}], ",", " ", 
     RowBox[{"Real", " ", "b_"}]}], "]"}], "\[Rule]", "Real"}], ":=", 
  RowBox[{"Integrate", "[", 
   RowBox[{
    RowBox[{"x", " ", 
     RowBox[{"Sin", "[", "x", "]"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", "a", ",", "b"}], "}"}]}], "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"f", "[", 
  RowBox[{"1.", ",", "2."}], "]"}]], "Input"],

Cell[BoxData["1.44042242098021`"], "Output",
 CellChangeTimes->{3.508662315384547*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "f"}]], "Input"],

Cell["Global`f", "Print", "PrintUsage",
 CellChangeTimes->{3.5086623155251703`*^9},
 CellTags->"Info3508665915-4434317"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"f", "[", 
          RowBox[{"a_", ",", "b_"}], "]"}], ":=", 
         RowBox[{
          SubsuperscriptBox["\[Integral]", "a", "b"], 
          RowBox[{
           RowBox[{"x", " ", 
            RowBox[{"Sin", "[", "x", "]"}]}], 
           RowBox[{"\[DifferentialD]", "x"}]}]}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`f],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086623155251703`*^9},
 CellTags->"Info3508665915-4434317"]
}, Open  ]],

Cell["Generate C++ code and compile it to an executable file", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BuildCode", "[", 
  RowBox[{"EvaluateFunctions", "\[Rule]", 
   RowBox[{"{", "f", "}"}]}], "]"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 1 function(s)\"\>"], "Print",
 CellChangeTimes->{3.50866231829076*^9}]
}, Open  ]],

Cell[TextData[{
 "The option EvaluateFunctions tells ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " to let ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " expand the function body as much as possible. Everything works fine since \
the result belongs to the compilable subset"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Integrate", "[", 
  RowBox[{
   RowBox[{"x", " ", 
    RowBox[{"Sin", "[", "x", "]"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "a", ",", "b"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", " ", 
   RowBox[{"Cos", "[", "a", "]"}]}], "-", 
  RowBox[{"b", " ", 
   RowBox[{"Cos", "[", "b", "]"}]}], "-", 
  RowBox[{"Sin", "[", "a", "]"}], "+", 
  RowBox[{"Sin", "[", "b", "]"}]}]], "Output",
 CellChangeTimes->{3.5086623218688393`*^9}]
}, Open  ]],

Cell[TextData[{
 "The generated executable is connected to ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086623219938374`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"f", "[", 
  RowBox[{"1.", ",", "2."}], "]"}]], "Input"],

Cell[BoxData["1.4404224209802097`"], "Output",
 CellChangeTimes->{3.5086623220094624`*^9}]
}, Open  ]],

Cell[TextData[{
 "Uninstall the executable, remove the temporary files, and remove the \
function ",
 StyleBox["f",
  FontFamily->"Courier",
  FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"CleanMathCodeFiles", "[", 
   RowBox[{
    RowBox[{"Confirm", "\[Rule]", "False"}], ",", 
    RowBox[{"CleanAllBut", "\[Rule]", 
     RowBox[{"{", "}"}]}]}], "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"Remove", "[", "f", "]"}], ";"}]}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Call-backs to ",
 StyleBox["Mathematica",
  FontSlant->"Italic"]
}], "Subsubsection"],

Cell["\<\
Consider the following function whose definition includes the Zeta function. \
The Zeta function does not belong to the compilable subset.\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"f", "[", "x_", "]"}], ":=", 
  RowBox[{
   FractionBox[
    RowBox[{
     RowBox[{"Sin", "[", "x", "]"}], " ", 
     RowBox[{"Cos", "[", "x", "]"}]}], 
    RowBox[{"1", "+", 
     SuperscriptBox[
      RowBox[{"Tan", "[", "x", "]"}], "2"]}]], 
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{"-", 
     SuperscriptBox["x", "2"]}]], 
   RowBox[{"Zeta", "[", "x", "]"}]}]}]], "Input"],

Cell["A plot of the function", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"f", "[", "x", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "2", ",", "4"}], "}"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}]}], "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwVlnk8VN8bx+8os1xmRrJVliKR7CWFjqeF9oVIlphBiJJKlC1LtlCJSpbs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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{2., 0},
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.508662323118823*^9}]
}, Open  ]],

Cell["\<\
To generate C++ code we have to declare the function arguments and return \
values.\
\>", "Text"],

Cell[BoxData[
 RowBox[{"Declare", "[", 
  RowBox[{
   RowBox[{"f", "[", 
    RowBox[{"Real", " ", "x_"}], "]"}], "\[Rule]", "Real"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"Declare", "[", 
  RowBox[{
   RowBox[{"Zeta", "[", 
    RowBox[{"Real", " ", "x_"}], "]"}], "\[Rule]", "Real"}], "]"}]], "Input"],

Cell[TextData[{
 "These declare statements does not change the way ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " computes the function"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"f", "[", "2.5", "]"}], ",", 
   RowBox[{"f", "[", 
    FractionBox["5", "2"], "]"}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "0.000796931712141105`"}], ",", 
   FractionBox[
    RowBox[{
     RowBox[{"Cos", "[", 
      FractionBox["5", "2"], "]"}], " ", 
     RowBox[{"Sin", "[", 
      FractionBox["5", "2"], "]"}], " ", 
     RowBox[{"Zeta", "[", 
      FractionBox["5", "2"], "]"}]}], 
    RowBox[{
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"25", "/", "4"}]], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       SuperscriptBox[
        RowBox[{"Tan", "[", 
         FractionBox["5", "2"], "]"}], "2"]}], ")"}]}]]}], "}"}]], "Output",
 CellChangeTimes->{3.508662323150073*^9}]
}, Open  ]],

Cell[TextData[{
 "Generate C++ code and compile it to an executable file. The option \
CallBackFunctions tells ",
 StyleBox["MathCode C++ ",
  FontSlant->"Italic"],
 "which functions that has to be evaluated by ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BuildCode", "[", 
  RowBox[{"CallBackFunctions", "\[Rule]", 
   RowBox[{"{", "Zeta", "}"}]}], "]"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 2 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5086623256500406`*^9}]
}, Open  ]],

Cell[TextData[{
 "The generated executable is connected to ",
 StyleBox["Mathematica",
  FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.508662329212495*^9}]
}, Open  ]],

Cell["Now the external code is used to compute the function", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"f", "[", "2.5", "]"}], ",", 
   RowBox[{"f", "[", 
    FractionBox["5", "2"], "]"}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "0.0007969317121411052`"}], ",", 
   RowBox[{"f", "[", 
    FractionBox["5", "2"], "]"}]}], "}"}]], "Output",
 CellChangeTimes->{3.50866232922812*^9}]
}, Open  ]],

Cell[TextData[{
 "In this case the external code calls ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " when the Zeta function has to be evaluated. After evaluation the \
computations proceeds in the external code."
}], "Text"],

Cell["A plot using the external code", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"f", "[", "x", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "2", ",", "4"}], "}"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}]}], "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwVlnk8FW8bxueIs4yzSKGFlCUShUShp7uFFqmIZIlzkERJJcoWiayVpZIl
u1C27NuZZHcoKUulRSiSFOpXlN55/znz+X6eOc/cz3Vd9z2zyv606TEhDMN0
yJ//X7XcvUXICzR1RXHcVblEJuW/5G0UDDxtm+uM1bnE4thz2kFCGMwl2QUp
aHGJ6bKTx6kiGBxI9zCp0+MSD+ds25g4BmGf5IYT93IJ9as7opYuxmAm+I0Z
8ziXSF3ySNFaAgNRsw4lHxcuIZanz0+SxKBCKI0+dIpLTHZsnJRdisHfSzlT
yR5cokBc2WT1CgzyjMO5jwK5xNq7TIkNyhjU0qr8KxK4RML6sHyPNRg4bUuU
qkjmEni9iGG5CgY9NW61halc4vPQvOdmNQwmU11UIrO5RM6a7/1bNcn6zDqS
vhdzidVlPUn79DHYW/5r2ekWLnHL0HTjtS0YvIPcG//auAS1/0nnU4TBiwN+
UmEdXGJktmXedBsGgT6dVlefkfpANdfSEIPWUq029ddcQq7jroLTQQxE2pcP
60xwCWkDufbHJhisqgk5fHWSS0jxs07LHsIgY/uS4iffuQSz6EF1vzkGctmZ
8gY/ucR/sdUHjawxuD5JkX82zyWmmOjnPRsMdKeE5t5hPOJLcH2isC0GDjr/
NX0U4hEfvFo/1nExCHn0WHeAyiM6rHt91Z0weBTyV+DE5hHNLyzloo5j4JLu
cXyjGI+oN37TMuaMQY6YQOjPQh5RvnVYPMMVg26hNXBGgkekyU/lSJzBwEj2
B/3Pch6RlHR+/9mzGEStvNgbLsMjbkn8nn5yDoM7li8eLpTlEZE0bGuoJwZ+
WfKpNDke4TnO6pn1wYDzgns1TolHuDtGe1v4YTBkmS4yrMwjXN8sXlnqj4Hb
qvsZa1V4BPfpctdTgRiYDyzfnaHKI4xK1lDeh2CwwcjunJYGjzBUzc/WD8Xg
+Wi4uoEmj4As9X13wjBQXWe6cv8GHqF9W+e2aSQGEbzctN0becRKHwO1pmgM
THmNZ99v4hHLZ1q7V8VisEB2WLRkM4+QPLXvgn8cBu/ddn7x0+URonaHGnRu
k/o/vxv1R49H+t3rHBePwXYVSbMifR6BmVixv9/BQDa49aTNFh7xczvPMi8J
g9hW+dRkxCO+1QzPU+9i0L434JHmVh4xruWc6ZCCwepv/ugRyR/zx/c8SsVA
TwOXMQQe8X716UnpdAzaZBfZNZL8OmUq7mIGBqM20TS9bTyid4mnbm8mBmxO
0NI8kgW4X/D1exh8FrwMPrOd/H+2qVhPDgbzMs8NG0me2a6cuCwPg7j4ex5i
O3gE491fRe59DBgB7xceJnmFz/OirAcYBAyKrrpJsqZUrt54PgaDEbWZnSTv
KvFvVi8k800PT/lHss0BMxPPIgzKbootXruT9G98zUBNMZlX6e/CB0kOvvrP
iVKCwXG1Zx5uJCfI93w3LMXg2k8rpxCSC4g838gy0k+0ZeA2yQ3WAbTucgzu
Wa99mU5y/3/mMVKVGGxZ/8buHslfYtfKHK3CwF+acy6LZIo6JSe9msyb0GWJ
ZJIlO3o1R2vI522c23ON5LXOD+rU6jBo8dy56CLJIHx59zk+6dfkWo+jJJul
WjyvJDAwU7pyWo/kE/pqtvOPMDiMKdAXkezfLzS24zEGt/I/wjB53hiP/nNh
DRicyb8vXUjyPbGC+SeN5Dx4fPjuOZJrHwSFLW7GIGv8aasGyc92Wy62asFg
s92ClDFS/4/D61JSWjHALw/KJ5I8GyCsMtJGzqNBWytDkjkyr0pVBBiIq3oY
jJN+KlQVbnXvwMBq7YrRMJI3mQe3l3VioFJ71ECOZPso9ffQhcFDVKYHZD68
1lBdQ56RerKHXjWReYpsev1D0E3OH5luMCC57G+IqEUPBvd3GJ3SIPPYfsfm
VlIvBjqPcd27ZF7fbtRc9aEPg6fbfr4UJpnu9kb71CvS/7MfUwgy7zJ4Sf3D
1xjY6W7tWESyZnbovl8DpJ6rTvbzyP6wfrvBPugdOY8eLEoeJ/upYH941J1h
0h+lVZrlZL81fLZb8m4Eg8ec0IqXWqS/IRszFD5hYJPQqP0f2Z8U4n1l4RgG
GlsXqK8k+9dsvc5w01cMQu03u6qs4xHOAuZp0W8YjK/tk16hxiP8jn/4ffA7
Buc11IZEyXmQnRLFGZjGoGR1SU3PGh7xmzOsO/ULg0X+ZX/kFHnE/pope41Z
8rxDMlMT8jwi3YkS4T6HwdfIfJMSch7trVvx+utfUp+tYTZqK8l55mLpMy5E
gdbPdfNvl5H5aXpSM8ykwBtLHxc9Do+Ic38zJM+mgGu52a0CFo8YW/5F1IFD
gf26D+ylmWRezjJsBhdS4NL95s+f6DxiSHbnnzeSFEhZ1rZsKzmPQ7yr9PtW
UuBoyDPx79Nc4pViq6OkHAUKm4gJzhSXWP+sN9JcngLje4XS13zjEv1KMwPP
FSlQbEDPMv7CJdb0rPPrUqHAwaUqPkeGuUTH+sy6Ni0KsBJm7/o+5xKLRq6h
2t0U0A6vHlxSQL4fvDLPSe6lgH3mtkqH+1yiG6/OcTeiwOmjjszcHC5xQ31E
XHE/BVRvK22Tz+ASbF+9T1GHyP1FCKfe21yCvmj0up0tBe54LfmtF8Al5mHb
oJAHBW4kHd+jZMwl3jw/LHn0PAX8j7r5CPZwiVqnk0YVnhSIb0sUuBhyCZ/I
22WuFylw8U7fq4StXOJX30TYC38KhE5Sd9ZrkN8Hbgma2WGkfg3f69ZKcInR
pOmgPSkUKMNmq7WL7YgPptZ3I1Mp8PG40IrBXDviNb2h8mkaBZgbPFZcSbcj
nnjETJhnUsA8du/S0lg7omSfhoV9Lqn35a4PteftCP8/bio+JRQQity6+uIm
O0LC5nPXgxYK9Pme9OirtCW2Lx+UEftGgfNB94Yp944SFr6hv+/sEgLB0gjH
roXWhGX+OfmIZCEgiuReG+w/QrzP/JRWNiEEmk/lTGIzzQm5SxL2GTsXAOUi
xX+4yZQYk+3wbL+5AJKSjJZv0zhIJN1ofnlkZAHIFYpJOqcaEfiwt3+enjB0
FxxscRs2JP7+rfDaGCkMh5d+8e2d2E7cny14GDQkDNb8VaZSBxBh2DKmt3mz
CEjoe+zJmtYhYnq4/13QFYH8fq+CgTc6xLuhvoeVeiJQmbyCKteqQ1zAmtds
QiLQHnc6oDdJh8jbnC6ls0MEWrHbBh8NdQj2A6tpLWMRSN11Vv5usjbRe0OQ
p84TAedBhZTD5huJY1b5S5XDRKB2CaVbYkKT0D+49sPlcBFYmvJVSvalJrHY
MDfvTYQIlGYWZ+o0aRKPNbL0Yq+JgOUm8L2VrEmsZCQfnY8VgRvDS4PxA5rE
QEVk2ou7IrDX22C2uVSDMJU4uSawVAS8OjPtnCPUCehS2fzqnQgckog7P9Gi
SmiOzX6UGRSBM4P33M7kqhLyQoI43gcRWJ3w3ZIVoUpQtVy/jQ6LwG2K1UDk
flWi4/b9e/+NicCWK3TH3N61xBE7VUmJGRHwlWpmHxhXIdy/qk0foFPhSV9x
ArZ6DZHC1MhvUqeC9IEvltFvFQlpFt1ZQZMK+9SHNHMbFIl41ju5oA1UGPgw
PvoyR5G4wY6KB20qJCvceBh8TpG4LDYWVKNHBepzrrIIrkg4LU6zKjakwiqh
kblbegqEqvRCerINFS4zvmNpRXJEjvTo47mjVNBOW+BXGS9HKMoQflZ2VCiT
P0j7FCBHrFhxalrKngq0qCxmmIkcIbay7W3McSpcfVu18930KmJaPrAs9CwV
1Bg9RIv+KqJ67Xd7j1AqaEzPndjFlyU89Sg1HWFUSBBfGtUcJ0toGi1cpBhB
BbM8hxw3V1kiz0WjoSeKCvLcTqvNS2WJxNwz8jqxVIj5r6G8zHMF4as0NfQ7
mQrPL929qblZhkDy046XSqggeglbEvB6GTGruaCur5QKEUscP4yVLyPKty+S
UC+ngust9q6LMcuIdfYbmt5XUuH+Lr4Hfe8yQjbtnOJ2PhU6V52+OFK9lMBk
Z0aE26ggnBg4ezljCVG/7IdT+FtS72nd8tjrkkS8sUWm4B0VPjw5VZFxRpJw
C6gaZA5SQcFtKv7ZIUli2Uc/m+tDVHizUT43ZIkkcbaYaho3SgUZxfrSQxkS
hMLuJVvuTpHr6l/gOH8xcfW83qKHVBpc7JbMHWIsImxzkg9O02jQGpQUeXVS
nNB6/S9Ki0EDoRLvOsMeceI9NNIrRGkwMRRRLpkmTmxiGs/XiNHAqTm8e3az
OPE53XasaRkNbpwoVX3svpAw7gokXq6jwSZZql3wEIco+PyZoa1Og70pbsSZ
Yg4hJmJmFqNBg4oVl/3iL3GI7s2rx/Zq0WDk3pL9cdIc4nBGu3jtZhpoCd0d
2WrJJmw9Fznd3UmDQEfdI/pvmIS7dBbTwYoGp25uo3mzceKZNtuCsKZBllTO
RbMhBqFp4pW2/CgNSrTX7HeoZBAzwXu0X9jRwPRgt4OMA4O4ODlhZ3CMBjVT
SzfMV9OJy4+1S5TcaaDBs9HZ60Ej4pxbLb8E0yB4+Z8dT0VFCE52UJjGVRo0
JsvcPzAmTIQPoSrPUBqkPuH2/moWJgJty5ZQImgg9yZzvPGyMOFmlt4ncYMG
H1DYRvizgNgFvuZbE2iwIqI2K/OnEDErpW4SU0CDlArRhxvoFOK8+XhAXyEN
Nn9Msgz7jBHfYrKLpItp4KAt5LOgEyM+sWTE7pXQ4OeznmzHGIx4sQB/WlNJ
A5pxJK1vBUbkTw7tG3lMA557dAbfd57Pbbm9e1MfDXqKG2m66+b4yxdHSnj1
02Dl8rDAK39m+b3cgA9lL2ngUxx+giWY5e+bdfbdMEADv+VGSS+dZ/k6arpF
6wZpkDh4Sd/83m8+O3ZASnGcBqdnQy3urf/Fr7Vd9WnhPxrcHloTq3byB9/r
vkTpQYwOQV9DqP3bfvA1fzECr1PocEbF6XOV1A9+TvTUMpYwHR74CanJN87w
Yxsb9tMZdNg/mtocLDvDd1FxKp8Xp0Pr1i8Ljr6f4i/5mRc8vpoOf1eHnh8N
/ca3Vcm2M1Cmg8ldi1WL7b/xM23TNqesoUN364nTF/W+8de33J4wVaXDf2ln
v7x7M8k3jL9iXq1BB9Vvc/dEi7/yz+vaKobp02GLeK/KlpNf+N3+C5tWm9Kh
zjqyvUJzlL+khJkScIgOu+65W4ZPfeLbfqJdfGVGh5wtntL1Dz/xxw7Mq16z
oMOlgBkPS61PfGzVeNwPGzo0lIvOK+t/5K9raHRsPE4+T9SUOXh0mB9BuyBs
70cH/T0iTqUf3/M/2zY1ZPrToadNwqys7D1/b7l40KdLdIjwOpBHCX7Px4/l
U05dpgP7/aatPgrv+eGPP8x7XSX12+nc/azzLT/Mz/h3VDQd3NfFiHVMveZf
nZb7WpVFnt/G84BTVi//4173/D/ZdBi92GRNNe/lG6bXndyaQ4eZnT3qa0R6
+VSTI+ONeXQIsBjsmXHq4YcURI0+LaSDWsJQRKz6C37wiV8fRqro8HLjlZ0n
Xz/jB73r7Fv4hHz+QTNfOSkB322FXObpp3QQDH4o+x3Wzj9y1NO9s4sO61fI
5L3828Zf93oFI+w5Hcr27zQvGG3lv+o9rYu9pEP9emFKcHszX/3JwuSvQ3Qw
q5XW5Nc95i9nOZ3YN0IHIYMb4wHwmE/dV70x7yMdXtlLIJ3Gev7rNvsnx8bo
UJ4wF33xySN+SFPJv4GvdFhaHDhqP1DHf1Nrbi/4TYfditpGDx6W81vnctet
maODX+shC6+eMv5D3fnZkD90eNGk0kWZLeWHVmbHbv9HB6f3fFej3SV8rdL/
GquEGTDsqCfxZqqQH37/jlIOhwFnQk8f77DI4OskvPlyRZEBhie+fq8Y8UH2
oy6HB1YzoKrFfRkoB6FI7V/EBmUGODG/uh4xDUWD3eKxH1QYsMm2pevs9mgU
Ibp7M6gz4LMDzfm2ZiJ65/sweE6XAZ6lKYk4/R7CBVsnTfUZsOopF2Oq5CCt
pZ1H8rYwgBW0OPHCnlwUVvZprSUwoCRDzj094D7a8FW6u9yAAUtf2j9I7SpE
V+2urjh7kAH11x3it/wtQQ/zF4e2mTDgjk1Lm4d3KRqYS/u+8hADQhrwQv+f
pUgjvrbxqTkDHljP3l44XoZed313UbNmQKzrkijH5gq0frtNxegxBpTP2Kug
XTWoT0HjoJ0PA4gSWdMH/EdI+pZYSoovA/5FRCvhY48Ql/Zt4p0fqc8F70dz
4vXo81hBBDeAAYtSv5cnHatHfwpUW3jBDJiJSRj7KPIYrdysjByvM0CjufGt
/IYG5JhHi8q8wYC6HPwT1aIB5S7/9Ho4mgH/qdQfvOrdgDTnsy4ei2NAwrHi
O/aPGpBBg3yZ0x0G3OP8lozd1YhcjGVVT2QwQMr+Gc9qZxMq4M9752YyYELp
X42FfROaXv+2bSyLAactCz3mLjUhX/Hk4y45DJBQubZ2U3UTuta3LMM1nwEt
11NKbqo0oxKe5DK3ClJ/s0A7i7lm9Kv7h3NBJQNqNEdMZxe1IP2dPRVfqxhw
Nlyfvk+1BTWvjjM/XcuAlS0H6/5at6D+8YXR7vUMyPwuYpJT0YL+nGfRzwkY
sEvJxiCU24o6nkw7POlgwCW1IO25860oUekVseYJA3bYxb6UimhFui+zvd53
kX4bWSQcLG1FF/Thk3EvuT8M3+0SakO7b67ekdvHgNmTEnVpEm1I6iszRfgl
A4wy0nasUm5D5SkvD9e8ZkDj37+9avva0IzQuWblQQaoHnHFo260oUZrS7kr
HxigIr3VcTC1DcWWbvV/N8QA+D2XP1rUhjSdmNq3PjJAe+jQLvGuNuTelpW1
4AsDOj50Db/H2xHIRVLsJhhwc/0OyYEl7Yjjc/Zo9VcGNLVHR19b3Y4KVLdK
nPnOAHdV7+Dl29rRl+v9V97+ZIDoyNelw2fbUe0o//3mXwx4lBt/OMe/HUVs
y9K/+ZsBJr9rcOXwdqQyfWbG6A8DdopYDVmntiPnw6KOVRQc0PWp8xlt7WhT
4Xdi8QIc5oXkYk2etyMavX+5uzAOWZ9MVYiBdpRdlfl8NQ2HYnuvmldf29Gw
NNoRx8SBmpT9Y6WYAJWeV0iZZOFgRDX36JcSoKAn+NxeDg7B+SUGh2QFSC6g
7yFFHAelk0WP4tQEaKq/jn10EQ7ch8+O2GgJUL1GpkvlYhwSWvJWjukKEHfI
Xe60FA6PLLIkzHYJkLq+hX/7EhwOye1R3WAsQP/itrxSXIbD38g8g5emAnTX
EI8ZkMahrPPdPt+jAuSW8m1CZwUOdzR2y3rbC9CWX717YmVxuB8QXwfHBWgg
N4OyVw6HDcInVqq4C1Cz4sD2a/I4lDyNnaZ7CFBh2uIr3Qo4VHbeXlvlJUCX
7wRTrZVwGP1I0bjuL0AnJfi7UpRxcNn4sKgsUIDMo3+GDq3Bgf+t63fBFQFS
DjsuelIVh+9y/3YtDxcgcZHUfUVqOIQ6Ns5dixSguYD+qJl1OJzzduztvSZA
w3/Enm5Sx+Gmpuy66RsC9OTCHjE/DRwKXzUcGY4RoMqZQJN6TRxO5hTfz4sT
oHT36hgRLRwer+r3MbwlQJFfpp7v2YhD0ZUKmcrbAnTeea3ENW0cxNfzRv7F
C5DtsMPhbh0caNQLq1ckCNBubtJtyc04bHqjrrUoUYA0B170W+niMPLGlPeO
ZOkjrGUpejgMT+hgQUkCRH1hYD2kj4OwZKf9PMmTB/yTlBAOz0djXpkkC1C/
oPyN61YcLo4mV/qR/HjX5IoiwEFy4pbxZZIfNChxZ7bhsDu8ZsyO5JtbuWmb
dpD+ebwdkCTZvyb+g+9OHK6NHve8R+7vrPNMvt4ABw3mYRFxkk1LGMdEduHQ
abGl/zBZn9767dl7duOQ+mv3Fk/yPAr3vT9F7cHh60a9y+53BIitVKLcvReH
jT2nZgzI8/9KHz8huQ8Hnn/3+5+kXoMrFO5bGZN+yE7HBt0UoPYEmy939+NQ
cHDK8kssmWfJm2pDB3CQq4jw1yD1vxvT6aZkgsOk82LjQ6Q/oWxqkaspDukH
fMRMSf+sqV6aM2Y4+L9IT/kUJkA7Lxee23QYh3/vDP75XxWgdfOfSn0tcDhO
RPrMkPkQ+nlER8QKh6vcixPelwToy5noC3uscWgOe3cu0leAeibaqqJscGhZ
4rrJ76IA5Yzo6kva4WAYonnn71kBiuGd87Pi4jA7taQh6rQA+by5z7/Lw6E3
3dtw3lWA9vdIb1NyxOHwixfugY4CNNM4b7DpBA6aKp3vpcwE6C3oXPV1IfUP
XOX08IAAtdSebn3kikOkiPeJ9UYClFD6fu8eNxwk5Nx+PNsmQNsyHx+wOofD
pRlXm0dk/7qs11oX74HD0yLf+RvKAhRXncXsPY8Ddui8/i55AfrYRb4uL+Dw
y5Tid2wJmc+/+3fs8SPnwXmXuNMUAeo9/Fp701Ucsv94qU48aUf/BvdJeIbi
8I3+Z/mG1nakfIo/XRJG5jvzloF9fTvyCUorWh+JQ4xQtLxfSTtaWXRcRSka
h/qeaez+rXbkwpiRkUzEoavq3g8Jy3YUG3fsz6EkHFSw1Fq6CTkfZfteRSfj
wChPZ47uJuerdvVtVirZLwprFp7a1I5K7QMWimThQDncKnNLsh3N1zBFZgrI
/MevGbr2tA3FuCl+efYYh5TG2iDexjZkOJjcp9KIQ+zPai9HlTY0e0iyIagJ
B9/7jvstZNuQw2ZawsbW/9cXcoHFaENawmO77nSS8+a1hH/961bUc+dBBref
1Ff4yIrz/q1oSdMG68kJcj1lpMCuugXdXb69nblUFB6oDyUOSjSj9dc//Qk4
IQoa/sqFC3sfI58RwzOVj0QhNsDhpk0NH+0TWq+7RpwJD96e7elur0T4pJUJ
xZ4JInrwa6KgBO0O4fOtHZhgp0Wv80goQSEycqrljkxQPpK1Dg8pQUL7xqiu
x5lg72UuzTtagn7neNW+OMkE4yNWOgSzBI3axynleDHhXeXjBn23h6i5t/Pv
/igmnNiR2a2rX4yC+JCXVMkE85UmKztoBejVhZ9/3lUxwUu5381uNB9pbnhw
QL6GCeeK1geLtuWjD/ekfuTWMUHq++xEfng+2h79dWvFYyb8M2scX8bJR0KO
ST3POpiwrPPjuuvSD9Blxi+MNsgEF8VJq9w9eSjQrNDiDIMFp7SiBvI/ZyPH
n64VT3EWsEaakdjTbGQYryy1jskCK81iTlRJNmIOpPV+ZrPgqBq9ucc3G8U7
xJk7LmZBmEv+nzSxbFR09uIhC1lyfSdPYemWLPT2xo4DWzayYGhb4Qnp3AxU
r4kVJmqzIMuA1VB0IwNlvqhlz+qwQPnQVPnRCxnIZYn2k3JdFjBFVmhQdmeg
HynKxupA1lfb9+jkaDpiFrKM5I1YMJcY29iqlo50O/sMGTwWhC9b17q6PRXZ
+Hfu4NmzIGBMeMvislR0aX0DVDmwwOFYRLBUaipqjC7QPeHEgirP1eZ2nqlo
/+HgdW2uLADM23BKPhU5vNOUDPNkwXHvElN1agqK/B41Qo9kgc/zRXVTaUmo
ID3oAzeKBVuefS+g+SWhZ4cuvqu8xgKjf2reO48kIclyx5fO0SzQku6S28hJ
Qmneep2tt1ggHxZ+fNI/EZUvGCsNTWPBQl9tmfxjCeidxM4r9AoWtCL1kQsH
4pH05winPZUs+O1rstJTIx4d4T/fHV5Fru8SI2IWxaNuJwcWs5YF9zTVm9f0
30aNFZdvcepZwBHNO9RhfxvlWdbnSApYYK1SuuOJ3y3kkYw6Fd6xoCIp9LJM
exwqOhNSeOw9C16wxCP7C+PQhMGT6OxBFrwWc5orvBmHnL4ePaw8zAKDLzH5
Jbw4ZLnV793aMRacq2vvEp2LRWiw5rvmNAvujLmVftaMRQzFzZLbaGzYsvty
e315NOp6f1Ukmc4GqcAOwWRyNIpP6p35xWDD58N9T/SCo5HyYo/nRUw2/Oq0
dFYzi0aGCwpvrBRnw+aIvU/PTd9AlwcVRDEZNoQbLe700LmB/iSLYYQmG2qC
nrBEc6NQo6Xd5DItNlTXDx7x8Y9CERIFbz03suFF9Y36lWZRaFmUUd26TWw4
//mHJI5FoU2+Id53t7Ch7rwmzcQqEp23+jPjt5sNCnczxS4viUBfJUfH9G3Z
sJcYp9CrQ1HXj1B+oB0bRHctXy8bH4qKX6yJbeayYeVY/mI7z1DkEeOqf9CB
DSOinBNHNoSiOdbkNQdnNuSU9Q2oFl5FdJGfG8LPsqHMUP+8bkEIkp9e4N8X
woYPIld7X7ZeQcLdmabSoWyo3ELxr71/BX0sMlDihbGBEejDabh2BeW5Xe36
HMGGr0zXo3rmV5D6Z1x+/gYbulMbsl4MBZH6L2xVSGRDcve67eG0IHTkqaz4
2UI2VJ3ufsJwDkTedb+DiorYIB0+45x2IBAl3n/+42sxGzLEFhlb6ASit1ev
vnQtZUPE3KyHDC0QOcBk6rEqsp7SrBecnADkVsxfb9nAht1U1uGht5fQldij
xtDHBveK64l5q/1QdqAO4d/PhnnD8xd+/PFFLacXatS9ZMOMgZjd+ee+CN/X
tFh3gA0Vn1wcFwX6ouvCagMbBtmw7dpvycdvfVDC+T8uSuNseLX08kRlhjcq
tEgMZf8j6zcYsIo0vYBs1HxpzhgHTE59WyqqegHRFxwNqadwYHA5fcsjkQuI
W7DiiocwB86UsnM6qrzI78H0S68YHOj5L+yLk4IX8niY45m9mANVc8Vdzgs8
0SZWuSNS4cDz9zzf8Y/n0PCHW0O313LAaOfv2prmc+hGpZf9d1UOeEacTKzO
PofGHDZzM9dzQLU1Z/eO4+dQQnWtNb6RA4H/Tol5fz6L/hxvMO3dygH51bZn
jH6fQfWPu8DNnAOoJOXXb213FFqzIKvoMAcOOAp5z8q4o/2l2oxpCw5knnib
ICPijl5lJT3zsuJA6Yu9yrfunUYzoc72gXYcWJoeV+M74oZWH8CuxJ3ggMOD
SxpNp06hyNfrW6r9OFAs1Gi7pMAFmb6wX/vXnwPGz+YuK15xQUs7b17fGsAB
G38hMTcrF5TNn7NovMyBWvO/W4poLohIaxntvMqB1dbxX745nkBTx+3wwWgO
/MwYT1FQcUZHZq4Z07I50NVtVX3yzTHU/GS7+ul7HMgrkjI0Lj+GtHJ/ivfl
cMBONIh66voxxLG17c++z4G9hYp85+3k/S1qDobFHLCVXGW8674j2pjYcSG4
lgNW648dNwt1QOLb8cwFzznwO0Yj0e4UDwVI80NcX3DA/KmZc4AxD339eebE
8x4OlEDLTI8aD7Xdf7kuo5/Uq2FacsUkFwVI5FZtf8uB6QPlCss8uGhydNfT
wDEOXCL0vf+x7ZBNw1zx2GcOJCGHdTE9tqg9uTDO5AsHRuSlFgYn2aJsUymr
VZMcKL9nqhS21hbZ1n4cfjTDgfELsXJzxkdRx62EFuWfHAgpy3l2Seoo0j2z
P+/Gf+S6eocRd9AGSawuP82d5cCpZem+mp426DLmYto6R+ptUiMnv80GfXsl
s1H9LwfERQrPnmLaINuyZ1Lx8xwo3PdhRKzfGnVcD5799488f6jTT2amNfof
D2r1cQ==
     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{2., 0},
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.508662329306244*^9}]
}, Open  ]],

Cell[TextData[{
 "Uninstall the executable, remove the temporary files, and remove the \
function ",
 StyleBox["f",
  FontFamily->"Courier",
  FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}], "\n", 
 RowBox[{"CleanMathCodeFiles", "[", 
  RowBox[{
   RowBox[{"Confirm", "\[Rule]", "False"}], ",", 
   RowBox[{"CleanAllBut", "\[Rule]", 
    RowBox[{"{", "}"}]}]}], "]"}], "\n", 
 RowBox[{"Remove", "[", "f", "]"}]}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["External functions", "Subsubsection"],

Cell[TextData[{
 "Three very simple external functions, performing ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["x", "2"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[ExponentialE]", "x"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"Sin", "[", "x", "]"}], TraditionalForm]]],
 " respectively. The first two are in one object file and the third is in \
another."
}], "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"extsqr", "[", 
     RowBox[{"Real", " ", "x_"}], "]"}], "\[Rule]", "Real"}], ":=", 
   RowBox[{"ExternalFunction", "[", "x", "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"extexp", "[", 
     RowBox[{"Real", " ", "x_"}], "]"}], "\[Rule]", "Real"}], ":=", 
   RowBox[{"ExternalFunction", "[", "x", "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"extoscillation", "[", 
     RowBox[{"Real", " ", "x_"}], "]"}], "\[Rule]", "Real"}], ":=", 
   RowBox[{"ExternalFunction", "[", "x", "]"}]}], ";"}]}], "Input"],

Cell["\<\
Define a function to create a list of numbers using the external functions.\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Makeplot", "[", 
    RowBox[{"Integer", " ", "n_"}], "]"}], "\[Rule]", 
   RowBox[{"Real", "[", "n", "]"}]}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"Integer", " ", "i"}], ",", 
      RowBox[{
       RowBox[{"Real", "[", "_", "]"}], " ", "arr"}]}], "}"}], ",", 
    RowBox[{
     RowBox[{"arr", "=", 
      RowBox[{"Table", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"extsqr", "[", 
          RowBox[{"extoscillation", "[", 
           RowBox[{"0.1", "i"}], "]"}], "]"}], 
         RowBox[{"extexp", "[", 
          RowBox[{"-", 
           RowBox[{"extsqr", "[", 
            RowBox[{"0.03", "i"}], "]"}]}], "]"}]}], ",", 
        RowBox[{"{", 
         RowBox[{"i", ",", "n"}], "}"}]}], "]"}]}], ";", "arr"}]}], 
   "]"}]}]], "Input"],

Cell["\<\
Compile the package. Since this is a very small example, we don't bother to \
create a special package for the code.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CompilePackage", "[", "]"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 4 function(s)\"\>"], "Print",
 CellChangeTimes->{3.508662332915573*^9}]
}, Open  ]],

Cell["\<\
For Windows 95/98/NT the external function resides in \"extfunc.obj\" and for \
UNIX the file name is \"extfunc.o\". We generate the object-files. If you are \
running UNIX you have to change \"nmake\" to \"make\" below.\
\>", "Text"],

Cell["Create the MathLink binary. ", "Text"],

Cell[BoxData[
 RowBox[{"MakeBinary", "[", 
  RowBox[{"NeedsExternalObjectModule", "\[Rule]", 
   RowBox[{"{", 
    RowBox[{"\"\<external1\>\"", ",", "\"\<external2\>\""}], "}"}]}], 
  "]"}]], "Input"],

Cell[TextData[{
 "Install the MathCode compiled code so it is called using ",
 StyleBox["MathLink.",
  FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086623366655245`*^9}]
}, Open  ]],

Cell["Plot an example.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ListPlot", "[", 
  RowBox[{
   RowBox[{"Makeplot", "[", "100", "]"}], ",", 
   RowBox[{"Joined", "\[Rule]", "True"}]}], "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[
  {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJw1kgkw3Fccx1dUbRNXG21E4tpmFHG0dazbF4t1Za1dbKrtZllRSd1UaLFV
kqi06KSpaQiZMQ2tDBJRV4PaZrQYxZBBmmwkjtAYR0VoO6qzv92ZnZ3Pft7n
vf//vWcWkxwet4fBYEh3v///Kj8rXiU1sqV2aaEX/YFNg/vlvaxaYiZshgRr
rjatxHoQBcudjVh9xPo4N6F2x0g+SGyAtJSnXczeMeLDkMeeZnJPTRKbwlWn
NznwwANiFgz4FZfZ3jPERzDoliYPuvWY2BysbA99y4E5YgtYZ9Utik4vEFth
/3KY+FrhE2JrxFvuDR3TWiS2ReVgc+MIQ8VvIodZ7/ODVDX+LZwamM/f8lHN
9zZuP9Ru0mlQrWcPH7sQwWSL6nkcUDbVdPZSpup5HZEzbrbzAVNB7ITaPZrx
47rTxGzUJz39tfLQBLEzboa0qnNjR4hdMBb9R5anyW/ErripHdO5XfAzsRsE
At5Pr/q2EbuDm3ci1VPcQOyB9Pk/Kx3zLxN7Ipaxoz1ypIDYCxJxY5lwOoAY
EL5uOHPVKVLJMiBl1q9g3uQz8t5wfmZfzKr5hrw3qiQPLVyXVPfBB+WjSYfW
XBvJ+yBDXnLe5VkLeV+85zySYGXWQd4XLZJrX/09dJs8B8NqEf5X3+8lz8FQ
6T71RwrV+/rBjlNRxVfQ/ZL5oWt7vdpJJCfvj2/Fea+1BxPL/MF5ZUX3ulx1
HwMw1Vd97+gozScLQMyNrn69c7QegwvTg++mLKx0k+ciNLy5OutMF/lATI7+
aF87TPstC4R7znaqufEt8kGwlUydWD7bTD4IbB126kao6jyC0cu6yJRYfkc+
GOPrdSeF3TXkQ1BUtLDQ6VtBPgRm4mr3f6+XkQ9Fx13PXGZCEflQWATkbwk3
zpA/hiA/k/korQTyx/C1UZldooGAPA/9aYu68nRr8rts7CbNqFN4Kn0YNhnP
peU8J6VHGGafSNu4S2E0Pgys5NTJ2WKxknvC4PahtMO2Jp7m5+N+Ge+lYqtk
6vn4a+1w4JwijXo+TDSODyTkZlLPh1r90cT+0Czqw8H7kt3YHUjvg3CMOk51
XMkmloVjOi0uemeVxveEw0vh8GnT8EfUC5DVXeqn6UDzQ4C5yTvsDu906gWQ
sD5ueMcwlXoBWvvPn3zUlki9EMt8p4uZHNo/CKGX4Rk4rRZHvRAvXjG8+8LL
EuqFYFdqGcXoRlMfAVai+97Kx0LqI6DTlpewLuRRH4GJLS7Hvt2f+gho6PPm
fW94UR+JKY/6zvb9jtRH7u7vG0ODYivqI7HqlHXvgMSY+kiYbjBmgtZ1qI+C
aUntZkHVlvI8EQWbpe9XL+ybVrIsCmmfF/bp/3NByT1RWP/9ExNNky46fxHE
itykyakH1ItwcLArvoCzQr0ILhohZVGjz6kXofRS0y9f2DNo/eOQ52T725ur
e/0HyuvC6g==
    "]]},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, Automatic},
  PlotRange->Automatic,
  PlotRangeClipping->True]], "Output",
 CellChangeTimes->{3.5086623366811495`*^9}]
}, Open  ]],

Cell["This is what our example function really implemented:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ListPlot", "[", 
  RowBox[{
   RowBox[{"Array", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       SuperscriptBox[
        RowBox[{"Sin", "[", 
         RowBox[{"0.1`", " ", "#1"}], "]"}], "2"], " ", 
       RowBox[{"Exp", "[", 
        RowBox[{"-", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"0.03`", " ", "#1"}], ")"}], "2"]}], "]"}]}], "&"}], ",", 
     "100"}], "]"}], ",", 
   RowBox[{"Joined", "\[Rule]", "True"}]}], "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[
  {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJw1kgkw3Fccx1dUbRNXG21E4tpmFHG0dazbF4t1Za1dbKrtZllRSd1UaLFV
kqi06KSpaQiZMQ2tDBJRV4PaZrQYxZBBmmwkjtAYR0VoO6qzv92ZnZ3Pft7n
vf//vWcWkxwet4fBYEh3v///Kj8rXiU1sqV2aaEX/YFNg/vlvaxaYiZshgRr
rjatxHoQBcudjVh9xPo4N6F2x0g+SGyAtJSnXczeMeLDkMeeZnJPTRKbwlWn
NznwwANiFgz4FZfZ3jPERzDoliYPuvWY2BysbA99y4E5YgtYZ9Utik4vEFth
/3KY+FrhE2JrxFvuDR3TWiS2ReVgc+MIQ8VvIodZ7/ODVDX+LZwamM/f8lHN
9zZuP9Ru0mlQrWcPH7sQwWSL6nkcUDbVdPZSpup5HZEzbrbzAVNB7ITaPZrx
47rTxGzUJz39tfLQBLEzboa0qnNjR4hdMBb9R5anyW/ErripHdO5XfAzsRsE
At5Pr/q2EbuDm3ci1VPcQOyB9Pk/Kx3zLxN7Ipaxoz1ypIDYCxJxY5lwOoAY
EL5uOHPVKVLJMiBl1q9g3uQz8t5wfmZfzKr5hrw3qiQPLVyXVPfBB+WjSYfW
XBvJ+yBDXnLe5VkLeV+85zySYGXWQd4XLZJrX/09dJs8B8NqEf5X3+8lz8FQ
6T71RwrV+/rBjlNRxVfQ/ZL5oWt7vdpJJCfvj2/Fea+1BxPL/MF5ZUX3ulx1
HwMw1Vd97+gozScLQMyNrn69c7QegwvTg++mLKx0k+ciNLy5OutMF/lATI7+
aF87TPstC4R7znaqufEt8kGwlUydWD7bTD4IbB126kao6jyC0cu6yJRYfkc+
GOPrdSeF3TXkQ1BUtLDQ6VtBPgRm4mr3f6+XkQ9Fx13PXGZCEflQWATkbwk3
zpA/hiA/k/korQTyx/C1UZldooGAPA/9aYu68nRr8rts7CbNqFN4Kn0YNhnP
peU8J6VHGGafSNu4S2E0Pgys5NTJ2WKxknvC4PahtMO2Jp7m5+N+Ge+lYqtk
6vn4a+1w4JwijXo+TDSODyTkZlLPh1r90cT+0Czqw8H7kt3YHUjvg3CMOk51
XMkmloVjOi0uemeVxveEw0vh8GnT8EfUC5DVXeqn6UDzQ4C5yTvsDu906gWQ
sD5ueMcwlXoBWvvPn3zUlki9EMt8p4uZHNo/CKGX4Rk4rRZHvRAvXjG8+8LL
EuqFYFdqGcXoRlMfAVai+97Kx0LqI6DTlpewLuRRH4GJLS7Hvt2f+gho6PPm
fW94UR+JKY/6zvb9jtRH7u7vG0ODYivqI7HqlHXvgMSY+kiYbjBmgtZ1qI+C
aUntZkHVlvI8EQWbpe9XL+ybVrIsCmmfF/bp/3NByT1RWP/9ExNNky46fxHE
itykyakH1ItwcLArvoCzQr0ILhohZVGjz6kXofRS0y9f2DNo/eOQ52T725ur
e/0HyuvC6g==
    "]]},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, Automatic},
  PlotRange->Automatic,
  PlotRangeClipping->True]], "Output",
 CellChangeTimes->{3.5086623367123995`*^9}]
}, Open  ]],

Cell["Timings. We compute the function values a million times.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"start", "=", 
   RowBox[{"AbsoluteTime", "[", "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"Makeplot", "[", "2000000", "]"}], ";"}], "\n", 
 RowBox[{"MathCodeTime", "=", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"AbsoluteTime", "[", "]"}], "-", "start"}], ")"}], "/", 
   "2000000"}]}]}], "Input"],

Cell[BoxData["2.4218439999999999997886`7.136721154341411*^-7"], "Output",
 CellChangeTimes->{3.508662337212393*^9}]
}, Open  ]],

Cell["Internal computation of the function values 10000 times.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"start", "=", 
   RowBox[{"AbsoluteTime", "[", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Array", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      SuperscriptBox[
       RowBox[{"Sin", "[", 
        RowBox[{"0.1", "#"}], "]"}], "2"], 
      RowBox[{"Exp", "[", 
       RowBox[{"-", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"0.03", "#"}], ")"}], "2"]}], "]"}]}], " ", "&"}], ",", 
    "100000"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"MathematicaTime", "=", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"AbsoluteTime", "[", "]"}], "-", "start"}], ")"}], "/", 
   "100000"}]}]}], "Input"],

Cell[BoxData["0.00003187459199999999999984058`7.954989627933038"], "Output",
 CellChangeTimes->{3.508662340415477*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MathematicaTime", "/", "MathCodeTime"}]], "Input"],

Cell[BoxData["131.6129032258064516238`7.075283459495574"], "Output",
 CellChangeTimes->{3.508662340462351*^9}]
}, Open  ]],

Cell[TextData[{
 "We observe that compiled code gives two orders of magnitude faster \
performance. However, using the Compile command in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " a lot better performance can be obtained in this case."
}], "Text"],

Cell["\<\
Uninstall the executable, remove the temporary files, and remove the functions\
\>", "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"CleanMathCodeFiles", "[", 
   RowBox[{
    RowBox[{"Confirm", "\[Rule]", "False"}], ",", 
    RowBox[{"CleanAllBut", "\[Rule]", 
     RowBox[{"{", "}"}]}]}], "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"Remove", "[", 
   RowBox[{"extsqr", ",", "extexp", ",", "extoscillation", ",", "Makeplot"}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Off", "[", 
   RowBox[{"DeleteFile", "::", "nffil"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"DeleteFile", "[", "\"\<external1.obj\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"DeleteFile", "[", "\"\<external2.obj\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"DeleteFile", "[", "\"\<external1.o\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"DeleteFile", "[", "\"\<external2.o\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"On", "[", 
   RowBox[{"DeleteFile", "::", "nffil"}], "]"}], ";"}]}], "Input"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Other Features", "Section"],

Cell[TextData[{
 "In this section we briefly describe some more features of ",
 StyleBox["MathCode",
  FontSlant->"Italic"],
 "."
}], "Text"],

Cell[CellGroupData[{

Cell["Common subexpession elimination", "Subsubsection"],

Cell[TextData[{
 "Consider the following function whose definition contains a number of \
common subexpressions (e.g. ",
 Cell[BoxData[
  FormBox[
   StyleBox[
    RowBox[{"1", "+", 
     SuperscriptBox["x", "2"]}],
    FontColor->RGBColor[1, 0, 0]], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   StyleBox[
    SqrtBox[
     StyleBox[
      RowBox[{"1", "+", 
       SuperscriptBox["x", "2"]}],
      FontColor->RGBColor[1, 0, 0]]],
    FontColor->RGBColor[0, 0, 1]], TraditionalForm]]],
 ")."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"g", "[", 
    RowBox[{"Real", " ", "x_"}], "]"}], "\[Rule]", "Real"}], ":=", 
  RowBox[{
   FractionBox[
    RowBox[{"x", " ", 
     RowBox[{"Cos", "[", "x", "]"}], " ", 
     RowBox[{"Cos", "[", 
      StyleBox[
       SqrtBox[
        StyleBox[
         RowBox[{"1", "+", 
          SuperscriptBox["x", "2"]}],
         FontColor->RGBColor[1, 0, 0]]],
       FontColor->RGBColor[0, 0, 1]], "]"}]}], 
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       StyleBox[
        RowBox[{"1", "+", 
         SuperscriptBox["x", "2"]}],
        FontColor->RGBColor[1, 0, 0]], ")"}], 
      RowBox[{"3", "/", "2"}]], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       SuperscriptBox[
        RowBox[{"Cos", "[", "x", "]"}], "2"]}], ")"}]}]], "-", 
   FractionBox[
    RowBox[{"2", " ", "x", " ", 
     RowBox[{"Cos", "[", "x", "]"}], " ", 
     RowBox[{"Sin", "[", 
      StyleBox[
       SqrtBox[
        StyleBox[
         RowBox[{"1", "+", 
          SuperscriptBox["x", "2"]}],
         FontColor->RGBColor[1, 0, 0]]],
       FontColor->RGBColor[0, 0, 1]], "]"}]}], 
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       StyleBox[
        RowBox[{"1", "+", 
         SuperscriptBox["x", "2"]}],
        FontColor->RGBColor[1, 0, 0]], ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       SuperscriptBox[
        RowBox[{"Cos", "[", "x", "]"}], "2"]}], ")"}]}]], "+", 
   FractionBox[
    RowBox[{"2", " ", 
     SuperscriptBox[
      RowBox[{"Cos", "[", "x", "]"}], "2"], " ", 
     RowBox[{"Sin", "[", "x", "]"}], " ", 
     RowBox[{"Sin", "[", 
      StyleBox[
       SqrtBox[
        StyleBox[
         RowBox[{"1", "+", 
          SuperscriptBox["x", "2"]}],
         FontColor->RGBColor[1, 0, 0]]],
       FontColor->RGBColor[0, 0, 1]], "]"}]}], 
    RowBox[{
     RowBox[{"(", 
      StyleBox[
       RowBox[{"1", "+", 
        SuperscriptBox["x", "2"]}],
       FontColor->RGBColor[1, 0, 0]], ")"}], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", 
        SuperscriptBox[
         RowBox[{"Cos", "[", "x", "]"}], "2"]}], ")"}], "2"]}]], "-", 
   FractionBox[
    RowBox[{
     RowBox[{"Sin", "[", "x", "]"}], " ", 
     RowBox[{"Sin", "[", 
      StyleBox[
       SqrtBox[
        StyleBox[
         RowBox[{"1", "+", 
          SuperscriptBox["x", "2"]}],
         FontColor->RGBColor[1, 0, 0]]],
       FontColor->RGBColor[0, 0, 1]], "]"}]}], 
    RowBox[{
     RowBox[{"(", 
      StyleBox[
       RowBox[{"1", "+", 
        SuperscriptBox["x", "2"]}],
       FontColor->RGBColor[1, 0, 0]], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       SuperscriptBox[
        RowBox[{"Cos", "[", "x", "]"}], "2"]}], ")"}]}]]}]}]], "Input"],

Cell["\<\
There are algorithms to evaluate functions containing common subexpressions \
very efficiently. The basic idea is to evaluate common subexpressions only \
once and put the results in temporary variables.\
\>", "Text"],

Cell["We plot the function", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"pl1", "=", 
  RowBox[{"Plot", "[", 
   RowBox[{
    RowBox[{"g", "[", "x", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", 
      RowBox[{"-", "5"}], ",", "5"}], "}"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwtm3k4lO/3x5kZys4zkoQoKYSQpehzP0UpSSpbkp0KSfYtREkqIXvIOvZl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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  PlotRange->{{-5, 5}, {-0.38799835911796116`, 0.38799987357568794`}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.5086623416029615`*^9}]
}, Open  ]],

Cell[TextData[{
 "We can tell ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " to do Common Subexpression Elimination (CSE) by the option \
EvaluateFunctions to CompileCode[ ] or BuildCode[ ]."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BuildCode", "[", 
  RowBox[{"EvaluateFunctions", "\[Rule]", 
   RowBox[{"{", "g", "}"}]}], "]"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 1 function(s)\"\>"], "Print",
 CellChangeTimes->{3.508662344134179*^9}]
}, Open  ]],

Cell[TextData[{
 "The generated executable is connected to ",
 StyleBox["Mathematica",
  FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086623477122583`*^9}]
}, Open  ]],

Cell["and we plot the externally evaluated function", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"pl2", "=", 
  RowBox[{"Plot", "[", 
   RowBox[{
    RowBox[{"g", "[", "x", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", 
      RowBox[{"-", "5"}], ",", "5"}], "}"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwtm3k4lN/7x5kZimR5RpIQUllCKBR9zlOUQlJZSrJTIYnsQpSQhGzZso49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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  PlotRange->{{-5, 5}, {-0.38799835911796116`, 0.38799987357568794`}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.508662347759133*^9}]
}, Open  ]],

Cell["We take a look at the generated C++ file", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<Global.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"Global.h\"

#include \"Global.icc\"

#include <math.h>
void Global_TGlobalInit ()
{
; 
}

double Global_Tg ( const double &x)
{
    double mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    double mc_T5;
    double mc_T6;
    double mc_T7;
    double mc_T8;
    double mc_T9;
    double mc_T10;
    double mc_T11;
    double mc_T12;
    double mc_T13;
    double mc_T14;
    double mc_T15;
    double mc_T16;
    double mc_T17;
    double mc_T18;
    double mc_T19;
    double mc_T20;
    double mc_T21;
    double mc_T22;
    double mc_T23;
    double mc_T24;
    double mc_T25;
    mc_T1 = (x*x);
    mc_T2 = 1+mc_T1;
    mc_T3 = cos(x);
    mc_T4 = (mc_T3*mc_T3);
    mc_T5 = 1+mc_T4;
    mc_T6 = mc_T5*mc_T2;
    mc_T7 =                        0.5;
    mc_T8 = pow(mc_T2, mc_T7);
    mc_T9 = sin(mc_T8);
    mc_T10 = sin(x);
    mc_T11 = mc_T10*mc_T9;
    mc_T12 = mc_T11/mc_T6;
    mc_T13 = -mc_T12;
    mc_T14 = pow(mc_T5, -2);
    mc_T15 = 2*mc_T4*mc_T14*mc_T10*mc_T9;
    mc_T16 = mc_T15/mc_T2;
    mc_T17 = pow(mc_T2, -2);
    mc_T18 = -2*x*mc_T17*mc_T3*mc_T9;
    mc_T19 = mc_T18/mc_T5;
    mc_T20 = cos(mc_T8);
    mc_T21 =                       -1.5;
    mc_T22 = pow(mc_T2, mc_T21);
    mc_T23 = x*mc_T22*mc_T3*mc_T20;
    mc_T24 = mc_T23/mc_T5;
    mc_T25 = mc_T24+mc_T19+mc_T16+mc_T13;
    return mc_T25;
}
\
\>", "Print",
 CellChangeTimes->{3.5086623477903824`*^9}]
}, Open  ]],

Cell["\<\
Note how the computation of the function has been divided into small \
subexpressions that are only evaluated once and then stored in temporary \
variables for future use. This gives very efficient code for large functions!\
\
\>", "Text"],

Cell[TextData[{
 "Uninstall the executable, remove the temporary files, and remove the \
function ",
 StyleBox["g",
  FontFamily->"Courier",
  FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"CleanMathCodeFiles", "[", 
   RowBox[{
    RowBox[{"Confirm", "\[Rule]", "False"}], ",", 
    RowBox[{"CleanAllBut", "\[Rule]", 
     RowBox[{"{", "}"}]}]}], "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"Remove", "[", "g", "]"}], ";"}]}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Stand-Alone Code", "Subsubsection"],

Cell[TextData[{
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " can be used to generate C++ programs that can be executed independently of \
",
 StyleBox["Mathematica,",
  FontSlant->"Italic"],
 " that is to produce stand-alone programs. We define a simple function that \
prints something as result (hence, return type ",
 StyleBox["Null",
  FontFamily->"Courier"],
 ")"
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f", "[", 
    RowBox[{"Real", " ", "x_"}], "]"}], "\[Rule]", "Null"}], ":=", 
  RowBox[{"Print", "[", 
   SuperscriptBox["x", "2"], "]"}]}]], "Input"],

Cell[TextData[{
 "The option ",
 StyleBox["MainFileAndFunction",
  FontFamily->"Times New Roman"],
 StyleBox[" ",
  FontSlant->"Italic"],
 "is used to specify the main-file. The functions defined in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " must have the prefix Global_T (",
 StyleBox["packagename",
  FontSlant->"Italic"],
 "_T in general) to be recognized in the main file."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CompilePackage", "[", 
  RowBox[{
  "MainFileAndFunction", "\[Rule]", 
   "\"\<\n\t\t\tint main()\n\t\t\t\t{\n\t\t\t\t\tGlobal_Tf(1.2);\n            \
        return 0;\n\t\t\t\t}\n\>\""}], "]"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 1 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5086623513840866`*^9}]
}, Open  ]],

Cell["The stand-alone executable is generated", "Text"],

Cell[BoxData[
 RowBox[{"MakeBinary", "[", 
  RowBox[{"StandAloneExecutable", "\[Rule]", "True"}], "]"}]], "Input"],

Cell["\<\
This generates just three files and compiles them to an executable\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<Global.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"Global.h\"

#include \"Global.icc\"

#include <math.h>
void Global_Tf ( const double &x)
{
    cout << ((x*x));
    cout << \"\\n\";; 
}

void Global_TGlobalInit ()
{
; 
}
\
\>", "Print",
 CellChangeTimes->{3.508662354462172*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<Global.h\>\"", "]"}]], "Input"],

Cell["\<\
#ifndef Global_h
#define Global_h

#define LM_NNNN
/* Level of the library */
#include \"lightmat.h\"

void Global_Tf ( const double &);

void Global_TGlobalInit ();

#endif\
\>", "Print",
 CellChangeTimes->{3.5086623544777966`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<Globalmain.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"Global.h\"

\t\t\tint main()
\t\t\t\t{
\t\t\t\t\tGlobal_Tf(1.2);
                    return 0;
\t\t\t\t}
\
\>", "Print",
 CellChangeTimes->{3.5086623544934216`*^9}]
}, Open  ]],

Cell["\<\
We run the program \"Global\" and put the output in the file named result.txt\
\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Run", "[", "\"\<./Global.exe > result.txt\>\"", "]"}], 
  ";"}]], "Input"],

Cell["The content of result.txt", "Text"],

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<result.txt\>\"", "]"}]], "Input"],

Cell[TextData[{
 "Uninstall the executable, remove the temporary files, and remove the \
function ",
 StyleBox["f",
  FontFamily->"Courier",
  FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"CleanMathCodeFiles", "[", 
   RowBox[{
    RowBox[{"Confirm", "\[Rule]", "False"}], ",", 
    RowBox[{"CleanAllBut", "\[Rule]", 
     RowBox[{"{", "}"}]}]}], "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"Remove", "[", "f", "]"}], ";"}]}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Extended matrix operations", "Subsubsection"],

Cell["\<\
When dealing with matrices it is very convenient to have short notation for \
part extraction. The functionality of Part[] or [[ ]] has been extended to \
obtain this.\
\>", "Text"],

Cell["\<\
Consider the following 5\[Cross]5 matrix:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"A", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"a", "[", 
      RowBox[{"i", ",", "j"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"i", ",", "4"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"j", ",", "5"}], "}"}]}], "]"}]}], ";", 
  RowBox[{"A", "//", "MatrixForm"}]}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "5"}], "]"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086623556184072`*^9}]
}, Open  ]],

Cell["We extract row 2 to 4", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"A", "\[LeftDoubleBracket]", 
   RowBox[{"2", "|", "4"}], "\[RightDoubleBracket]"}], " ", "//", " ", 
  "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "5"}], "]"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086623556340322`*^9}]
}, Open  ]],

Cell["\<\
We extract the elements in all rows which also belongs to column 3 and higher\
\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"A", "\[LeftDoubleBracket]", 
   RowBox[{"_", ",", 
    RowBox[{"3", "|", "_"}]}], "\[RightDoubleBracket]"}], "//", 
  "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "5"}], "]"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086623556652813`*^9}]
}, Open  ]],

Cell[TextData[{
 "We can assign values to a submatrix of ",
 Cell[BoxData[
  FormBox["A", TraditionalForm]]],
 "."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"A", "\[LeftDoubleBracket]", 
    RowBox[{
     RowBox[{"2", "|", "3"}], ",", 
     RowBox[{"2", "|", "3"}]}], "\[RightDoubleBracket]"}], " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "2"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", "4"}], "}"}]}], "}"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"A", " ", "//", " ", "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "1"}], "]"}], "1", "2", 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "1"}], "]"}], "3", "4", 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "5"}], "]"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086623556809063`*^9}]
}, Open  ]],

Cell["All these operations belongs to the compilable subset.", "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Code Generation from a Package", "Section"],

Cell["\<\
In this section we generate code for a package called MySurface. The example \
is similar to what is shown in the SinSurface.nb demo notebook.\
\>", "Text"],

Cell[CellGroupData[{

Cell["The Package", "Subsection"],

Cell[CellGroupData[{

Cell["Beginning of Package", "Subsubsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"BeginPackage", "[", 
     RowBox[{"\"\<MySurface`\>\"", ",", 
      RowBox[{"{", "MathCodeContexts", "}"}]}], "]"}], ";"}], "\n", 
   RowBox[{"Off", "[", 
    RowBox[{"General", "::", "\"\<spell1\>\""}], "]"}], "\n", 
   RowBox[{
    RowBox[{"Clear", "[", "\"\<MySurface`*\>\"", "]"}], ";"}]}], 
  " "}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Exported Symbols", "Subsubsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Begin", "[", "\"\<MySurface`\>\"", "]"}], ";"}], "\n", 
 RowBox[{"\t", 
  RowBox[{"xyMatrix", ";"}]}], "\n", 
 RowBox[{"\t", 
  RowBox[{"sin", ";"}]}], "\n", 
 RowBox[{"\t", 
  RowBox[{"cos", ";"}]}], "\n", 
 RowBox[{"\t", 
  RowBox[{"arcTan", ";"}]}], "\n", 
 RowBox[{"\t", 
  RowBox[{"sinxy", ";"}]}], "\n", 
 RowBox[{"\t", 
  RowBox[{"computeGridValues", ";"}]}], "\n", 
 RowBox[{
  RowBox[{"End", "[", "]"}], ";"}]}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compilation Options", "Subsubsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCompilationOptions", "[", "\n", "\t", 
   RowBox[{
    RowBox[{"EvaluateFunctions", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"sin", ",", "cos", ",", "arcTan"}], "}"}]}], ",", "\n", "\t", 
    RowBox[{"CallBackFunctions", "\[Rule]", 
     RowBox[{"{", "FresnelS", "}"}]}], ",", "\n", "\t", 
    RowBox[{"StandAloneExecutable", "\[Rule]", "True"}], ",", "\n", "\t", 
    RowBox[{
    "MainFileAndFunction", "\[Rule]", 
     "\"\<\n\t\t\tint main(){\n\t\t\t\tint i, j;\n\t\t\t\ti=-5;\n\t\t\t\t\
while(i<=5){\n\t\t\t\t\t\ti++; j=-5;\n\t\t\t\t\t\twhile(j<=5){\n\t\t\t\t\t\t\t\
\tj++;\n\t\t\t\t\t\t\t\tprintf(\\\"%f  \\\",MySurface_Tsinxy(0.4*i,0.4*j)); }\
\n\t\t\t\t\t\tprintf(\\\"\\\\n\\\"); } \n             return 0; }\n\t\t\t\
\>\""}]}], "\n", "]"}], ";"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Private Context", "Subsubsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Begin", "[", "\"\<MySurface`Private`\>\"", "]"}], ";"}]], "Input"],

Cell["A global variable.", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Declare", "[", " ", 
   RowBox[{
    RowBox[{"Real", "[", 
     RowBox[{"21", ",", "21"}], "]"}], " ", "xyMatrix"}], " ", "]"}], 
  ";"}]], "Input"],

Cell[TextData[{
 "An function that have to be computed by a call-back to ",
 StyleBox["Mathematica.",
  FontSlant->"Italic"]
}], "Text"],

Cell[BoxData[
 RowBox[{"Declare", "[", 
  RowBox[{
   RowBox[{"FresnelS", "[", 
    RowBox[{"Real", " ", "x_"}], "]"}], "\[Rule]", "Real"}], "]"}]], "Input"],

Cell[TextData[{
 "A number of functions with right-hand-sides (rhs) not in the compilable \
subset. However, the rhs can be evaluated by ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " to expressions that belong to the compilable subset."
}], "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"sin", "[", 
    RowBox[{"Real", " ", "x_"}], " ", "]"}], "\[Rule]", "Real"}], ":=", 
  RowBox[{
   RowBox[{"Normal", "[", 
    RowBox[{"Series", "[", 
     RowBox[{
      RowBox[{"Sin", "[", "z", "]"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"z", ",", "0", ",", "10"}], "}"}]}], "]"}], "]"}], " ", "/.", 
   " ", 
   RowBox[{"z", " ", "\[Rule]", " ", "x", " "}]}]}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"cos", "[", 
    RowBox[{"Real", " ", "x_"}], " ", "]"}], "\[Rule]", "Real"}], ":=", 
  RowBox[{"1", "-", 
   RowBox[{"Integrate", "[", 
    RowBox[{
     RowBox[{"sin", "[", "z", "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"z", ",", "0", ",", "x"}], "}"}]}], "]"}]}]}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"arcTan", "[", 
    RowBox[{
     RowBox[{"Real", " ", "x_"}], ",", 
     RowBox[{"Real", " ", "y_"}]}], "]"}], "\[Rule]", "Real"}], ":=", " ", 
  RowBox[{
   RowBox[{"If", "[", 
    RowBox[{
     RowBox[{"x", "<", "0"}], ",", "Pi", ",", "0"}], "]"}], "+", 
   RowBox[{"If", "[", 
    RowBox[{
     RowBox[{"x", "==", "0"}], ",", 
     RowBox[{
      RowBox[{"Sign", "[", "y", "]"}], "*", 
      RowBox[{"Pi", "/", "2"}]}], ",", " ", 
     RowBox[{"ArcTan", "[", 
      RowBox[{"y", "/", "x"}], "]"}]}], " ", "]"}]}]}]}], "Input"],

Cell["The main function", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"sinxy", "[", 
     RowBox[{
      RowBox[{"Real", " ", "x_"}], ",", 
      RowBox[{"Real", " ", "y_"}]}], "]"}], "\[Rule]", "Real"}], ":=", " ", 
   "\n", "\t", 
   RowBox[{"Module", "[", "\n", "\t\t", 
    RowBox[{
     RowBox[{"{", " ", 
      RowBox[{"Real", " ", 
       RowBox[{"{", 
        RowBox[{"r", ",", "xx", ",", "yy"}], "}"}]}], " ", "}"}], ",", "\n", 
     "\t\t", 
     RowBox[{
      RowBox[{"r", "=", 
       RowBox[{"Sqrt", "[", 
        RowBox[{
         RowBox[{"x", "^", "2"}], "+", 
         RowBox[{"y", "^", "2"}]}], "]"}]}], ";", "\n", "\t\t", 
      RowBox[{"xx", "=", " ", 
       RowBox[{"r", "*", 
        RowBox[{"cos", "[", 
         RowBox[{"arcTan", "[", 
          RowBox[{"x", ",", "y"}], "]"}], "]"}]}]}], ";", "\n", "\t\t", 
      RowBox[{"yy", "=", 
       RowBox[{"r", "*", 
        RowBox[{"sin", "[", 
         RowBox[{"arcTan", "[", 
          RowBox[{"x", ",", "y"}], "]"}], "]"}]}]}], ";", "\n", "\t\t", 
      RowBox[{"sin", "[", 
       RowBox[{"xx", "+", "yy"}], "]"}]}]}], "\n", "  ", "]"}]}], 
  ";"}]], "Input"],

Cell["\<\
A function that computes function values of sinxy[ ] over a grid. A scaling \
using the FresnelS[ ] function is also performed. \
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"computeGridValues", "[", 
    RowBox[{"Integer", " ", "n_"}], "]"}], " ", "\[Rule]", " ", 
   RowBox[{"Real", "[", 
    RowBox[{"21", ",", " ", "21"}], "]"}]}], " ", ":=", " ", "\n", "  ", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"Integer", " ", 
      RowBox[{"{", 
       RowBox[{"i", ",", " ", "j", ",", " ", "k"}], "}"}]}], "}"}], ",", "\n",
     "  ", 
    RowBox[{
     RowBox[{"For", "[", 
      RowBox[{
       RowBox[{"k", "=", "1"}], ",", 
       RowBox[{"k", "\[LessEqual]", "n"}], ",", 
       RowBox[{"xyMatrix", "=", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"sinxy", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"-", "2."}], "+", 
             RowBox[{"0.2", "*", "i"}]}], ",", 
            RowBox[{
             RowBox[{"-", " ", "2."}], "+", 
             RowBox[{"0.2", "*", "j"}]}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "1", ",", "21"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"j", ",", "1", ",", "21"}], "}"}]}], "]"}]}], ",", 
       RowBox[{"k", "++"}]}], "]"}], ";", "\n", "\t\t", 
     RowBox[{
      RowBox[{"FresnelS", "[", "1.", "]"}], "*", "xyMatrix"}]}]}], 
   "]"}]}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["End of Private Context", "Subsubsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"End", "[", "]"}], ";"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["End of Package", "Subsubsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"EndPackage", "[", "]"}], ";"}], " "}]], "Input"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Generate code for all functions in the package", "Subsection"],

Cell["\<\
Generate C++ code according to the compilation options given in the package\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CompilePackage", "[", "\"\<MySurface`\>\"", "]"}], ";"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 6 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5086623584777455`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["The generated code", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<MySurface.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"MySurface.h\"

#include \"MySurface.icc\"

#include <math.h>
doubleNN MySurface_TcomputeGridValues ( const int &n)
{
    int i;
    int j;
    int k;
    int j_D7540_D7543;
    int i_D7539_D7545;
    k = 1;
    while (k <= n)
    {
        k = k+1;
        MySurface_TxyMatrix.SetShape(Max(21,0), Max(21,0) );
        i_D7539_D7545 = 1;
        while (i_D7539_D7545 <= 21)
        {
            j_D7540_D7543 = 1;
            while (j_D7540_D7543 <= 21)
            {
                MySurface_TxyMatrix(1+i_D7539_D7545+-1, 1+j_D7540_D7543+-1) = \

                    MySurface_Tsinxy (-2.+0.2*i_D7539_D7545, -2.+0.2*
                    j_D7540_D7543);
                j_D7540_D7543 = j_D7540_D7543+1;
            }
            i_D7539_D7545 = i_D7539_D7545+1;
        }
    }
    return FresnelS (1.)*MySurface_TxyMatrix;
}

void MySurface_TMySurfaceInit ()
{
; 
}

double MySurface_Tsinxy ( const double &x, const double &y)
{
    double r;
    double xx;
    double yy;
    r = sqrt((x*x)+(y*y));
    xx = r*MySurface_Tcos (MySurface_TarcTan (x, y));
    yy = r*MySurface_Tsin (MySurface_TarcTan (x, y));
    return MySurface_Tsin (xx+yy);
}

double FresnelS ( const double &_D4)
;  /* EXTERNAL   */

double MySurface_TarcTan ( const double &x, const double &y)
{
    int mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    int mc_T5;
    double mc_T6;
    double mc_T7;
    int mc_T8;
    double mc_T9;
    double mc_T10;
    mc_T1 = x < 0;
    if (mc_T1)
    {
        mc_T2 = 3.14159265358979323846;
    }
    else
    {
        mc_T2 = 0;
    }
    mc_T3 = y/x;
    mc_T4 = atan(mc_T3);
    mc_T5 = sign (y);
    mc_T6 = 3.14159265358979323846*mc_T5;
    mc_T7 = mc_T6/2;
    mc_T8 = x == 0;
    if (mc_T8)
    {
        mc_T9 = mc_T7;
    }
    else
    {
        mc_T9 = mc_T4;
    }
    mc_T10 = mc_T9+mc_T2;
    return mc_T10;
}

double MySurface_Tcos ( const double &x)
{
    double mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    double mc_T5;
    double mc_T6;
    double mc_T7;
    double mc_T8;
    double mc_T9;
    double mc_T10;
    double mc_T11;
    double mc_T12;
    double mc_T13;
    double mc_T14;
    mc_T1 = pow(x, 10);
    mc_T2 =    -2.755731922398589e-007;
    mc_T3 = mc_T2*mc_T1;
    mc_T4 = pow(x, 8);
    mc_T5 = mc_T4/40320;
    mc_T6 = pow(x, 6);
    mc_T7 =      -0.001388888888888889;
    mc_T8 = mc_T7*mc_T6;
    mc_T9 = (x*x*x*x);
    mc_T10 = mc_T9/24;
    mc_T11 = (x*x);
    mc_T12 =                       -0.5;
    mc_T13 = mc_T12*mc_T11;
    mc_T14 = 1+mc_T13+mc_T10+mc_T8+mc_T5+mc_T3;
    return mc_T14;
}

double MySurface_Tsin ( const double &x)
{
    double mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    double mc_T5;
    double mc_T6;
    double mc_T7;
    double mc_T8;
    double mc_T9;
    double mc_T10;
    double mc_T11;
    mc_T1 = pow(x, 9);
    mc_T2 = mc_T1/362880;
    mc_T3 = pow(x, 7);
    mc_T4 =     -0.0001984126984126984;
    mc_T5 = mc_T4*mc_T3;
    mc_T6 = pow(x, 5);
    mc_T7 = mc_T6/120;
    mc_T8 = (x*x*x);
    mc_T9 =        -0.1666666666666667;
    mc_T10 = mc_T9*mc_T8;
    mc_T11 = x+mc_T10+mc_T7+mc_T5+mc_T2;
    return mc_T11;
}

doubleNN MySurface_TxyMatrix(21, 21);\
\>", "Print",
 CellChangeTimes->{3.508662358571494*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Use function expansion to avoid the auxiliary functions", "Subsection"],

Cell["\<\
Generate C++ code where the body of sinxy[ ] is symbolically expanded before \
code generation. This makes the auxiliary functions sin, cos, arcTan \
unnecessary and we choose not to compile them.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CompilePackage", "[", 
   RowBox[{"\"\<MySurface`\>\"", ",", "\n", "\t", 
    RowBox[{"EvaluateFunctions", "\[Rule]", 
     RowBox[{"{", "sinxy", "}"}]}], ",", "\n", "\t", 
    RowBox[{"UnCompiledFunctions", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"sin", ",", "cos", ",", "arcTan"}], "}"}]}]}], "]"}], 
  ";"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 3 function(s)\"\>"], "Print",
 CellChangeTimes->{3.508662363462057*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["The generated code", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<MySurface.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"MySurface.h\"

#include \"MySurface.icc\"

#include <math.h>
doubleNN MySurface_TcomputeGridValues ( const int &n)
{
    int i;
    int j;
    int k;
    int j_D8315_D8318;
    int i_D8314_D8320;
    k = 1;
    while (k <= n)
    {
        k = k+1;
        MySurface_TxyMatrix.SetShape(Max(21,0), Max(21,0) );
        i_D8314_D8320 = 1;
        while (i_D8314_D8320 <= 21)
        {
            j_D8315_D8318 = 1;
            while (j_D8315_D8318 <= 21)
            {
                MySurface_TxyMatrix(1+i_D8314_D8320+-1, 1+j_D8315_D8318+-1) = \

                    MySurface_Tsinxy (-2.+0.2*i_D8314_D8320, -2.+0.2*
                    j_D8315_D8318);
                j_D8315_D8318 = j_D8315_D8318+1;
            }
            i_D8314_D8320 = i_D8314_D8320+1;
        }
    }
    return FresnelS (1.)*MySurface_TxyMatrix;
}

void MySurface_TMySurfaceInit ()
{
; 
}

double FresnelS ( const double &_D5)
;  /* EXTERNAL   */

double MySurface_Tsinxy ( const double &x, const double &y)
{
    int mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    int mc_T5;
    double mc_T6;
    double mc_T7;
    int mc_T8;
    double mc_T9;
    double mc_T10;
    double mc_T11;
    double mc_T12;
    double mc_T13;
    double mc_T14;
    double mc_T15;
    double mc_T16;
    double mc_T17;
    double mc_T18;
    double mc_T19;
    double mc_T20;
    double mc_T21;
    double mc_T22;
    double mc_T23;
    double mc_T24;
    double mc_T25;
    double mc_T26;
    double mc_T27;
    double mc_T28;
    double mc_T29;
    double mc_T30;
    double mc_T31;
    double mc_T32;
    double mc_T33;
    double mc_T34;
    double mc_T35;
    double mc_T36;
    double mc_T37;
    double mc_T38;
    double mc_T39;
    double mc_T40;
    double mc_T41;
    double mc_T42;
    double mc_T43;
    double mc_T44;
    double mc_T45;
    double mc_T46;
    double mc_T47;
    double mc_T48;
    double mc_T49;
    mc_T1 = x < 0;
    if (mc_T1)
    {
        mc_T2 = 3.14159265358979323846;
    }
    else
    {
        mc_T2 = 0;
    }
    mc_T3 = y/x;
    mc_T4 = atan(mc_T3);
    mc_T5 = sign (y);
    mc_T6 = 3.14159265358979323846*mc_T5;
    mc_T7 = mc_T6/2;
    mc_T8 = x == 0;
    if (mc_T8)
    {
        mc_T9 = mc_T7;
    }
    else
    {
        mc_T9 = mc_T4;
    }
    mc_T10 = mc_T9+mc_T2;
    mc_T11 = pow(mc_T10, 8);
    mc_T12 = pow(mc_T10, 6);
    mc_T13 = -90*mc_T12;
    mc_T14 = (mc_T10*mc_T10*mc_T10*mc_T10);
    mc_T15 = 5040*mc_T14;
    mc_T16 = (mc_T10*mc_T10);
    mc_T17 = -151200*mc_T16;
    mc_T18 = 1814400+mc_T17+mc_T15+mc_T13+mc_T11;
    mc_T19 =    -2.755731922398589e-007;
    mc_T20 = mc_T19*mc_T16*mc_T18;
    mc_T21 = 1+mc_T20;
    mc_T22 =                        0.5;
    mc_T23 = (y*y);
    mc_T24 = (x*x);
    mc_T25 = mc_T24+mc_T23;
    mc_T26 = pow(mc_T25, mc_T22);
    mc_T27 = mc_T26*mc_T21;
    mc_T28 = pow(mc_T10, 9);
    mc_T29 = mc_T28/362880;
    mc_T30 = pow(mc_T10, 7);
    mc_T31 =     -0.0001984126984126984;
    mc_T32 = mc_T31*mc_T30;
    mc_T33 = pow(mc_T10, 5);
    mc_T34 = mc_T33/120;
    mc_T35 = (mc_T10*mc_T10*mc_T10);
    mc_T36 =        -0.1666666666666667;
    mc_T37 = mc_T36*mc_T35;
    mc_T38 = mc_T9+mc_T2+mc_T37+mc_T34+mc_T32+mc_T29;
    mc_T39 = mc_T26*mc_T38;
    mc_T40 = mc_T39+mc_T27;
    mc_T41 = pow(mc_T40, 9);
    mc_T42 = mc_T41/362880;
    mc_T43 = pow(mc_T40, 7);
    mc_T44 = mc_T31*mc_T43;
    mc_T45 = pow(mc_T40, 5);
    mc_T46 = mc_T45/120;
    mc_T47 = (mc_T40*mc_T40*mc_T40);
    mc_T48 = mc_T36*mc_T47;
    mc_T49 = mc_T39+mc_T27+mc_T48+mc_T46+mc_T44+mc_T42;
    return mc_T49;
}

doubleNN MySurface_TxyMatrix(21, 21);\
\>", "Print",
 CellChangeTimes->{3.5086623635401807`*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Generate Stand-Alone Code", "Subsection"],

Cell[TextData[{
 "Generate C++ code without code for FresnelS[ ] and computeGridValues[ ] \
since the main function does not use these (FresnalS[ ] cannot be evaluated \
without call-backs to ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ")."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CompilePackage", "[", 
   RowBox[{"\"\<MySurface`\>\"", ",", "\n", "\t", 
    RowBox[{"CallBackFunctions", "\[Rule]", 
     RowBox[{"{", "}"}]}], ",", "\n", "\t", 
    RowBox[{"UnCompiledFunctions", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"FresnelS", ",", "computeGridValues"}], "}"}]}]}], "]"}], 
  ";"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 4 function(s)\"\>"], "Print",
 CellChangeTimes->{3.508662366133898*^9}]
}, Open  ]],

Cell[TextData[{
 "For a standalone application three files are needed: ",
 StyleBox["MySurface.cc, MySurfaceain.cc,  MySurface.h ",
  FontWeight->"Bold"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<MySurfacemain.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"MySurface.h\"

\t\t\tint main(){
\t\t\t\tint i, j;
\t\t\t\ti=-5;
\t\t\t\twhile(i<=5){
\t\t\t\t\t\ti++; j=-5;
\t\t\t\t\t\twhile(j<=5){
\t\t\t\t\t\t\t\tj++;
\t\t\t\t\t\t\t\tprintf(\"%f  \",MySurface_Tsinxy(0.4*i,0.4*j)); }
\t\t\t\t\t\tprintf(\"\\n\"); } 
             return 0; }
\t\t\t\
\>", "Print",
 CellChangeTimes->{3.508662366180772*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<MySurface.h\>\"", "]"}]], "Input"],

Cell["\<\
#ifndef MySurface_h
#define MySurface_h

#define LM_NNNN
/* Level of the library */
#include \"lightmat.h\"

void MySurface_TMySurfaceInit ();

double MySurface_Tsinxy ( const double &, const double &);

double MySurface_TarcTan ( const double &, const double &);

double MySurface_Tcos ( const double &);

double MySurface_Tsin ( const double &);

extern doubleNN MySurface_TxyMatrix;
#endif\
\>", "Print",
 CellChangeTimes->{3.508662366196397*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{"MakeBinary", "[", "]"}]], "Input"],

Cell["\<\
We run the external program and put the result in a file: res.txt\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Run", "[", 
   RowBox[{
    RowBox[{"ToFileName", "[", 
     RowBox[{"\"\<.\>\"", ",", "\"\<MySurface.exe\>\""}], "]"}], "<>", 
    "\"\< > res.txt\>\""}], "]"}], ";"}]], "Input"],

Cell["The contents of res.txt", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<res.txt\>\"", "]"}]], "Input"],

Cell["\<\
-0.064994  -0.402676  -0.703618  -0.916872  -0.999783  -0.930739  -0.716180  \
-0.388589  0.000546  0.389759  0.717546  
-0.473108  -0.738141  -0.926789  -0.999940  -0.929806  -0.715918  -0.388663  \
0.000409  0.389647  0.717477  0.932088  
-0.813801  -0.953893  -0.999562  -0.924901  -0.714510  -0.388641  0.000273  \
0.389544  0.717418  0.932064  0.999576  
-0.994127  -0.984358  -0.889195  -0.696528  -0.387540  0.000136  0.389460  \
0.717376  0.932048  0.999577  0.909338  
-0.999578  -0.932041  -0.717358  -0.389420  0.000000  0.389419  0.717358  \
0.932041  0.999578  0.909345  0.675825  
-0.932040  -0.717356  -0.389418  -0.000000  0.389418  0.717356  0.932039  \
0.999578  0.909347  0.675830  0.336963  
-0.717356  -0.389418  -0.000000  0.389418  0.717356  0.932039  0.999578  \
0.909347  0.675831  0.336965  -0.049913  
-0.389418  -0.000000  0.389418  0.717356  0.932039  0.999578  0.909347  \
0.675831  0.336965  -0.049913  -0.412125  
-0.000000  0.389418  0.717356  0.932039  0.999578  0.909347  0.675831  \
0.336965  -0.049913  -0.412125  -0.661728  
0.389418  0.717356  0.932039  0.999578  0.909347  0.675831  0.336965  \
-0.049913  -0.412125  -0.661728  -0.685814  
0.717356  0.932039  0.999578  0.909347  0.675831  0.336965  -0.049913  \
-0.412125  -0.661728  -0.685814  -0.319099  \
\>", "Print",
 CellChangeTimes->{3.508662369305732*^9}]
}, Open  ]],

Cell["and a plot of the result", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ListPlot3D", "[", 
  RowBox[{"ReadList", "[", 
   RowBox[{"\"\<res.txt\>\"", ",", "Number", ",", 
    RowBox[{"RecordLists", "\[Rule]", "True"}]}], "]"}], "]"}]], "Input"],

Cell[BoxData[
 Graphics3DBox[GraphicsComplex3DBox[CompressedData["
1:eJxtnAdUVFcTx5dV3BVXXWGBBRbZxtoiYk006o499t4S7CI2LGCLWLDG3hWx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   "], {{
     {EdgeForm[None], GraphicsGroup3DBox[{Polygon3DBox[CompressedData["
1:eJwtmAn41lMexX/33kqptGlVWrRo074XqZRQKi0UsmvRXoqQNUoZFFlCxiAa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         "]], Polygon3DBox[CompressedData["
1:eJw1mHkcl1MWxu+9P2lqskS7Fi0KSfumRYoWLVoV7VkaWcaMIRM1Y5kZxjSN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         "]], Polygon3DBox[CompressedData["
1:eJwtlnmcTmUUx997r7E21sgyEhr7WrYwBlkbY4xd1mQrMclYkjHZprKFkWEw
ZY0I2XdKoUxZkyRZG7RJSZLl+/uc54/v55zze59733vPc855bum+Ce2H+qFQ
aCaEwfsE07xQqFkQCjWHFtAVLRGtFn5u/DzQAyahNUQrhl8c0mEyWhRaI4iG
9miD0Krg/4u9A33QZmBbolXErwSNoAyaj7YauwY6oQ3F1kDz8QPoBVPQmqKV
xX8SoqA02hyePw98hN8bbSr2WdZF4peDBnpOtBTWZIflen60x7EziHPBSvw4
tD7YA8RRcBO/M1oCtib3y4YfBithIVp+GMm6OsQfwgLifJAX/ideBRn4BaEA
3CNeDmn4eSA33CVeAfPxwyGR+9UirgDl4QOXs9t6Xiip/4G30G663Kfg/8d1
d+BP4hehKNoE4onwwOVLeVvk8nPLtz2fju9BCP7RXiinej+X77+xEVAC5rm9
v4FdBu/h54JhXF6TOBoiibORq0+wG7T/aJVhMcxSHWCXwGz87JDAtTWIO8IQ
4upcG8L3oDtMRLuG1BWKEjeH6sof6/Zi98HzaGOxF1nTAQoRx8NAtK+Jm8Jt
/LZoPbH7iRvADdUJWhW9B/fbid0F7dD6Yb9iTWO4pbpDq4jNwbqt2G3QF60w
dixrkuA+8QvuPVuzrip+NagPj6G9yZpssBS/ifZYOWDdZuwWWIo2B5tT/6M9
JU6Gtvifcd2ncJR4LoxDu0x8Bepzj56uJhrjl8J/Aprp/1VnaLuxezyrjWLY
yVw3Cdb49h5F0JKJx6mPiBur/tDCuHYjdhN0QxuBrY32CH642yft13TXsxex
5+EnzRO3pxe0j9AJ/0uWH4LjxPfUJzAVfzD2nOoLOrvcR8M1/H6qe+x04mmw
1vXWIrRA9QbxPNN4tHbKJWu2Qlm0tcQfwwA3Zza4nMbib2TNJijNuv7KGVoq
8WxYTzxcswNtMfESKBBY/87VcxPfg9jAej/d9ZB6KS6wvUz1rCfVmzGBzSLN
pAzfZq36TfNTc3S+bzNU/TsBOuAfdHMoC/9t9TX2O+JTcJr4B+Izqj/8Vqp1
5V15hCT8GOwR3/ISj/8F130Ox4iPER+FUfhNsIdUU9BGPcWaenBJNYsW53qm
PlzWM8ApeF39iD2MfQNa4u9kzQ7IJP6W+CSMVi1qT7F3Vdcwxbfz4Sw2C36G
mb6dN1exOSA7vAuvoV3BpkAPPTf3PwIn1S/6b7Q/iH+HX5RH6IqWSXwYThC/
AwOUL+Lv4YzmFyQp58S93NxIhdFov+pecF19Bb3RTqhulX/fzkKdibNcT1zT
TIOR+NdZ0w3yKVYveVZnqreDxK9CbbRVOnfU13pHqIeWobMY8gf2Hl3QdhPv
gfKB5bkFWnH8EhARWF66ezYfNCcqBFYvfT3rNfVcJbQuLpdP4+fEz+VmSTLa
Xzrv4DfiNNWg6lOzT3kl3kG8HQb7Ni93urz08qymVFsVA9vD4Z7Lcci+HWJc
nRTDP63fPJsNr3iWT+VVc7+1q8Ui+MeVb8/yF40thPYoFNa3CVodtIVcFwEH
8Nu4HijJ72exP0IrtEbYgmiZ6g94ztVi0cBqWLXcEi1K81ezQDXjWb3096x+
VEeVWR/reqoU/nnNNe0N2i69i5u1e3yb5ZrpL/t2Zmz3bX5qjr7k29zf5tvM
KeXZN9gsjY/AZk5DtPXE69SvxEeIv4FE9z4HXA3Vwk9nzQLIG1gNPaN5p7kO
+4kT4Cm0FfrugX3Ew6Au2mrVlMvfILRynp0JOhs099cTr4OBvp3rm9z9auKn
sWYehLNOH5I10JYRl1ENenbW6MwZ4tt5vde3/GsfRrg9VS9sgc2+/b/Om62u
/sbjZ3GvqzqrA+vLUdpX4nNQN7DeH4N2nviCZlZg50xHz2aQZlFkYPWmuhvj
eifT7VdVz2a9Zr7OSJ0zEZpNxFMhhNbBt7OpGv597APPvkf1XTrDt2/DS9iH
hcliMQ==
         "]]}]}, {}, {}, {}, {}}, {
     {GrayLevel[0], Line3DBox[CompressedData["
1:eJwt0DsvQ2EAx+FXi7Yc6pKUhKVmFZt21VkHU43U3LrfQiIkjZAIiS9glfAV
iLUJX4FYxSph8byJ4Xee/3ByzntOcaW12OwKISzrQHHfuiS854PG7GdO8YWv
KtkfLPOLVf7wVzU7nwphiROc1Ko9yxZ73bPGDLNat3PcYB/7tWkn3OIAtznI
vHbsIe5ymCPas0e5z4r3NDnDkhr2OOvxWUy0YH+7f56fnOM73zRtd1jkEx9V
sO/i2XgTz6lrXcWz61JZXeg8fpvO4reqrR6d6kTdOlZaRzpU6v///wEeCyRM

       "]]}, {
      Line3DBox[{417, 276, 416, 682, 610, 766, 418, 683, 570, 419, 684, 611, 
       767, 420, 685, 571, 421, 283, 422, 686, 612, 768, 423, 687, 572, 424, 
       688, 613, 769, 425, 689, 573, 426}], 
      Line3DBox[{430, 690, 574, 428, 691, 292, 432, 692, 575, 434, 694, 695, 
       576, 436, 696, 577, 438, 697, 578, 440, 698, 299, 442, 699, 579, 444, 
       701, 702, 580, 446, 703, 581, 448}], 
      Line3DBox[{447, 779, 621, 445, 778, 620, 700, 443, 777, 619, 441, 776, 
       298, 439, 775, 618, 437, 774, 617, 435, 773, 616, 693, 433, 772, 615, 
       431, 771, 291, 427, 770, 614, 429}], 
      Line3DBox[{450, 622, 780, 449, 704, 623, 781, 451, 307, 452, 705, 624, 
       782, 453, 706, 582, 454, 625, 783, 455, 707, 626, 784, 456, 314, 457, 
       708, 627, 785, 458, 709, 583, 459}], 
      Line3DBox[{463, 710, 584, 461, 711, 630, 789, 465, 712, 585, 467, 713, 
       323, 469, 714, 586, 471, 715, 587, 473, 716, 635, 796, 475, 717, 588, 
       477, 718, 330, 479, 719, 589, 481}], 
      Line3DBox[{480, 799, 637, 478, 798, 329, 476, 797, 636, 474, 795, 794, 
       634, 472, 793, 633, 470, 792, 632, 468, 791, 322, 466, 790, 631, 464, 
       788, 787, 629, 460, 786, 628, 462}], 
      Line3DBox[{56, 192, 57, 193, 194, 58, 195, 59, 196, 197, 60, 198, 61, 
       199, 62, 200, 201, 63, 202, 64, 203, 204, 65, 205, 66}], 
      Line3DBox[{483, 638, 800, 482, 720, 639, 801, 484, 640, 802, 485, 721, 
       641, 803, 486, 338, 487, 642, 804, 488, 722, 643, 805, 489, 644, 806, 
       490, 723, 645, 807, 491, 345, 492}], 
      Line3DBox[{494, 346, 493, 724, 646, 808, 495, 725, 590, 496, 726, 647, 
       809, 497, 727, 591, 498, 353, 499, 728, 648, 810, 500, 729, 592, 501, 
       730, 649, 811, 502, 731, 593, 503}], 
      Line3DBox[{507, 732, 594, 505, 733, 362, 509, 734, 595, 511, 736, 737, 
       596, 513, 738, 597, 515, 739, 598, 517, 740, 369, 519, 741, 599, 521, 
       743, 744, 600, 523, 745, 601, 525}], 
      Line3DBox[{524, 821, 657, 522, 820, 656, 742, 520, 819, 655, 518, 818, 
       368, 516, 817, 654, 514, 816, 653, 512, 815, 652, 735, 510, 814, 651, 
       508, 813, 361, 504, 812, 650, 506}], 
      Line3DBox[{527, 658, 822, 526, 746, 659, 823, 528, 377, 529, 747, 660, 
       824, 530, 748, 602, 531, 661, 825, 532, 749, 662, 826, 533, 384, 534, 
       750, 663, 827, 535, 751, 603, 536}], 
      Line3DBox[{540, 752, 604, 538, 753, 666, 831, 542, 754, 605, 544, 755, 
       393, 546, 756, 606, 548, 757, 607, 550, 758, 671, 838, 552, 759, 608, 
       554, 760, 400, 556, 761, 609, 558}], 
      Line3DBox[{557, 841, 673, 555, 840, 399, 553, 839, 672, 551, 837, 836, 
       670, 549, 835, 669, 547, 834, 668, 545, 833, 392, 543, 832, 667, 541, 
       830, 829, 665, 537, 828, 664, 539}], 
      Line3DBox[{560, 674, 842, 559, 762, 675, 843, 561, 676, 844, 562, 763, 
       677, 845, 563, 408, 564, 678, 846, 565, 764, 679, 847, 566, 680, 848, 
       567, 765, 681, 849, 568, 415, 569}]}, {
      Line3DBox[{122, 276, 123, 770, 290, 690, 150, 780, 304, 164, 786, 318, 
       710, 178, 800, 332, 192, 346, 206, 812, 360, 732, 220, 822, 374, 234, 
       828, 388, 752, 248, 842, 402, 262}], 
      Line3DBox[{124, 277, 682, 126, 291, 691, 151, 305, 704, 165, 787, 319, 
       711, 179, 333, 720, 193, 347, 724, 207, 361, 733, 221, 375, 746, 235, 
       829, 389, 753, 249, 403, 762, 263}], 
      Line3DBox[{128, 279, 683, 129, 772, 293, 692, 153, 307, 167, 790, 321, 
       712, 181, 802, 335, 195, 349, 725, 209, 814, 363, 734, 223, 377, 237, 
       832, 391, 754, 251, 844, 405, 265}], 
      Line3DBox[{130, 280, 684, 132, 294, 693, 694, 154, 308, 705, 168, 322, 
       713, 182, 336, 721, 196, 350, 726, 210, 364, 735, 736, 224, 378, 747, 
       238, 392, 755, 252, 406, 763, 266}], 
      Line3DBox[{134, 282, 685, 135, 774, 296, 696, 156, 310, 706, 170, 792, 
       324, 714, 184, 338, 198, 352, 727, 212, 816, 366, 738, 226, 380, 748, 
       240, 834, 394, 756, 254, 408, 268}], 
      Line3DBox[{6, 421, 17, 437, 438, 28, 454, 39, 470, 471, 50, 487, 61, 
       498, 72, 514, 515, 83, 531, 94, 547, 548, 105, 564, 116}], 
      Line3DBox[{136, 283, 137, 775, 297, 697, 157, 783, 311, 171, 793, 325, 
       715, 185, 804, 339, 199, 353, 213, 817, 367, 739, 227, 825, 381, 241, 
       835, 395, 757, 255, 846, 409, 269}], 
      Line3DBox[{138, 284, 686, 140, 298, 698, 158, 312, 707, 172, 794, 326, 
       716, 186, 340, 722, 200, 354, 728, 214, 368, 740, 228, 382, 749, 242, 
       836, 396, 758, 256, 410, 764, 270}], 
      Line3DBox[{142, 286, 687, 143, 777, 300, 699, 160, 314, 174, 797, 328, 
       717, 188, 806, 342, 202, 356, 729, 216, 819, 370, 741, 230, 384, 244, 
       839, 398, 759, 258, 848, 412, 272}], 
      Line3DBox[{144, 287, 688, 146, 301, 700, 701, 161, 315, 708, 175, 329, 
       718, 189, 343, 723, 203, 357, 730, 217, 371, 742, 743, 231, 385, 750, 
       245, 399, 760, 259, 413, 765, 273}], 
      Line3DBox[{148, 289, 689, 149, 779, 303, 703, 163, 317, 709, 177, 799, 
       331, 719, 191, 345, 205, 359, 731, 219, 821, 373, 745, 233, 387, 751, 
       247, 841, 401, 761, 261, 415, 275}], 
      Line3DBox[{264, 404, 843, 250, 390, 831, 830, 236, 376, 823, 222, 362, 
       813, 208, 348, 808, 194, 334, 801, 180, 320, 789, 788, 166, 306, 781, 
       152, 292, 771, 127, 278, 766, 125}], 
      Line3DBox[{267, 407, 845, 253, 393, 833, 239, 379, 824, 225, 737, 365, 
       815, 211, 351, 809, 197, 337, 803, 183, 323, 791, 169, 309, 782, 155, 
       695, 295, 773, 133, 281, 767, 131}], 
      Line3DBox[{271, 411, 847, 257, 397, 838, 837, 243, 383, 826, 229, 369, 
       818, 215, 355, 810, 201, 341, 805, 187, 327, 796, 795, 173, 313, 784, 
       159, 299, 776, 141, 285, 768, 139}], 
      Line3DBox[{274, 414, 849, 260, 400, 840, 246, 386, 827, 232, 744, 372, 
       820, 218, 358, 811, 204, 344, 807, 190, 330, 798, 176, 316, 785, 162, 
       702, 302, 778, 147, 288, 769, 145}]}, {}, {}}},
   VertexNormals->CompressedData["
1:eJyUvAkw1uH3969USlFUoogiKVlaVFJOCBXaVChJRCQRESlRqSShsitkCyXZ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    "]],
  Axes->True,
  BoxRatios->{1, 1, 0.4},
  Method->{"RotationControl" -> "Globe"},
  PlotRange->{{1., 11.}, {1., 11.}, {-0.99994, 0.999578}},
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.508662369555729*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["The generated code", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<MySurface.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"MySurface.h\"

#include \"MySurface.icc\"

#include <math.h>
void MySurface_TMySurfaceInit ()
{
; 
}

double MySurface_Tsinxy ( const double &x, const double &y)
{
    double r;
    double xx;
    double yy;
    r = sqrt((x*x)+(y*y));
    xx = r*MySurface_Tcos (MySurface_TarcTan (x, y));
    yy = r*MySurface_Tsin (MySurface_TarcTan (x, y));
    return MySurface_Tsin (xx+yy);
}

double MySurface_TarcTan ( const double &x, const double &y)
{
    int mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    int mc_T5;
    double mc_T6;
    double mc_T7;
    int mc_T8;
    double mc_T9;
    double mc_T10;
    mc_T1 = x < 0;
    if (mc_T1)
    {
        mc_T2 = 3.14159265358979323846;
    }
    else
    {
        mc_T2 = 0;
    }
    mc_T3 = y/x;
    mc_T4 = atan(mc_T3);
    mc_T5 = sign (y);
    mc_T6 = 3.14159265358979323846*mc_T5;
    mc_T7 = mc_T6/2;
    mc_T8 = x == 0;
    if (mc_T8)
    {
        mc_T9 = mc_T7;
    }
    else
    {
        mc_T9 = mc_T4;
    }
    mc_T10 = mc_T9+mc_T2;
    return mc_T10;
}

double MySurface_Tcos ( const double &x)
{
    double mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    double mc_T5;
    double mc_T6;
    double mc_T7;
    double mc_T8;
    double mc_T9;
    double mc_T10;
    double mc_T11;
    double mc_T12;
    double mc_T13;
    double mc_T14;
    mc_T1 = pow(x, 10);
    mc_T2 =    -2.755731922398589e-007;
    mc_T3 = mc_T2*mc_T1;
    mc_T4 = pow(x, 8);
    mc_T5 = mc_T4/40320;
    mc_T6 = pow(x, 6);
    mc_T7 =      -0.001388888888888889;
    mc_T8 = mc_T7*mc_T6;
    mc_T9 = (x*x*x*x);
    mc_T10 = mc_T9/24;
    mc_T11 = (x*x);
    mc_T12 =                       -0.5;
    mc_T13 = mc_T12*mc_T11;
    mc_T14 = 1+mc_T13+mc_T10+mc_T8+mc_T5+mc_T3;
    return mc_T14;
}

double MySurface_Tsin ( const double &x)
{
    double mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    double mc_T5;
    double mc_T6;
    double mc_T7;
    double mc_T8;
    double mc_T9;
    double mc_T10;
    double mc_T11;
    mc_T1 = pow(x, 9);
    mc_T2 = mc_T1/362880;
    mc_T3 = pow(x, 7);
    mc_T4 =     -0.0001984126984126984;
    mc_T5 = mc_T4*mc_T3;
    mc_T6 = pow(x, 5);
    mc_T7 = mc_T6/120;
    mc_T8 = (x*x*x);
    mc_T9 =        -0.1666666666666667;
    mc_T10 = mc_T9*mc_T8;
    mc_T11 = x+mc_T10+mc_T7+mc_T5+mc_T2;
    return mc_T11;
}

doubleNN MySurface_TxyMatrix(21, 21);\
\>", "Print",
 CellChangeTimes->{3.508662369618228*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Macro rules", "Subsection"],

Cell[TextData[{
 "Instead of using call-backs to ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " to evaluate functions not in the compilable subset so called macro rules \
can be used to replace these functions with compilable expressions. We \
illustrate this by substituting the FresbelS[ ] function above with a \
polynomial."
}], "Text"],

Cell["Generate some data", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"data", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"FresnelS", "[", "x", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"-", "2"}], ",", "2", ",", ".1"}], "}"}]}], "]"}]}], 
  ";"}]], "Input"],

Cell["\<\
Define the function ownFresnelS[ ] whose right hand side is a polynomial fit \
of 10th degree to the data with \"small\" terms removed\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ownFresnelS", "[", "x_", "]"}], "=", 
  RowBox[{
   RowBox[{"Fit", "[", 
    RowBox[{"data", ",", 
     SuperscriptBox["x", 
      RowBox[{"Range", "[", 
       RowBox[{"0", ",", "10"}], "]"}]], ",", "x"}], "]"}], "//", 
   "Chop"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", "0.036648265519705764`"}], " ", "x"}], "+", 
  RowBox[{"0.6998662074506663`", " ", 
   SuperscriptBox["x", "3"]}], "-", 
  RowBox[{"0.21386224373707644`", " ", 
   SuperscriptBox["x", "5"]}], "-", 
  RowBox[{"0.014640164493687376`", " ", 
   SuperscriptBox["x", "7"]}], "+", 
  RowBox[{"0.0069172028964939935`", " ", 
   SuperscriptBox["x", "9"]}]}]], "Output",
 CellChangeTimes->{3.508662369680727*^9}]
}, Open  ]],

Cell["We check the result", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"FresnelS", "[", "x", "]"}], ",", 
     RowBox[{"ownFresnelS", "[", "x", "]"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "2.5"}], ",", "2.5"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwtnHk0ld/3xy9XoyEhU4WiyVAppTLsW0oplSgkJUMKRZQyRCpJRaaQIkMy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     "]]}, 
   {Hue[0.9060679774997897, 0.6, 0.6], LineBox[CompressedData["
1:eJwVV3c81e8Xv0QD9RUltKwWSaVI6NySolQ2SZnZIUUIUZJKIYVKEmWljJCV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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  PlotRange->{{-2.5, 2.5}, {-2.2301481132346543`, 2.225524784732937}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.508662370180721*^9}]
}, Open  ]],

Cell["Generate C++ code with ", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CompilePackage", "[", 
   RowBox[{"\"\<MySurface`\>\"", ",", "\n", "\t", 
    RowBox[{"CallBackFunctions", "\[Rule]", 
     RowBox[{"{", "}"}]}], ",", "\n", "\t", 
    RowBox[{"MacroRules", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"FresnelS", "[", "x_", "]"}], "\[Rule]", 
       RowBox[{"ownFresnelS", "[", "x", "]"}]}], "}"}]}]}], "]"}], 
  ";"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 5 function(s)\"\>"], "Print",
 CellChangeTimes->{3.508662372836937*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["The generated code", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<MySurface.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"MySurface.h\"

#include \"MySurface.icc\"

#include <math.h>
doubleNN MySurface_TcomputeGridValues ( const int &n)
{
    int i;
    int j;
    int k;
    int j_D16752_D16755;
    int i_D16751_D16757;
    k = 1;
    while (k <= n)
    {
        k = k+1;
        MySurface_TxyMatrix.SetShape(Max(21,0), Max(21,0) );
        i_D16751_D16757 = 1;
        while (i_D16751_D16757 <= 21)
        {
            j_D16752_D16755 = 1;
            while (j_D16752_D16755 <= 21)
            {
                MySurface_TxyMatrix(1+i_D16751_D16757+-1, 1+j_D16752_D16755+-1)
                     = MySurface_Tsinxy (-2.+0.2*i_D16751_D16757, -2.+0.2*
                    j_D16752_D16755);
                j_D16752_D16755 = j_D16752_D16755+1;
            }
            i_D16751_D16757 = i_D16751_D16757+1;
        }
    }
    return (        0.4416327365966907)*MySurface_TxyMatrix;
}

void MySurface_TMySurfaceInit ()
{
; 
}

double MySurface_Tsinxy ( const double &x, const double &y)
{
    double r;
    double xx;
    double yy;
    r = sqrt((x*x)+(y*y));
    xx = r*MySurface_Tcos (MySurface_TarcTan (x, y));
    yy = r*MySurface_Tsin (MySurface_TarcTan (x, y));
    return MySurface_Tsin (xx+yy);
}

double MySurface_TarcTan ( const double &x, const double &y)
{
    int mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    int mc_T5;
    double mc_T6;
    double mc_T7;
    int mc_T8;
    double mc_T9;
    double mc_T10;
    mc_T1 = x < 0;
    if (mc_T1)
    {
        mc_T2 = 3.14159265358979323846;
    }
    else
    {
        mc_T2 = 0;
    }
    mc_T3 = y/x;
    mc_T4 = atan(mc_T3);
    mc_T5 = sign (y);
    mc_T6 = 3.14159265358979323846*mc_T5;
    mc_T7 = mc_T6/2;
    mc_T8 = x == 0;
    if (mc_T8)
    {
        mc_T9 = mc_T7;
    }
    else
    {
        mc_T9 = mc_T4;
    }
    mc_T10 = mc_T9+mc_T2;
    return mc_T10;
}

double MySurface_Tcos ( const double &x)
{
    double mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    double mc_T5;
    double mc_T6;
    double mc_T7;
    double mc_T8;
    double mc_T9;
    double mc_T10;
    double mc_T11;
    double mc_T12;
    double mc_T13;
    double mc_T14;
    mc_T1 = pow(x, 10);
    mc_T2 =    -2.755731922398589e-007;
    mc_T3 = mc_T2*mc_T1;
    mc_T4 = pow(x, 8);
    mc_T5 = mc_T4/40320;
    mc_T6 = pow(x, 6);
    mc_T7 =      -0.001388888888888889;
    mc_T8 = mc_T7*mc_T6;
    mc_T9 = (x*x*x*x);
    mc_T10 = mc_T9/24;
    mc_T11 = (x*x);
    mc_T12 =                       -0.5;
    mc_T13 = mc_T12*mc_T11;
    mc_T14 = 1+mc_T13+mc_T10+mc_T8+mc_T5+mc_T3;
    return mc_T14;
}

double MySurface_Tsin ( const double &x)
{
    double mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    double mc_T5;
    double mc_T6;
    double mc_T7;
    double mc_T8;
    double mc_T9;
    double mc_T10;
    double mc_T11;
    mc_T1 = pow(x, 9);
    mc_T2 = mc_T1/362880;
    mc_T3 = pow(x, 7);
    mc_T4 =     -0.0001984126984126984;
    mc_T5 = mc_T4*mc_T3;
    mc_T6 = pow(x, 5);
    mc_T7 = mc_T6/120;
    mc_T8 = (x*x*x);
    mc_T9 =        -0.1666666666666667;
    mc_T10 = mc_T9*mc_T8;
    mc_T11 = x+mc_T10+mc_T7+mc_T5+mc_T2;
    return mc_T11;
}

doubleNN MySurface_TxyMatrix(21, 21);\
\>", "Print",
 CellChangeTimes->{3.508662372883811*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean-up", "Section"],

Cell["\<\
Uninstall the executable, remove the temporary files, and remove the \
functions in the MySurface` package.\
\>", "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"CleanMathCodeFiles", "[", 
   RowBox[{
    RowBox[{"Confirm", "\[Rule]", "False"}], ",", 
    RowBox[{"CleanAllBut", "\[Rule]", 
     RowBox[{"{", "}"}]}]}], "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"Off", "[", 
   RowBox[{"Remove", "::", "rmnsm"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Remove", "[", "\"\<MySurface`*\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"On", "[", 
   RowBox[{"Remove", "::", "rmnsm"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Off", "[", 
   RowBox[{"DeleteFile", "::", "nffil"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"DeleteFile", "[", "\"\<res.txt\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"DeleteFile", "[", "\"\<result.txt\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"On", "[", 
   RowBox[{"DeleteFile", "::", "nffil"}], "]"}], ";"}]}], "Input"],

Cell["Delete the temporary directory and all files it contains.", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Conclusions", "Section"],

Cell[TextData[{
 "Simple usage: the most basic usage of ",
 StyleBox["MathCode C++ ",
  FontSlant->"Italic"],
 "can be summarized by three steps\n\t1)  Add type declarations\n\t2)  \
BuildCode[ ] generates C++ code and an executable program\n\t3)  InstallCode[ \
] connects the executable program to ",
 StyleBox["Mathematica",
  FontSlant->"Italic"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " provides an easy way of calling external programs without knowledge about ",
 StyleBox["MathLink.",
  FontSlant->"Italic"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "The combination of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " and",
 StyleBox[" MathCode C++",
  FontSlant->"Italic"],
 " is an excellent tool for prototyping of numerical algorithms."
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Speeds up large numerical computations that can be performed by compiled \
code outside ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " is an application package for ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " that generates optimized C++ code for ",
 StyleBox["numerical computations",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 StyleBox[".",
  FontSlant->"Italic"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell["Boost in performance for large numerical computations.", "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Transparent calling of compiled code from within ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " "
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " independent stand-alone code and programs can be generated "
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "Keep track of the compilable subset of ",
 StyleBox["Mathematica\n- Procedural programming constructs ",
  FontSlant->"Italic"],
 "(For, While, Do, If, ...)",
 StyleBox["\n- Elementary functions ",
  FontSlant->"Italic"],
 "(Sin, Exp, Log, ...)",
 StyleBox["\n- Arithmetic and logic expressions ",
  FontSlant->"Italic"],
 "(+,-,*,/,||,<,!=,...) \n- Map, Apply, List, Table, Array, Module, Sum, \
Product, Dot, Max, Transpose, DiagonalMatrix, ..."
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
 "How to use functions which does not belong to the compilable subset?\n",
 StyleBox["- Symbolic evaluation\n- Call-backs to Mathematica\n- Use external \
code\n- Reimplement the function using functions in the compilable subset",
  FontSlant->"Italic"]
}], "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}],

Cell["\<\
Simple usage: Type declarations, BuildCode[ ], InstallCode[ ]\
\>", "Text",
 CellDingbat->"\[FilledSmallCircle]",
 CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
}, Closed]]
},
AutoGeneratedPackage->Automatic,
ScreenStyleEnvironment->"Working",
WindowToolbars->{},
WindowSize->{924, 887},
WindowMargins->{{27, Automatic}, {Automatic, 5}},
ShowSelection->True,
ShowCellLabel->False,
CellLabelAutoDelete->True,
Magnification->1,
FrontEndVersion->"6.0 for Microsoft Windows (32-bit) (April 20, 2007)",
StyleDefinitions->Notebook[{
   Cell[
    CellGroupData[{
      Cell["Style Definitions", "Subtitle"], 
      Cell[
      "Modify the definitions below to change the default appearance of all \
cells in a given style.  Make modifications to any definition using commands \
in the Format menu.", "Text"], 
      Cell[
       CellGroupData[{
         Cell["Style Environment Names", "Section"], 
         Cell[
          StyleData[All, "Working"], PageWidth -> WindowWidth, ScriptMinSize -> 
          9], 
         Cell[
          StyleData[All, "Presentation"], PageWidth -> WindowWidth, 
          ScriptMinSize -> 12, FontSize -> 16], 
         Cell[
          StyleData[All, "Condensed"], PageWidth -> WindowWidth, 
          CellBracketOptions -> {"Margins" -> {1, 1}, "Widths" -> {0, 5}}, 
          ScriptMinSize -> 8, FontSize -> 11], 
         Cell[
          StyleData[All, "Printout"], PageWidth -> PaperWidth, ScriptMinSize -> 
          7, FontSize -> 10, 
          PrivateFontOptions -> {"FontType" -> "Outline"}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Notebook Options", "Section"], 
         Cell[
         "The options defined for the style below will be used at the \
Notebook level.", "Text"], 
         Cell[
          StyleData["Notebook"], PageHeaders -> {{
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"], None, 
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"]}, {
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"], None, 
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"]}}, 
          PageHeaderLines -> {True, True}, 
          PrintingOptions -> {
           "FacingPages" -> True, "FirstPageHeader" -> False}, 
          CellFrameLabelMargins -> 6, StyleMenuListing -> None]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headings", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 0}, CounterIncrements -> "Title", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subtitle", 0}, {"Subsubtitle", 0}}, FontSize -> 
             34, FontColor -> GrayLevel[1], Background -> RGBColor[0, 0, 1]], 
            
            Cell[
             StyleData["Title", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 0}, FontSize -> 
             44], 
            Cell[
             StyleData["Title", "Condensed"], CellMargins -> {{0, 0}, {0, 0}},
              FontSize -> 20], 
            Cell[
             StyleData["Title", "Printout"], CellMargins -> {{0, 0}, {0, 0}}, 
             FontSize -> 24, FontTracking -> "Plain", Background -> 
             GrayLevel[0]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 3}, ParagraphIndent -> -96, CounterIncrements -> 
             "Subtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subsubtitle", 0}}, FontFamily -> "Helvetica", 
             FontSize -> 18, FontColor -> GrayLevel[1], Background -> 
             RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Subtitle", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 6}, 
             ParagraphIndent -> -157, FontSize -> 30], 
            Cell[
             StyleData["Subtitle", "Condensed"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -78, 
             FontSize -> 14], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -85, 
             FontSize -> 16, Background -> GrayLevel[0.6]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {30, 10}}, PageBreakBelow -> False, 
             CounterIncrements -> "Subsubtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}}, FontFamily -> "Helvetica", FontSize -> 14, 
             FontSlant -> "Italic"], 
            Cell[
             StyleData["Subsubtitle", "Presentation"], 
             CellMargins -> {{8, 10}, {40, 20}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Subsubtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 8}}, FontSize -> 12], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{9, 10}, {50, 10}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], CellFrame -> {{6, 0}, {0, 1}}, CellDingbat -> 
             None, CellMargins -> {{12, Inherited}, {4, 24}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameMargins -> 6, CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             FontFamily -> "Helvetica", FontSize -> 16, FontWeight -> "Bold", 
             FontColor -> RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Section", "Presentation"], 
             CellMargins -> {{10, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["Section", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Section", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, FontSize -> 14, FontTracking -> 
             "Plain", FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], 
             CellMargins -> {{12, Inherited}, {8, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsection", "Presentation"], 
             CellMargins -> {{11, 10}, {8, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 22], 
            Cell[
             StyleData["Subsection", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Subsection", "Printout"], 
             CellMargins -> {{9, 0}, {4, 40}}, FontSize -> 12]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> "\[FilledSquare]", 
             CellMargins -> {{25, Inherited}, {8, 12}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsubsection", "Presentation"], 
             CellMargins -> {{29, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Subsubsection", "Condensed"], 
             CellMargins -> {{22, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Subsubsection", "Printout"], 
             CellMargins -> {{21, 0}, {4, 20}}, FontSize -> 11]}, Open]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Body Text", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Text"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Text", FontFamily -> "Times"], 
            Cell[
             StyleData["Text", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Text", "Condensed"], CellMargins -> {{8, 10}, {4, 4}},
              LineSpacing -> {1, 1}, ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Text", "Printout"], CellMargins -> {{9, 0}, {4, 4}}, 
             ParagraphSpacing -> {0, 6}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SmallText"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 6}, 
             CounterIncrements -> "SmallText", FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["SmallText", "Condensed"], 
             CellMargins -> {{8, 10}, {2, 2}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Printout"], 
             CellMargins -> {{9, 0}, {4, 4}}, FontSize -> 7]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Input/Output", "Section"], 
         Cell[
         "The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Input"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             CounterIncrements -> "Input", FontWeight -> "Bold", Background -> 
             RGBColor[1, 1, 0]], 
            Cell[
             StyleData["Input", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Input", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["Input", "Printout"], CellMargins -> {{44, 0}, {6, 6}},
              Background -> GrayLevel[0.8]]}, Closed]], 
         Cell[
          StyleData["InlineInput"], Evaluatable -> True, CellGroupingRules -> 
          "InputGrouping", CellHorizontalScrolling -> True, PageBreakWithin -> 
          False, GroupPageBreakWithin -> False, DefaultFormatType -> 
          DefaultInputFormatType, AutoItalicWords -> {}, FormatType -> 
          InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
          CounterIncrements -> "Input", FontWeight -> "Bold"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Output"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, CellEditDuplicate -> True, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, 
             CellLabelMargins -> {{3, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Output", Background -> 
             RGBColor[1, 0.900008, 0.900008]], 
            Cell[
             StyleData["Output", "Presentation"], 
             CellMargins -> {{62, Inherited}, {12, 5}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Output", "Condensed"], 
             CellMargins -> {{40, Inherited}, {4, 1}}], 
            Cell[
             StyleData["Output", "Printout"], 
             CellMargins -> {{44, 0}, {6, 2}}, Background -> 
             GrayLevel[0.900008]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Message"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Message", StyleMenuListing -> 
             None, FontColor -> RGBColor[1, 0, 0]], 
            Cell[
             StyleData["Message", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Message", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Message", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Print"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, ShowCellLabel -> 
             False, DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Print", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Print", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Print", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Print", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Graphics"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "GraphicsGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Graphics", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Graphics", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Graphics", "Condensed"], 
             CellMargins -> {{52, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {175, 175}], 
            Cell[
             StyleData["Graphics", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {250, 250}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["CellLabel"], StyleMenuListing -> None, FontFamily -> 
             "Helvetica", FontSize -> 11, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["CellLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["CellLabel", "Condensed"], FontSize -> 8], 
            Cell[
             StyleData["CellLabel", "Printout"], FontSize -> 8, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Unique Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {2, 10}}, FontSize -> 16, FontSlant -> 
             "Italic"], 
            Cell[
             StyleData["Author", "Presentation"], 
             CellMargins -> {{12, 10}, {2, 12}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}, FontSize -> 20], 
            Cell[
             StyleData["Author", "Condensed"], 
             CellMargins -> {{8, 10}, {1, 4}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}, FontSize -> 12], 
            Cell[
             StyleData["Author", "Printout"], 
             CellMargins -> {{9, 0}, {4, 12}}, ParagraphSpacing -> {0, 6}, 
             FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Copyright"], ShowCellBracket -> False, 
             CellMargins -> {{10, 10}, {40, 2}}, FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Presentation"], 
             CellMargins -> {{12, 10}, {50, 2}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["Copyright", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 1}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Printout"], 
             CellMargins -> {{9, 0}, {72, 4}}, FontSize -> 7]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Example"], CellMargins -> {{12, 10}, {5, 12}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Example", FontFamily -> "Times", 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["Example", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 20}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Example", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 8}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Example", "Printout"], 
             CellMargins -> {{9, 0}, {4, 10}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Definition"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Definition", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Definition", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Definition", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Definition", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.6]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Theorem"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Theorem", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Theorem", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Theorem", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Theorem", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.4]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseMain"], CellFrame -> {{6, 0}, {0, 1}}, 
             CellMargins -> {{12, Inherited}, {4, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameColor -> RGBColor[0.571389, 0.19675, 0.570504], 
             CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 16, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["ExerciseMain", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["ExerciseMain", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["ExerciseMain", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, CellFrameColor -> 
             GrayLevel[0.500008], FontSize -> 14, FontTracking -> "Plain", 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Exercise"], CellDingbat -> "\[FilledDownTriangle]", 
             CellMargins -> {{23, Inherited}, {4, 18}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Exercise", "Presentation"], 
             CellMargins -> {{33, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Exercise", "Condensed"], 
             CellMargins -> {{17, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Exercise", "Printout"], 
             CellFrame -> {{0, 0}, {0.5, 0}}, CellDingbat -> None, 
             CellMargins -> {{9, 0}, {6, 20}}, FontSize -> 11, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseText"], CellMargins -> {{62, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 8}, 
             CounterIncrements -> "ExcersiceText", FontFamily -> "Times"], 
            Cell[
             StyleData["ExerciseText", "Presentation"], 
             CellMargins -> {{74, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["ExerciseText", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 2}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["ExerciseText", "Printout"], 
             CellMargins -> {{54, 0}, {4, 4}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Itemize"], CellDingbat -> "\[FilledSmallCircle]", 
             CellMargins -> {{65, Inherited}, {8, 8}}, FontFamily -> "Times", 
             FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Itemize", "Presentation"], 
             CellMargins -> {{Inherited, 10}, {8, 12}}, LineSpacing -> {1, 0},
              FontSize -> 18], 
            Cell[
             StyleData["Itemize", "Condensed"], 
             CellMargins -> {{Inherited, Inherited}, {2, 12}}, FontSize -> 
             10], 
            Cell[
             StyleData["Itemize", "Printout"], 
             CellMargins -> {{Inherited, 0}, {4, 20}}, FontSize -> 11]}, 
           Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayCode"], CellFrame -> 2, 
             CellMargins -> {{52, 10}, {8, 8}}, CellHorizontalScrolling -> 
             True, PageBreakWithin -> True, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             FontWeight -> "Bold", Background -> RGBColor[1, 1, 0]], 
            Cell[
             StyleData["DisplayCode", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["DisplayCode", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["DisplayCode", "Printout"], 
             CellMargins -> {{44, 0}, {6, 6}}, Background -> GrayLevel[0.8]]},
            Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Formulas and Programming", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayFormula"], CellMargins -> {{62, 10}, {2, 10}}, 
             CellHorizontalScrolling -> True, ScriptLevel -> 0, 
             SingleLetterItalics -> True, 
             UnderoverscriptBoxOptions -> {LimitsPositioning -> True}], 
            Cell[
             StyleData["DisplayFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 18], 
            Cell[
             StyleData["DisplayFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["DisplayFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ChemicalFormula"], CellMargins -> {{62, 10}, {2, 10}},
              AutoSpacing -> False, ScriptBaselineShifts -> {0.6, Automatic}, 
             ScriptLevel -> 1, SingleLetterItalics -> False, ZeroWidthTimes -> 
             True], 
            Cell[
             StyleData["ChemicalFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Program"], CellMargins -> {{12, 10}, {Inherited, 6}}, 
             FontFamily -> "Courier"], 
            Cell[
             StyleData["Program", "Presentation"], 
             CellMargins -> {{13, 30}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Condensed"], 
             CellMargins -> {{8, 10}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Printout"], 
             CellMargins -> {{9, 0}, {Inherited, 4}}, FontSize -> 9.5]}, 
           Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Automatic Numbering", "Section"], 
         Cell[
         "The following styles are useful for numbered equations, figures, \
etc.  They automatically give the cell a FrameLabel containing a reference to \
a particular counter, and also increment that counter.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedEquation"], 
             CellMargins -> {{62, 10}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, 
                Cell[
                 TextData[{"(", 
                   CounterBox["NumberedEquation"], ")"}]]}, {None, None}}, 
             DefaultFormatType -> DefaultInputFormatType, CounterIncrements -> 
             "NumberedEquation", FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedEquation", "Presentation"], 
             CellMargins -> {{74, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Condensed"], 
             CellMargins -> {{52, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Printout"], 
             CellMargins -> {{54, 0}, {Inherited, Inherited}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedFigure"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Figure ", 
                   CounterBox["NumberedFigure"]}]], None}}, CounterIncrements -> 
             "NumberedFigure", 
             ImageMargins -> {{43, Inherited}, {Inherited, 0}}, 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedFigure", "Presentation"]], 
            Cell[
             StyleData["NumberedFigure", "Condensed"]], 
            Cell[
             StyleData["NumberedFigure", "Printout"]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedTable"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Table ", 
                   CounterBox["NumberedTable"]}]], None}}, TextAlignment -> 
             Center, CounterIncrements -> "NumberedTable", 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedTable", "Presentation"]], 
            Cell[
             StyleData["NumberedTable", "Condensed"]], 
            Cell[
             StyleData["NumberedTable", "Printout"]]}, Open]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headers and Footers", "Section"], 
         Cell[
          StyleData["Header"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["Footer"], CellMargins -> {{0, 0}, {0, 4}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 6],
          
         Cell[
          StyleData["PageNumber"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontWeight -> "Bold"]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Hyperlink Styles", "Section"], 
         Cell[
         "The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Hyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonNote -> 
               ButtonData}], 
            Cell[
             StyleData["Hyperlink", "Presentation"]], 
            Cell[
             StyleData["Hyperlink", "Condensed"]], 
            Cell[
             StyleData["Hyperlink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
         "The following styles are for linking automatically to the on-line \
help system.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MainBookLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MainBook", #]}]& )}], 
            Cell[
             StyleData["MainBookLink", "Presentation"]], 
            Cell[
             StyleData["MainBookLink", "Condensed"]], 
            Cell[
             StyleData["MainBookLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["AddOnsLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["AddOns", #]}]& )}], 
            Cell[
             StyleData["AddOnsLink", "Presentation"]], 
            Cell[
             StyleData["AddOnsLink", "Condensed"]], 
            Cell[
             StyleData["AddOnLink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuideLink", #]}]& )}], 
            Cell[
             StyleData["RefGuideLink", "Presentation"]], 
            Cell[
             StyleData["RefGuideLink", "Condensed"]], 
            Cell[
             StyleData["RefGuideLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["GettingStartedLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["GettingStarted", #]}]& )}], 
            Cell[
             StyleData["GettingStartedLink", "Presentation"]], 
            Cell[
             StyleData["GettingStartedLink", "Condensed"]], 
            Cell[
             StyleData["GettingStartedLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["OtherInformationLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["OtherInformation", #]}]& )}], 
            Cell[
             StyleData["OtherInformationLink", "Presentation"]], 
            Cell[
             StyleData["OtherInformationLink", "Condensed"]], 
            Cell[
             StyleData["OtherInformationLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]]}, Closed]]},
      Open]]}, Visible -> False, FrontEndVersion -> 
  "6.0 for Microsoft Windows (32-bit) (April 20, 2007)", StyleDefinitions -> 
  "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "Info3508665915-4434317"->{
  Cell[37037, 1066, 120, 2, 42, "Print",
   CellTags->"Info3508665915-4434317"],
  Cell[37160, 1070, 1028, 29, 27, "Print",
   CellTags->"Info3508665915-4434317"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"Info3508665915-4434317", 298523, 7204}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[580, 21, 92, 2, 55, "Title"],
Cell[675, 25, 50, 0, 37, "Subtitle"],
Cell[728, 27, 85, 2, 54, "Copyright"],
Cell[CellGroupData[{
Cell[838, 33, 27, 0, 54, "Section"],
Cell[868, 35, 805, 23, 48, "Text"]
}, Closed]],
Cell[CellGroupData[{
Cell[1710, 63, 98, 5, 34, "Section"],
Cell[1811, 70, 421, 14, 28, "Text"],
Cell[2235, 86, 223, 6, 28, "Text"],
Cell[2461, 94, 277, 8, 28, "Text"],
Cell[2741, 104, 182, 4, 28, "Text"],
Cell[2926, 110, 279, 8, 28, "Text"],
Cell[3208, 120, 170, 2, 28, "Text"],
Cell[3381, 124, 410, 13, 44, "Text"]
}, Closed]],
Cell[CellGroupData[{
Cell[3828, 142, 111, 5, 34, "Section"],
Cell[3942, 149, 460, 14, 28, "Text"],
Cell[4405, 165, 343, 10, 28, "Text"],
Cell[4751, 177, 426, 14, 28, "Text"],
Cell[5180, 193, 401, 11, 28, "Text"],
Cell[5584, 206, 342, 11, 28, "Text"]
}, Closed]],
Cell[CellGroupData[{
Cell[5963, 222, 30, 0, 34, "Section"],
Cell[5996, 224, 32, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[6053, 228, 74, 1, 48, "Input"],
Cell[6130, 231, 448, 10, 23, "Print"]
}, Open  ]],
Cell[6593, 244, 222, 6, 25, "Text"],
Cell[6818, 252, 130, 1, 48, "Input"],
Cell[6951, 255, 120, 3, 25, "Text"],
Cell[7074, 260, 390, 14, 63, "Input"],
Cell[7467, 276, 48, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[7540, 280, 197, 6, 48, "Input"],
Cell[7740, 288, 8583, 148, 255, "Output"]
}, Open  ]],
Cell[16338, 439, 70, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[16433, 443, 57, 1, 48, "Input"],
Cell[16493, 446, 122, 1, 23, "Print"]
}, Open  ]],
Cell[16630, 450, 127, 5, 25, "Text"],
Cell[CellGroupData[{
Cell[16782, 459, 77, 2, 48, "Input"],
Cell[16862, 463, 228, 5, 23, "Print"]
}, Open  ]],
Cell[17105, 471, 72, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[17202, 475, 197, 6, 48, "Input"],
Cell[17402, 483, 8577, 148, 255, "Output"]
}, Open  ]],
Cell[25994, 634, 56, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[26075, 638, 78, 1, 48, "Input"],
Cell[26156, 641, 239, 17, 247, "Print"]
}, Open  ]],
Cell[26410, 661, 174, 6, 26, "Text"],
Cell[26587, 669, 342, 10, 89, "Input"],
Cell[26932, 681, 105, 3, 25, "Text"],
Cell[27040, 686, 168, 2, 25, "Text"],
Cell[27211, 690, 224, 7, 26, "Text"],
Cell[27438, 699, 235, 7, 26, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[27710, 711, 36, 0, 54, "Section"],
Cell[27749, 713, 229, 6, 25, "Text"],
Cell[27981, 721, 115, 3, 48, "Input",
 Evaluatable->False],
Cell[28099, 726, 70, 0, 25, "Text"],
Cell[28172, 728, 385, 11, 48, "Input",
 Evaluatable->False],
Cell[28560, 741, 89, 2, 25, "Text"],
Cell[28652, 745, 147, 2, 25, "Text"],
Cell[28802, 749, 333, 12, 48, "Input",
 Evaluatable->False],
Cell[29138, 763, 133, 2, 25, "Text"],
Cell[29274, 767, 135, 4, 48, "Input",
 Evaluatable->False],
Cell[29412, 773, 417, 15, 48, "Input",
 Evaluatable->False],
Cell[29832, 790, 343, 12, 43, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[30212, 807, 40, 0, 54, "Section"],
Cell[30255, 809, 328, 14, 41, "Text"],
Cell[30586, 825, 317, 9, 43, "Text"],
Cell[30906, 836, 200, 5, 25, "Text"],
Cell[31109, 843, 261, 7, 43, "Text"],
Cell[31373, 852, 233, 5, 25, "Text"],
Cell[31609, 859, 380, 11, 87, "Text"],
Cell[31992, 872, 303, 9, 56, "Text"],
Cell[32298, 883, 314, 9, 56, "Text"],
Cell[32615, 894, 283, 9, 56, "Text"],
Cell[32901, 905, 269, 7, 56, "Text"],
Cell[33173, 914, 482, 14, 117, "Text"],
Cell[33658, 930, 322, 11, 56, "Text"],
Cell[33983, 943, 261, 9, 56, "Text"],
Cell[34247, 954, 280, 9, 56, "Text"],
Cell[34530, 965, 284, 9, 56, "Text"],
Cell[34817, 976, 401, 13, 25, "Text"],
Cell[35221, 991, 199, 6, 25, "Text"],
Cell[35423, 999, 91, 2, 25, "Text"],
Cell[35517, 1003, 138, 4, 26, "Text"],
Cell[35658, 1009, 208, 6, 25, "Text"],
Cell[CellGroupData[{
Cell[35891, 1019, 58, 0, 44, "Subsection"],
Cell[CellGroupData[{
Cell[35974, 1023, 62, 0, 35, "Subsubsection"],
Cell[36039, 1025, 338, 10, 43, "Text"],
Cell[36380, 1037, 380, 12, 48, "Input"],
Cell[CellGroupData[{
Cell[36785, 1053, 79, 2, 48, "Input"],
Cell[36867, 1057, 86, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[36990, 1063, 44, 1, 48, "Input"],
Cell[37037, 1066, 120, 2, 42, "Print",
 CellTags->"Info3508665915-4434317"],
Cell[37160, 1070, 1028, 29, 27, "Print",
 CellTags->"Info3508665915-4434317"]
}, Open  ]],
Cell[38203, 1102, 70, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[38298, 1106, 131, 3, 48, "Input"],
Cell[38432, 1111, 119, 1, 23, "Print"]
}, Open  ]],
Cell[38566, 1115, 303, 9, 43, "Text"],
Cell[CellGroupData[{
Cell[38894, 1128, 200, 6, 48, "Input"],
Cell[39097, 1136, 282, 8, 47, "Output"]
}, Open  ]],
Cell[39394, 1147, 127, 5, 25, "Text"],
Cell[CellGroupData[{
Cell[39546, 1156, 77, 2, 48, "Input"],
Cell[39626, 1160, 228, 5, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[39891, 1170, 79, 2, 48, "Input"],
Cell[39973, 1174, 90, 1, 47, "Output"]
}, Open  ]],
Cell[40078, 1178, 174, 6, 26, "Text"],
Cell[40255, 1186, 342, 10, 89, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[40634, 1201, 103, 4, 36, "Subsubsection"],
Cell[40740, 1207, 162, 3, 25, "Text"],
Cell[40905, 1212, 419, 14, 66, "Input"],
Cell[41327, 1228, 38, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[41390, 1232, 218, 6, 48, "Input"],
Cell[41611, 1240, 12126, 204, 225, "Output"]
}, Open  ]],
Cell[53752, 1447, 107, 3, 25, "Text"],
Cell[53862, 1452, 150, 4, 48, "Input"],
Cell[54015, 1458, 153, 4, 48, "Input"],
Cell[54171, 1464, 156, 5, 25, "Text"],
Cell[CellGroupData[{
Cell[54352, 1473, 155, 5, 64, "Input"],
Cell[54510, 1480, 626, 20, 85, "Output"]
}, Open  ]],
Cell[55151, 1503, 279, 9, 25, "Text"],
Cell[CellGroupData[{
Cell[55455, 1516, 134, 3, 48, "Input"],
Cell[55592, 1521, 122, 1, 23, "Print"]
}, Open  ]],
Cell[55729, 1525, 121, 4, 25, "Text"],
Cell[CellGroupData[{
Cell[55875, 1533, 77, 2, 48, "Input"],
Cell[55955, 1537, 226, 5, 23, "Print"]
}, Open  ]],
Cell[56196, 1545, 69, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[56290, 1549, 155, 5, 64, "Input"],
Cell[56448, 1556, 206, 6, 64, "Output"]
}, Open  ]],
Cell[56669, 1565, 233, 6, 25, "Text"],
Cell[56905, 1573, 46, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[56976, 1577, 218, 6, 48, "Input"],
Cell[57197, 1585, 12130, 204, 225, "Output"]
}, Open  ]],
Cell[69342, 1792, 174, 6, 26, "Text"],
Cell[69519, 1800, 302, 8, 89, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[69858, 1813, 43, 0, 35, "Subsubsection"],
Cell[69904, 1815, 429, 15, 25, "Text"],
Cell[70336, 1832, 608, 18, 89, "Input"],
Cell[70947, 1852, 99, 2, 25, "Text"],
Cell[71049, 1856, 848, 27, 69, "Input"],
Cell[71900, 1885, 140, 3, 25, "Text"],
Cell[CellGroupData[{
Cell[72065, 1892, 62, 1, 48, "Input"],
Cell[72130, 1895, 120, 1, 23, "Print"]
}, Open  ]],
Cell[72265, 1899, 244, 4, 43, "Text"],
Cell[72512, 1905, 44, 0, 25, "Text"],
Cell[72559, 1907, 200, 5, 48, "Input"],
Cell[72762, 1914, 135, 4, 25, "Text"],
Cell[CellGroupData[{
Cell[72922, 1922, 77, 2, 48, "Input"],
Cell[73002, 1926, 228, 5, 23, "Print"]
}, Open  ]],
Cell[73245, 1934, 32, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[73302, 1938, 160, 4, 48, "Input"],
Cell[73465, 1944, 1777, 34, 254, "Output"]
}, Open  ]],
Cell[75257, 1981, 69, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[75351, 1985, 493, 16, 50, "Input"],
Cell[75847, 2003, 1777, 34, 254, "Output"]
}, Open  ]],
Cell[77639, 2040, 72, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[77736, 2044, 344, 11, 89, "Input"],
Cell[78083, 2057, 115, 1, 47, "Output"]
}, Open  ]],
Cell[78213, 2061, 72, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[78310, 2065, 686, 23, 92, "Input"],
Cell[78999, 2090, 118, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[79154, 2096, 74, 1, 48, "Input"],
Cell[79231, 2099, 110, 1, 47, "Output"]
}, Open  ]],
Cell[79356, 2103, 260, 6, 43, "Text"],
Cell[79619, 2111, 102, 2, 25, "Text"],
Cell[79724, 2115, 1061, 30, 209, "Input"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[80846, 2152, 33, 0, 54, "Section"],
Cell[80882, 2154, 141, 5, 25, "Text"],
Cell[CellGroupData[{
Cell[81048, 2163, 56, 0, 35, "Subsubsection"],
Cell[81107, 2165, 518, 20, 32, "Text"],
Cell[81628, 2187, 2756, 95, 103, "Input"],
Cell[84387, 2284, 227, 4, 43, "Text"],
Cell[84617, 2290, 36, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[84678, 2294, 219, 7, 48, "Input"],
Cell[84900, 2303, 17988, 301, 267, "Output"]
}, Open  ]],
Cell[102903, 2607, 212, 6, 25, "Text"],
Cell[CellGroupData[{
Cell[103140, 2617, 131, 3, 48, "Input"],
Cell[103274, 2622, 120, 1, 23, "Print"]
}, Open  ]],
Cell[103409, 2626, 121, 4, 25, "Text"],
Cell[CellGroupData[{
Cell[103555, 2634, 77, 2, 48, "Input"],
Cell[103635, 2638, 228, 5, 23, "Print"]
}, Open  ]],
Cell[103878, 2646, 61, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[103964, 2650, 219, 7, 48, "Input"],
Cell[104186, 2659, 17990, 301, 267, "Output"]
}, Open  ]],
Cell[122191, 2963, 56, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[122272, 2967, 78, 1, 48, "Input"],
Cell[122353, 2970, 1416, 67, 1047, "Print"]
}, Open  ]],
Cell[123784, 3040, 249, 5, 43, "Text"],
Cell[124036, 3047, 174, 6, 26, "Text"],
Cell[124213, 3055, 342, 10, 89, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[124592, 3070, 41, 0, 35, "Subsubsection"],
Cell[124636, 3072, 388, 12, 43, "Text"],
Cell[125027, 3086, 197, 6, 50, "Input"],
Cell[125227, 3094, 401, 13, 43, "Text"],
Cell[CellGroupData[{
Cell[125653, 3111, 230, 5, 169, "Input"],
Cell[125886, 3118, 122, 1, 23, "Print"]
}, Open  ]],
Cell[126023, 3122, 55, 0, 25, "Text"],
Cell[126081, 3124, 114, 2, 48, "Input"],
Cell[126198, 3128, 90, 2, 25, "Text"],
Cell[CellGroupData[{
Cell[126313, 3134, 78, 1, 48, "Input"],
Cell[126394, 3137, 250, 18, 263, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[126681, 3160, 77, 1, 48, "Input"],
Cell[126761, 3163, 242, 14, 199, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[127040, 3182, 82, 1, 48, "Input"],
Cell[127125, 3185, 185, 10, 135, "Print"]
}, Open  ]],
Cell[127325, 3198, 103, 3, 25, "Text"],
Cell[127431, 3203, 109, 3, 48, "Input"],
Cell[127543, 3208, 41, 0, 25, "Text"],
Cell[127587, 3210, 79, 1, 48, "Input"],
Cell[127669, 3213, 174, 6, 26, "Text"],
Cell[127846, 3221, 342, 10, 89, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[128225, 3236, 51, 0, 35, "Subsubsection"],
Cell[128279, 3238, 191, 4, 25, "Text"],
Cell[128473, 3244, 65, 2, 25, "Text"],
Cell[CellGroupData[{
Cell[128563, 3250, 337, 11, 48, "Input"],
Cell[128903, 3263, 1923, 60, 89, "Output"]
}, Open  ]],
Cell[130841, 3326, 37, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[130903, 3330, 163, 4, 48, "Input"],
Cell[131069, 3336, 1591, 49, 74, "Output"]
}, Open  ]],
Cell[132675, 3388, 103, 3, 25, "Text"],
Cell[CellGroupData[{
Cell[132803, 3395, 178, 5, 48, "Input"],
Cell[132984, 3402, 1403, 44, 89, "Output"]
}, Open  ]],
Cell[134402, 3449, 126, 5, 25, "Text"],
Cell[134531, 3456, 377, 12, 48, "Input"],
Cell[CellGroupData[{
Cell[134933, 3472, 69, 1, 48, "Input"],
Cell[135005, 3475, 1683, 52, 89, "Output"]
}, Open  ]],
Cell[136703, 3530, 70, 0, 25, "Text"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[136822, 3536, 49, 0, 54, "Section"],
Cell[136874, 3538, 166, 3, 25, "Text"],
Cell[CellGroupData[{
Cell[137065, 3545, 33, 0, 44, "Subsection"],
Cell[CellGroupData[{
Cell[137123, 3549, 45, 0, 35, "Subsubsection"],
Cell[137171, 3551, 374, 11, 89, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[137582, 3567, 41, 0, 35, "Subsubsection"],
Cell[137626, 3569, 466, 16, 189, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[138129, 3590, 44, 0, 35, "Subsubsection"],
Cell[138176, 3592, 808, 16, 369, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[139021, 3613, 40, 0, 35, "Subsubsection"],
Cell[139064, 3615, 101, 2, 48, "Input"],
Cell[139168, 3619, 34, 0, 25, "Text"],
Cell[139205, 3621, 183, 6, 48, "Input"],
Cell[139391, 3629, 136, 4, 25, "Text"],
Cell[139530, 3635, 157, 4, 48, "Input"],
Cell[139690, 3641, 257, 6, 43, "Text"],
Cell[139950, 3649, 1303, 41, 89, "Input"],
Cell[141256, 3692, 33, 0, 25, "Text"],
Cell[141292, 3694, 1122, 34, 189, "Input"],
Cell[142417, 3730, 152, 3, 25, "Text"],
Cell[142572, 3735, 1281, 37, 109, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[143890, 3777, 47, 0, 35, "Subsubsection"],
Cell[143940, 3779, 69, 2, 48, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[144046, 3786, 39, 0, 35, "Subsubsection"],
Cell[144088, 3788, 95, 3, 48, "Input"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[144232, 3797, 68, 0, 44, "Subsection"],
Cell[144303, 3799, 99, 2, 25, "Text"],
Cell[CellGroupData[{
Cell[144427, 3805, 102, 2, 48, "Input"],
Cell[144532, 3809, 122, 1, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[144691, 3815, 43, 0, 35, "Subsubsection"],
Cell[CellGroupData[{
Cell[144759, 3819, 81, 1, 48, "Input"],
Cell[144843, 3822, 3277, 156, 2455, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[148181, 3985, 77, 0, 44, "Subsection"],
Cell[148261, 3987, 220, 4, 43, "Text"],
Cell[CellGroupData[{
Cell[148506, 3995, 361, 9, 89, "Input"],
Cell[148870, 4006, 120, 1, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[149027, 4012, 43, 0, 35, "Subsubsection"],
Cell[CellGroupData[{
Cell[149095, 4016, 81, 1, 48, "Input"],
Cell[149179, 4019, 3680, 163, 2567, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[152920, 4189, 47, 0, 44, "Subsection"],
Cell[152970, 4191, 261, 7, 43, "Text"],
Cell[CellGroupData[{
Cell[153256, 4202, 356, 9, 89, "Input"],
Cell[153615, 4213, 120, 1, 23, "Print"]
}, Open  ]],
Cell[153750, 4217, 165, 4, 25, "Text"],
Cell[CellGroupData[{
Cell[153940, 4225, 85, 1, 48, "Input"],
Cell[154028, 4228, 362, 15, 215, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[154427, 4248, 80, 1, 48, "Input"],
Cell[154510, 4251, 459, 21, 311, "Print"]
}, Open  ]],
Cell[154984, 4275, 58, 1, 48, "Input"],
Cell[155045, 4278, 89, 2, 25, "Text"],
Cell[155137, 4282, 214, 6, 48, "Input"],
Cell[155354, 4290, 39, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[155418, 4294, 76, 1, 48, "Input"],
Cell[155497, 4297, 1364, 24, 183, "Print"]
}, Open  ]],
Cell[156876, 4324, 40, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[156941, 4328, 195, 4, 48, "Input"],
Cell[157139, 4334, 50151, 825, 315, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[207327, 5164, 43, 0, 35, "Subsubsection"],
Cell[CellGroupData[{
Cell[207395, 5168, 81, 1, 48, "Input"],
Cell[207479, 5171, 2452, 123, 1943, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[209992, 5301, 33, 0, 44, "Subsection"],
Cell[210028, 5303, 348, 8, 43, "Text"],
Cell[210379, 5313, 34, 0, 25, "Text"],
Cell[210416, 5315, 321, 11, 48, "Input"],
Cell[210740, 5328, 158, 3, 25, "Text"],
Cell[CellGroupData[{
Cell[210923, 5335, 280, 9, 50, "Input"],
Cell[211206, 5346, 454, 12, 47, "Output"]
}, Open  ]],
Cell[211675, 5361, 35, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[211735, 5365, 292, 9, 48, "Input"],
Cell[212030, 5376, 40284, 667, 255, "Output"]
}, Open  ]],
Cell[252329, 6046, 39, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[252393, 6050, 414, 11, 89, "Input"],
Cell[252810, 6063, 120, 1, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[252967, 6069, 43, 0, 35, "Subsubsection"],
Cell[CellGroupData[{
Cell[253035, 6073, 81, 1, 48, "Input"],
Cell[253119, 6076, 3260, 152, 2407, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[256452, 6236, 27, 0, 54, "Section"],
Cell[256482, 6238, 131, 3, 25, "Text"],
Cell[256616, 6243, 943, 27, 189, "Input"],
Cell[257562, 6272, 73, 0, 25, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[257672, 6277, 30, 0, 54, "Section"],
Cell[257705, 6279, 457, 11, 118, "Text"],
Cell[258165, 6292, 298, 8, 28, "Text"],
Cell[258466, 6302, 321, 10, 28, "Text"],
Cell[258790, 6314, 269, 8, 28, "Text"],
Cell[259062, 6324, 421, 14, 28, "Text"],
Cell[259486, 6340, 165, 2, 28, "Text"],
Cell[259654, 6344, 230, 7, 28, "Text"],
Cell[259887, 6353, 235, 6, 28, "Text"],
Cell[260125, 6361, 572, 14, 148, "Text"],
Cell[260700, 6377, 376, 7, 148, "Text"],
Cell[261079, 6386, 180, 4, 28, "Text"]
}, Closed]]
}
]
*)

(* End of internal cache information *)
