(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     53568,       1427]*)
(*NotebookOutlinePosition[     83797,       2530]*)
(*  CellTagsIndexPosition[     83753,       2526]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["The Seesaw/Pendulum Process", "Title",
  CellMargins->{{0, -87}, {Inherited, Inherited}},
  Evaluatable->False,
  CellLabelMargins->{{0, Inherited}, {Inherited, Inherited}},
  TextAlignment->Left,
  AspectRatioFixed->True],

Cell[TextData[StyleBox["Modeling, Controller Design, Code Generation,\nand \
Simulation",
  FontSize->25]], "Subtitle",
  TextAlignment->Left,
  TextJustification->0],

Cell["Copyright \[Copyright] MathCore AB 1999-2004", "Copyright"],

Cell[TextData[StyleBox["This notebook will not work correctly in Mathematica \
6.0 and higher.",
  FontSize->18,
  FontWeight->"Bold",
  FontColor->RGBColor[1, 0, 0]]], "Text"],

Cell[BoxData[
    \(\(\(CSP\)\(=\)\(False\)\(\ \)\( (*\ 
        Whether\ Control\ System\ Professional\ is\ installed\ *) \)\(\ \
\)\)\)], "Input"],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell["\<\
This notebook illustrates how to build a dynamic model of a \
physical system and then generate efficient code for simulation.\
\>", "Text"],

Cell["\<\
The physical system is a seesaw where two carts C1 and C2 can be \
moved along two tracks by controlling the forces acting on them. Cart C1 \
carries a weight and cart C2 has an inverted pendulum attached by a friction \
free joint.\
\>", "Text"],

Cell[TextData[{
  "The model of the open loop system has the forces ",
  Cell[BoxData[
      \(TraditionalForm\`F\_1\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`F\_2\)]],
  " acting on each cart as inputs. We first model the system by Lagrangian \
methodology and then linearize the obtained nonlinear model. The linearized \
model is used for control system design where the Control System Professional \
(CSP) application package is used. The closed loop system are then simulated \
both within ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " and by external code generated using ",
  StyleBox["MathCode C++",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell[GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40009/00018`P10`400001
00000800002008200000P800P020P<30`<3L`:K:l000DP00[@1B001BDP1B
[@1Bo`2]002]DP2][@2]o`3oDP3o[E800580DU80[E80oe9B059BDU9B[E9B
oe:]05:]DU:][E:]oe;o05;oDU;o[E;oojd00:d0DZd0[Jd0ojeB0:eBDZeB
[JeBojf]0:f]DZf][Jf]ojgo0:goDZgo[Jgoool0D_l0[OmB0?mBD_mB[OmB
oon]0?n]D_n][On]ooooD_oo[@0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000?okl:2PY820P?l0003o0?oo0000ool0o`3o
oooooooooom^o`00ooooofko0007o`030?oo01co00<0ool08_l00`3oo`0Q
o`030?oo023o00<0ool06ol00`3oo`0Qo`030?oo01ko00<0ool07ol00`3o
o`0Po`030?oo027o00<0ool07Ol00`3oo`0So`030?oo01Wo00<0ool09?l0
0`3oo`0Mo`030?oo02;o00<0ool07?l00`3oo`0No`001ol00`3oo`0Ko`03
0?oo02?o00<0ool08?l00`3oo`0Po`030?oo01co00<0ool08Ol00`3oo`0M
o`030?oo01oo00<0ool08Ol00`3oo`0Qo`030?oo01go00<0ool08_l00`3o
o`0Io`030?oo02Co00<0ool07Ol00`3oo`0Ro`030?oo01go00<0ool07_l0
00Oo00<0ool06ol00`3oo`0Ro`030?oo027o00<0ool08?l00`3oo`0Lo`03
0?oo023o00<0ool07_l00`3oo`0Oo`030?oo027o00<0ool08?l00`3oo`0M
o`030?oo02?o00<0ool06Ol00`3oo`0To`030?oo01go00<0ool08_l00`3o
o`0Lo`030?oo01oo0006o`030?oo01co00<0ool08_l00`3oo`0Qo`030?oo
01oo00<0ool07?l00`3oo`0Qo`030?oo01go00<0ool08?l00`3oo`0Po`03
0?oo027o00<0ool07Ol00`3oo`0Ro`030?oo01[o00<0ool09?l00`3oo`0M
o`030?oo02;o00<0ool07?l00`3oo`0No`400006o`030?oo01_o00<0ool0
8_l00`3oo`0Qo`030?oo023o00<0ool07?l00`3oo`0Po`030?oo01ko00<0
ool07ol00`3oo`0Qo`030?oo027o00<0ool07Ol00`3oo`0Ro`030?oo01Wo
00<0ool09?l00`3oo`0Mo`030?oo02;o00<0ool07?l00`3oo`0Oo`400005
o`030?oo01co00<0ool08_l00`3oo`0Qo`030?oo023o00<0ool06ol00`3o
o`0Qo`030?oo01ko00<0ool07ol00`3oo`0Po`030?oo027o00<0ool07Ol0
0`3oo`0So`030?oo01Wo00<0ool09?l00`3oo`0Mo`030?oo02;o00<0ool0
7?l00`3oo`0Oo`400005o`030?oo01_o00<0ool08ol00`3oo`0Po`030?oo
023o00<0ool07?l00`3oo`0Qo`030?oo01go00<0ool07ol00`3oo`0Qo`03
0?oo027o00<0ool07Ol00`3oo`0Ro`030?oo01Wo00<0ool09?l00`3oo`0M
o`030?oo02;o00<0ool07Ol00`3oo`0No`400Ol000Go00<0ool06ol00`3o
o`0Ro`030?oo027o00<0ool08?l00`3oo`0Lo`030?oo023o00<0ool07_l0
0`3oo`0Oo`030?oo027o00<0ool08?l00`3oo`0Mo`030?oo02?o00<0ool0
6Ol00`3oo`0To`030?oo01go00<0ool08_l00`3oo`0Lo`030?oo01oo0@01
o`001?l00`3oo`0Lo`030?oo02;o00<0ool08Ol00`3oo`0Oo`030?oo01co
00<0ool08Ol00`3oo`0Mo`030?oo023o00<0ool08?l00`3oo`0Qo`030?oo
01go00<0ool08_l00`3oo`0Jo`030?oo02Co00<0ool07Ol00`3oo`0Ro`03
0?oo01co00<0ool07_l1007o0Ol000Co00<0ool06ol00`3oo`0Ro`030?oo
027o00<0ool08?l00`3oo`0Lo`030?oo023o00<0ool07_l00`3oo`0Oo`03
0?oo027o00<0ool08Ol00`3oo`0Mo`030?oo02;o00<0ool06Ol00`3oo`0T
o`030?oo01go00<0ool08_l00`3oo`0Lo`030?oo01oo0@01o`7o0003o`03
0?oo01co00<0ool08_l00`3oo`0Qo`030?oo023o00<0ool06ol00`3oo`0Q
o`030?oo01ko00<0ool07ol00`3oo`0Po`030?oo027o00<0ool07Ol00`3o
o`0So`030?oo01Wo00<0ool09?l00`3oo`0Mo`030?oo02;o00<0ool07?l0
0`3oo`0Oo`400Ol1o`000ol00`3oo`0Ko`030?oo02?o00<0ool08?l00`3o
o`0Po`030?oo01co00<0ool08Ol00`3oo`0Mo`030?oo01oo00<0ool08Ol0
0`3oo`0Qo`030?oo01go00<0ool08_l00`3oo`0Io`030?oo02Co00<0ool0
7Ol00`3oo`0Ro`030?oo01go00<0ool07_l00`3oo`01o`000ol00`3oo`0K
o`030?oo02;o00<0ool08Ol00`3oo`0Po`030?oo01co00<0ool08?l00`3o
o`0No`030?oo01oo00<0ool08Ol00`3oo`0Po`030?oo01go00<0ool08ol0
0`3oo`0Io`030?oo02Co00<0ool07Ol00`3oo`0Ro`030?oo01co00<0ool0
7ol00`3oo`01o`000_l00`3oo`0Lo`030?oo02;o00<0ool08Ol00`3oo`0O
o`030?oo01co00<0ool08Ol00`3oo`0Mo`030?oo023o00<0ool08?l00`3o
o`0Qo`030?oo01go00<0ool08_l00`3oo`0Jo`030?oo02Co00<0ool07Ol0
0`3oo`0Ro`030?oo01co00<0ool07_l00`3oo`02o`000_l00`3oo`0Ko`03
0?oo02;o00<0ool08Ol00`3oo`0Po`030?oo01co00<0ool08?l00`3oo`0N
o`030?oo01oo00<0ool08Ol00`3oo`0Qo`030?oo01go00<0ool08_l00`3o
o`0Io`030?oo02Co00<0ool07Ol00`3oo`0Ro`030?oo01co00<0ool07ol0
0`3oo`02o`00o`3o06/00ol00?oo:_l600[o1P3oobko003oobCo1P0Fo`H0
oolXo`00oolNo`H08_l60?oo8_l00?oo6?l602ko1P3ooaco003ooa;o1P0j
o`H0oolFo`00ool<o`H0A_l60?oo4?l00?oo1_l605;o1P3oo`[o003oo`H0
G_l60?oo1?l00?Wo1P1Zo`H0oOl00??o1P1fo`H0mol00>go1P22o`H0lOl0
0>Oo1P2>o`H0jol00>7o1P2Jo`H0iOl00=_o1P2Vo`H0gol00=Go1P2bo`H0
fOl00<oo1P2no`H0dol00<Wo1P3:o`H0cOl00<?o1P3Fo`H0aol00;go1P3R
o`H0`Ol00;Oo1P3^o`H0^ol00;7o1P3jo`H0]Ol00:_o1P3oo`Oo1P2_o`00
YOl60?oo4ol60:Wo002Oo`H0oolOo`H0Xol009Wo1P3oob_o1P2Mo`00Tol6
0?oo=ol609Oo002=o`H0oom3o`H0TOl008Oo1P3oodoo1P2;o`00POl60?oo
Fol608Go001ko`H0oomWo`H0Ool007Go1P3oog?o1P1io`00Kol60?ooOol6
07?o001Yo`H0oon;o`H0KOl006Ko0`3ooiOo0`1Zo`00I_l00`3oo`3ooiWo
00<0ool0J?l006Ko00<0ool0oonIo`030?oo06So001Vo`030?oo0?ooVOl0
0`3oo`1Xo`00I_l00`3oo`3ooiWo00<0ool0J?l006Ko00<0ool0a_l50<go
00<0ool0J?l006Ko00<0ool0a?l90<_o00<0ool0J?l006Ko00<0ool0a?l3
00Co0`3:o`030?oo06So001Vo`030?oo0<?o0P07o`80b_l00`3oo`1Xo`00
I_l00`3oo`33o`801ol20<[o00<0ool0J?l006Ko00<0ool0c?l20<[o00<0
ool0J?l006Ko00<0ool0bOl40<_o00<0ool0J?l006Ko00<0ool0a_l50<go
00<0ool0J?l006Ko00<0ool0aOl30=3o00<0ool0J?l006Ko00<0ool0a?l2
0=;o00<0ool0J?l006Ko00<0ool0a?l200Ko0P3:o`030?oo06So001Vo`03
0?oo0<Co0`04o`<0b_l00`3oo`1Xo`00I_l00`3oo`35o`P0bol00`3oo`1X
o`00I_l00`3oo`36o`H0c?l00`3oo`1Xo`00I_l00`3oo`3ooiWo00<0ool0
J?l006Ko00<0ool0oonIo`030?oo06So001Vo`030?oo0?ooVOl00`3oo`1X
o`00I_l00`3oo`3ooiWo00<0ool0J?l006Ko00<0ool0oonIo`030?oo06So
001Vo`030?oo0?ooVOl00`3oo`1Xo`00I_l00`3oo`3ooiWo00<0ool0J?l0
06Ko00<0ool0oonIo`030?oo06So001Vo`030?oo0?ooVOl00`3oo`1Xo`00
I_l00`3oo`3ooiWo00<0ool0J?l006Ko00<0ool0oonIo`030?oo06So001V
o`030?oo0?ooVOl00`3oo`1Xo`00I_l00`3oo`3ooiWo00<0ool0J?l006Ko
00<0ool0oonIo`030?oo06So001Vo`030?oo0?ooVOl00`3oo`1Xo`00I_oo
09d0J_l00=So00<0ool00ol201;o00<0ool00ol204go1@0Eo`@0ool8o`00
eol00`3oo`06o`030?oo00ko00<0ool01_l00`3oo`19o`030?oo00?o00<0
ool04?l200Co0P3oo`Ko003Fo`030?oo00Oo00<0ool03Ol00`3oo`07o`03
0?oo04Oo0P07o`030?oo00ko00<0ool01_l00`3oo`3oo`?o003Fo`030?oo
00So00<0ool02ol00`3oo`09o`030?oo04Ko00<0ool01ol00`3oo`0=o`03
0?oo00Oo00<0ool0ool2o`00eOl00`3oo`09o`030?oo00_o00<0ool02Ol0
0`3oo`15o`030?oo00Wo00<0ool02ol00`3oo`08o`030?oo0?oo0_l00=Go
00<0ool02Ol00`3oo`0;o`030?oo00Wo00<0ool0AOl00`3oo`09o`030?oo
00_o00<0ool02Ol00`3oo`3oo`7o003<ocP0?ol00`3oo`09o`030?oo00_o
00<0ool02Ol00`3oo`3oo`7o003<o`030?oo00Oo00<0ool02?l00`3oo`0;
o`030?oo00Wo00<0ool01Ol00`3oo`0mo`030?oo00Wo00<0ool02ol00`3o
o`09o`030?oo0?oo0Ol00<co00<0ool01ol00`3oo`07o`030?oo00go00<0
ool01ol00`3oo`06o`030?oo03Co=`3ko`00c?l00`3oo`08o`030?oo00Ko
00<0ool03_l00`3oo`06o`030?oo00Ko00<0ool0=?l00`3oo`06o`802Ol0
0`3oo`0<o`030?oo00So00<0ool01Ol00`3oo`3io`00c?l00`3oo`09o`03
0?oo00?o0P0Ao`801Ol200Wo00<0ool0=?l00`3oo`07o`030?oo00Ko00<0
ool03_l00`3oo`06o`030?oo00Ko00<0ool0nOl00<co00<0ool02_l601Co
1@0;o`030?oo03Co00<0ool02?l200Go0P0Ao`801Ol00`3oo`07o`030?oo
0?Wo003<o`030?oo03Co00<0ool0=?l00`3oo`09o`H05?l600[o00<0ool0
nOl00<co00<0ool0=?l00`3oo`0do`030?oo03?o00<0ool0nOl00<co00<0
ool0=?l00`3oo`0do`030?oo03?o00<0ool0nOl00<co00<0ool0=?l00`3o
o`0do`030?oo03?o00<0ool0nOl00<co00<0ool0=?l00`3oo`0do`030?oo
03?o00<0ool0nOl00<co00<0ool0=?l00`3oo`0do`030?oo03?o00<0ool0
nOl00<co00<0ool0=?l00`3oo`0do`030?oo03?o00<0ool0nOl00<co00<0
ool0=?l00`3oo`0do`030?oo03?o00<0ool0nOl00<co00<0ool0=?l00`3o
o`0do`030?oo03?o00<0ool0nOl00<co00<0ool0=?l00`3oo`0do`030?oo
03?o00<0ool0nOl00<co00<0ool0=?l00`3oo`0do`030?oo03?o00<0ool0
nOl00<co00<0ool0=?l00`3oo`0do`030?oo03?o00<0ool0nOl00<co00<0
ool0=?l00`3oo`0do`030?oo03?o00<0ool0nOl00<co00<0ool0=?l00`3o
o`0do`030?oo03?o00<0ool0nOl00<co00<0ool0=?l00`3oo`0do`030?oo
03?o00<0ool0nOl00<co00<0ool0=?l00`3oo`0do`030?oo03?o00<0ool0
nOl00<co>00fo`030?oo03?o00<0ool0nOl00=;o:`0mocL0nol00=;o00<0
ool09ol00`3oo`1Fo`030?oo0?oo5Ol00=;o00<0ool09ol00`3oo`1Fo`03
0?oo0?oo5Ol00=;o00<0ool09ol00`3oo`1Go`030?oo0?oo5?l00=;o00<0
ool09ol00`3oo`1Go`030?oo0?oo5?l00=;o00<0ool09ol00`3oo`1Go`03
0?oo0?oo5?l00=;o00<0ool09ol00`3oo`1Go`030?oo0?oo5?l00=;o00<0
ool09ol00`3oo`1Go`030?oo0?oo5?l00=;o00<0ool09ol00`3oo`1Ho`03
0?oo0?oo4ol00=;o00<0ool09ol00`3oo`1Ho`030?oo0?oo4ol00=;o00<0
ool09ol00`3oo`1Ho`030?oo0?oo4ol004Oo100<o`80NOl00`3oo`0Wo`03
0?oo05So00<0ool09ol400Wo2@3Eo`00AOl800[o0P1io`030?oo02Oo00<0
ool0F?l00`3oo`0Uo`P01ol90=Go0014o`<01?l300Wo0P1io`030?oo02Oo
00<0ool0FOl00`3oo`0So`<01?l300Oo0P3Ko`00A?l200Ko0P09o`80NOl[
05_o00<0ool08ol200Ko0P08o`80f_l004?o0P08o`802?l20?oo00<0ool0
8_l200So0P08o`80fOl004?o0P0Bo`80ool00`3oo`0Ro`804ol20=So0013
o`804_l20?oo00<0ool08_l201Co0P3Go`00@ol201;o0P3oo`030?oo02;o
0P0Eo`80e_l004?o0P0Bo`80ool1o`030?oo027o0P0Fo`80eOl004?o0P0?
o`8000?o0000ool1o`030?oo027o0P0Fo`80eOl004Co0P07o`801Ol50?oo
0Ol00`3oo`0Ro`801ol200Go0P05o`80eOl004Co0`05o`802?l30?oo0Ol0
0`3oo`0Ro`<01Ol200Ko0`03o`<0eOl004Go200:o`80ool1o`030?oo02?o
2008o`L0e_l004Oo100<o`80ool2o`030?oo02Co100;o`D0eol00?ooFol0
0`3oo`3ooa3o003ooe_o00<0ool0ool@o`00oomKo`030?oo0?oo4?l00?oo
Fol00`3oo`3ooa3o003ooeco00<0ool0ool?o`00oomLo`030?oo0?oo3ol0
0?ooG?l00`3oo`3oo`oo003ooeco00<0ool0ool?o`00oomLo`030?oo0?oo
3ol00?ooGOl00`3oo`3oo`ko003ooego00<0ool0ool>o`00oomMo`030?oo
0?oo3_l00?ooGOl00`3oo`3oo`ko003ooego00<0ool0ool>o`00oomMo`03
0?oo0?oo3_l00?ooG_l00`3oo`3oo`go003ooeko00<0ool0ool=o`00oomN
o`030?oo0?oo3Ol00?ooG_l00`3oo`3oo`go003ooeko00<0ool0ool=o`00
oomOo`030?oo0?oo3?l00?ooGol00`3oo`3oo`co003ooeoo00<0ool0ool<
o`00oomOo`030?oo0?oo3?l00?ooGol00`3oo`3oo`co003oof3o00<0ool0
ool;o`00oomPo`030?oo0?oo2ol00?ooH?l00`3oo`3oo`_o003oof3o00<0
ool0ool;o`00oomPo`030?oo0?oo2ol00?ooHOl00`3oo`3oo`[o003oof7o
00<0ool0ool:o`00oomQo`030?oo0?oo2_l00?ooHOl00`3oo`3oo`[o003o
of7o00<0ool0ool:o`00oomQo`030?oo0?oo2_l00?ooH_l00`3oo`3oo`Wo
003oof;o00<0ool0ool9o`00oomRo`030?oo0?oo2Ol00?ooH_l00`3oo`3o
o`Wo003oof;o00<0ool0ool9o`00oomSo`030?oo0?oo2?l00?ooHol00`3o
o`3oo`So003oof?o00<0ool0ool8o`00oomSo`030?oo0?oo2?l00?ooHol0
0`3oo`3oo`So003oofCo00<0ool0ool7o`00oomTo`030?oo0?oo1ol00?oo
I?l00`3oo`3oo`Oo003oofCo00<0ool0ool7o`00oomTo`030?oo0?oo1ol0
0?ooI?l00`3oo`3oo`Oo003oofGo00<0ool0ool6o`00oomUo`030?oo0?oo
1_l00?ooIOl00`3oo`3oo`Ko003oofGo00<0ool0ool6o`00oomUo`030?oo
0?oo1_l00?ooI_l00`3oo`3oo`Go003oofKo00<0ool0ool5o`00oomVo`03
0?oo0?oo1Ol00?ooI_l00`3oo`3oo`Go003oofKo00<0ool0ool5o`00oomW
o`030?oo0?oo1?l00?ooIol00`3oo`3oo`Co003oofOo00<0ool0ool4o`00
oomWo`030?oo0?oo1?l00?ooIol00`3oo`3oo`Co003oofSo00<0ool0ool3
o`00oomXo`030?oo0?oo0ol00?ooJ?l00`3oo`3oo`?o003oofSo00<0ool0
ool3o`00oomXo`030?oo0?oo0ol00?ooJ?l00`3oo`3oo`?o003oofWo00<0
ool0ool2o`00oomYo`030?oo0?oo0_l00?ooJOl00`3oo`3oo`;o003oofWo
00<0ool0ool2o`00oomYo`030?oo0?oo0_l00?ooJ_l00`3oo`3oo`7o003o
of[o00<0ool0ool1o`00oomZo`030?oo0?oo0Ol00?ooJ_l00`3oo`3oo`7o
003oof[o00<0ool0ool1o`00oom[o`030?oo0?oo003oof_o00<0ool0ool0
0?ooJol00`3oo`3oo`00oom[o`030?oo0?oo003oof_o00<0ool0ool00?oo
K?l00`3oo`3no`00oom/o`030?oo0?ko003oofco00<0ool0o_l00?ooK?l0
0`3oo`3no`00oom/o`030?oo0?ko003oofco00<0ool0o_l00?ooKOl00`3o
o`3mo`00oom]o`030?oo0?go003oofgo00<0ool0oOl00?ooKOl00`3oo`3m
o`00oom]o`030?oo0?go003oofko00<0ool0o?l00?ooK_l00`3oo`3lo`00
oom^o`030?oo0?co003oofko00<0ool0o?l00?ooK_l00`3oo`3lo`00oom_
o`030?oo0?_o003oofoo00<0ool0nol00?ooKol00`3oo`3ko`00oom_o`03
0?oo0?_o003oofoo00<0ool0nol00?ooKol00`3oo`3ko`00oom`o`030?oo
0?[o003oog3o00<0ool0n_l00?ooL?l00`3oo`3jo`00oom`o`030?oo0?[o
003oog3o00<0ool0n_l00?ooLOl00`3oo`3io`00oomao`030?oo0?Wo003o
og7o00<0ool0nOl00?ooLOl00`3oo`3io`00oomao`030?oo0?Wo003oog;o
00<0ool0n?l00?ooL_l00`3oo`3ho`00oombo`030?oo0?So003oog;o00<0
ool0n?l00?ooL_l00`3oo`3ho`00oomco`030?oo0?Oo003oog?o00<0ool0
mol00?ooLol00`3oo`3go`00oomco`030?oo0?Oo003oog?o00<0ool0mol0
0?ooLol00`3oo`3go`00oomdo`030?oo0?Ko003oogCo00<0ool0m_l00?oo
M?l00`3oo`3fo`00oomdo`030?oo0?Ko003oogCo00<0ool0m_l00?ooMOl0
0`3oo`3eo`00oomeo`030?oo0?Go003oogGo00<0ool0mOl00?ooMOl00`3o
o`3eo`00oomeo`030?oo0?Go003oogKo00<0ool0m?l00?ooM_l00`3oo`3d
o`00oomfo`030?oo0?Co003oogKo00<0ool0m?l00?ooM_l00`3oo`3do`00
oomgo`030?oo0??o003oogOo00<0ool0lol00?ooMol00`3oo`3co`00oomg
o`030?oo0??o003oogOo00<0ool0lol00001\
\>"], "Text",
  ImageSize->{620, 291},
  ImageMargins->{{0, 0}, {0, 1}},
  ImageRegion->{{0, 1}, {0, 1}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Variables and Constants", "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "Variables:\n\t",
  Cell[BoxData[
      \(TraditionalForm\`x\_1\)]],
  "\ttranslation of cart 1 from center of track\n\t",
  Cell[BoxData[
      \(TraditionalForm\`x\_2\)]],
  "\ttranslation of cart 2 from center of track\n\t\[Theta]\tangle of seesaw \
with vertical\n\t\[Alpha]\tangle of pendulum with normal to track\n\t",
  Cell[BoxData[
      \(TraditionalForm\`F\_1\)]],
  "\tforce applied to cart 1\n\t",
  Cell[BoxData[
      \(TraditionalForm\`F\_2\)]],
  "\tforce applied to cart 2"
}], "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "Constants:\n\tJ\tinertia of seesaw with track att height h\n\t",
  Cell[BoxData[
      \(TraditionalForm\`m\_s\)]],
  "\tmass of seesaw with track\n\tc\tcenter of gravity\n\th\theight of track \
from pivot point\n\t",
  Cell[BoxData[
      \(TraditionalForm\`m\_1\)]],
  "\tmass of cart 1 (weight cart, on the back track)\n\t",
  Cell[BoxData[
      \(TraditionalForm\`m\_2\)]],
  "\tmass of cart 2\t (pendulum cart, on the front track)\n\t",
  Cell[BoxData[
      \(TraditionalForm\`m\_p\)]],
  "\tmass of pendulum\n\t",
  Cell[BoxData[
      \(TraditionalForm\`l\_p\)]],
  "\tcenter of mass of pendulum rod (half of full length)\n\tg\tgravitational \
acceleration"
}], "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(physicalvalues := {\ \ J \[Rule] 1.6, m\_s \[Rule] 6.6, c \[Rule] 0.06, 
        h \[Rule] 0.115, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tm\_1 \[Rule] 
          0.48 + 0.38, m\_2 \[Rule] 0.48, 
        m\_p \[Rule] 
          0.2, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\ \ l\_p \[Rule] 
          0.61\/2 + 0.03, g \[Rule] 9.81}\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Mathematical Modeling", "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell["\<\
The Lagrangian methodology for deriving dynamic equations of a \
mechanical system consists of introducing so called generalized coordinates \
that describes (angular) positions and (angular) velocities of subsystems of \
the true system. Using these coordinates expressions for the kinetic and \
potential energy for each subsystem can be formed. Once this is done the \
dynamic equations describing the whole system can be derived automatically. \
\
\>", "Text"],

Cell[CellGroupData[{

Cell["Computation of the Lagrangian", "Subsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell["Coordinates of the center of mass of the seesaw", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(x\_s := c\ Sin[\[Theta][t]]\), "\n", 
    \(y\_s := c\ Cos[\[Theta][t]]\)}], "Input"],

Cell["The potential and kinetic energies of the seesaw", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(V\_s := m\_s\ g\ y\_s\), "\n", 
    \(T\_s := 
      Simplify[1\/2\ J\ \((\[PartialD]\_t \[Theta][t])\)\^2]\)}], "Input"],

Cell["Coordinates of the center of track", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(x\_c := h\ Sin[\[Theta][t]]\), "\n", 
    \(y\_c := h\ Cos[\[Theta][t]]\)}], "Input"],

Cell["Coordinates of cart 1", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(x\_m1 := x\_c + x\_1[t]\ Cos[\[Theta][t]]\), "\n", 
    \(y\_m1 := y\_c - x\_1[t]\ Sin[\[Theta][t]]\)}], "Input"],

Cell["The potential and kinetic energies of cart 1", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(V\_m1 := m\_1\ g\ y\_m1\), "\n", 
    \(T\_m1 := 
      Simplify[1\/2\ m\_1\ \((\((\[PartialD]\_t x\_m1)\)\^2 + \
\((\[PartialD]\_t y\_m1)\)\^2)\)]\)}], "Input"],

Cell["Coordinates of cart 2", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(x\_m2 := x\_c + x\_2[t]\ Cos[\[Theta][t]]\), "\n", 
    \(y\_m2 := y\_c - x\_2[t]\ Sin[\[Theta][t]]\)}], "Input"],

Cell["The potential and kinetic energies of cart 2", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(V\_m2 := m\_2\ g\ y\_m2\), "\n", 
    \(T\_m2 := 
      Simplify[1\/2\ m\_2\ \((\((\[PartialD]\_t x\_m2)\)\^2 + \
\((\[PartialD]\_t y\_m2)\)\^2)\)]\)}], "Input"],

Cell["Coordinates of the center of mass of the pendulum", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(x\_p := x\_m2 + l\_p\ Sin[\[Alpha][t] + \[Theta][t]]\), "\n", 
    \(y\_p := y\_m2 + l\_p\ Cos[\[Alpha][t] + \[Theta][t]]\)}], "Input"],

Cell["The potential and kinetic energies of the pendulum", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(V\_p := m\_p\ g\ y\_p\), "\n", 
    \(T\_p := 
      Simplify[1\/2\ m\_p\ \((\((\[PartialD]\_t x\_p)\)\^2 + \
\((\[PartialD]\_t y\_p)\)\^2)\)]\)}], "Input"],

Cell["The total potential energy", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(V\_tot := V\_s + V\_m1 + V\_m2 + V\_p\)], "Input"],

Cell["The total kinetic energy", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(T\_tot := T\_s + T\_m1 + T\_m2 + T\_p\)], "Input"],

Cell["The Lagrangian of the system ", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(L := T\_tot - V\_tot\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Computation of the Equations of Motion", "Subsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "The equations of motion are given by\n\t ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(\(d\/\(d\ t\)\) \[PartialD]L\/\[PartialD]\(q\& . \)\_i - \
\[PartialD]L\/\[PartialD]q\_i \[Equal] Q\_i\),
          FontSize->18], TraditionalForm]]],
  "\n",
  "(This computation may take a while ...)"
}], "Text"],

Cell[BoxData[{
    \(eqn\_1 = 
      Simplify[\[PartialD]\_t\((\[PartialD]\_\(\[PartialD]\_t x\_1[t]\)L)\) - \
\[PartialD]\_\(x\_1[t]\)L == F\_1]; \), "\n", 
    \(eqn\_2 = 
      Simplify[\[PartialD]\_t\((\[PartialD]\_\(\[PartialD]\_t x\_2[t]\)L)\) - \
\[PartialD]\_\(x\_2[t]\)L == F\_2]; \), "\n", 
    \(eqn\_3 = 
      Simplify[\[PartialD]\_t\((\[PartialD]\_\(\[PartialD]\_t \[Theta][t]\)L)\
\) - \[PartialD]\_\(\[Theta][t]\)L == 0]; \), "\n", 
    \(eqn\_4 = 
      Simplify[\[PartialD]\_t\((\[PartialD]\_\(\[PartialD]\_t \[Alpha][t]\)L)\
\) - \[PartialD]\_\(\[Alpha][t]\)L == 0]; \)}], "Input"],

Cell[TextData[{
  "These equations are coupled ordinary differential equations (ODEs) of \
second order in the generalized coordinates ",
  Cell[BoxData[
      \(TraditionalForm\`x\_1, \ \[Theta], \ x\_2, \ \[Alpha]\)]],
  ". To rewrite these in standard state space form (a system of first order \
ODEs) we introduce the following states:\n\t",
  Cell[BoxData[
      FormBox[
        RowBox[{"x", "=", 
          RowBox[{"(", GridBox[{
                {\(x\_1\), "\[Theta]", \(x\_2\), 
                  "\[Alpha]", \(\(x\& . \)\_1\), \(\[Theta]\& . \), \(\(x\& . \
\)\_2\), \(\[Alpha]\& . \)}
                }], ")"}]}], TraditionalForm]],
    FontSize->14],
  ".\nThe inputs to the system are\n\t",
  Cell[BoxData[
      FormBox[
        RowBox[{"u", "=", 
          RowBox[{"(", GridBox[{
                {\(F\_1\), \(F\_2\)}
                }], ")"}]}], TraditionalForm]]],
  ".\nWe observe that the second order time derivatives of the positions and \
angles appears linearly which makes it easy to solve for these in terms the \
states."
}], "Text"],

Cell["Solve for second order derivatives", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(sol = 
      Solve[\ {\ eqn\_1, eqn\_2, eqn\_3, 
          eqn\_4\ }, \n\t\t\t\t\t\t\t\t\t\t\t\t{\ \[PartialD]\_{t, 2}x\_1[
              t], \[PartialD]\_{t, 2}x\_2[
              t], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\ \[PartialD]\_{t, 2}\[Theta][
              t], \[PartialD]\_{t, 2}\[Alpha][t]\ }\ ]; \)], "Input"],

Cell["\<\
A look at the solutions (these will become the right hand sides of \
the differential equations used for simulating the system)\
\>", "Text"],

Cell[BoxData[
    \(sol\)], "Input"],

Cell["Turn off the spell checking", "Text"],

Cell[BoxData[
    \(Off[General::"\<spell1\>"]\)], "Input"],

Cell["Conversion rules to get rid of explicit time dependence", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(ssvariables := {\ \ x\_1[t] \[Rule] 
          x\_1, \[Theta][
            t] \[Rule] \[Theta], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tx\_2[
            t] \[Rule] 
          x\_2, \[Alpha][
            t] \[Rule] \[Alpha], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
\[PartialD]\_t x\_1[t] \[Rule] dx\_1, \[PartialD]\_t \[Theta][t] \[Rule] 
          d\[Theta], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\[PartialD]\_t x\_2[
              t] \[Rule] dx\_2, \[PartialD]\_t \[Alpha][t] \[Rule] 
          d\[Alpha]\ }; \)], "Input"],

Cell[TextData[{
  "We compute the right hand side of the nonlinear state space equations:\t",
  Cell[BoxData[
      FormBox[
        StyleBox[\(\(x\& . \)[t] = f\ [x[t], u[t]]\),
          FontSize->16], TraditionalForm]]]
}], "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    RowBox[{
      RowBox[{"f", "=", 
        RowBox[{
          RowBox[{
            RowBox[{
              RowBox[{"(", GridBox[{
                    {\(\[PartialD]\_t\ x\_1[t]\)},
                    {\(\[PartialD]\_t \[Theta][t]\)},
                    {\(\[PartialD]\_t x\_2[t]\)},
                    {\(\[PartialD]\_t \[Alpha][t]\)},
                    {\(\[PartialD]\_{t, 2}x\_1[t]\)},
                    {\(\[PartialD]\_{t, 2}\[Theta][t]\)},
                    {\(\[PartialD]\_{t, 2}x\_2[t]\)},
                    {\(\[PartialD]\_{t, 2}\[Alpha][t]\)}
                    }], ")"}], 
              "/.", \(sol\[LeftDoubleBracket]1\[RightDoubleBracket]\)}], "/.",
             "ssvariables"}], "//", "Flatten"}]}], ";"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The Linearized Model (ONLY if Control Systems Professional IS \
installed)\
\>", "Subsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell["\<\
Load the Control System Professional (CSP) package (if you don't \
have this package skip this subsection and evaluate the cells in \"The \
Linearized Model (Non Control Systems Professional Users)\" instead)\
\>", \
"Text"],

Cell[BoxData[
    \(If[CSP, Needs["\<ControlSystems`Master`\>"]]; \)], "Input"],

Cell[TextData[{
  "Linearize the equations around the origin\n\t",
  Cell[BoxData[
      FormBox[
        RowBox[{
          RowBox[{"x", "=", 
            RowBox[{"(", GridBox[{
                  {"0", "0", "0", "0", "0", "0", "0", "0"}
                  }], ")"}]}], ",", " ", 
          RowBox[{"u", "=", 
            RowBox[{
              RowBox[{"(", GridBox[{
                    {"0", "0"}
                    }], ")"}], " ", "\[Implies]", 
              " ", \(\(x\& . \)[t] \[Equal] A\ x[t] + \ B\ u[t]\)}]}]}], 
        TraditionalForm]],
    FontSize->14],
  StyleBox[".",
    FontSize->14],
  "\nusing the Linearize function in CSP"
}], "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(If[CSP, 
      ss = Linearize[
            f, {x\_1, \[Theta], 
              x\_2, \[Alpha]}, \n\t\t{{x\_1, 0}, {\[Theta], 0}, {x\_2, 
                0}, {\[Alpha], 0}, {dx\_1, 0}, {d\[Theta], 0}, {dx\_2, 
                0}, {d\[Alpha], 0}}, \n\t\t{{F\_1, 0}, {F\_2, 0}}] // 
          Simplify; \nMap[MatrixForm, ss]]\)], "Input"],

Cell["\<\
We extract the matrices from the linearization and substitute with \
the physical values of the parameters\
\>", "Text"],

Cell[BoxData[
    \(If[CSP, {\[ScriptCapitalA], \[ScriptCapitalB], \[ScriptCapitalC], \
\[ScriptCapitalD]} = \(ss /. StateSpace \[Rule] List\) /. 
          physicalvalues]; \)], "Input"],

Cell[BoxData[
    \(\[ScriptCapitalA] // MatrixForm\)], "Input"],

Cell[BoxData[
    \(\[ScriptCapitalB] // MatrixForm\)], "Input"],

Cell[BoxData[
    \(\[ScriptCapitalC] // MatrixForm\)], "Input"],

Cell[BoxData[
    \(\[ScriptCapitalD] // MatrixForm\)], "Input"],

Cell["Check controllability", "Text"],

Cell[BoxData[
    \(If[CSP, Controllable[ss /. physicalvalues]]\)], "Input"],

Cell["\<\
Compute a full state feedback controller using the LQG-method \
(LQRegulatorGains is a CSP command).\
\>", "Text"],

Cell[BoxData[
    \(If[CSP, \[ScriptCapitalL] = 
        LQRegulatorGains[\n\t\t\t\t\t\tss /. 
            physicalvalues, \n\t\t\t\t\t\tDiagonalMatrix[{1, 5. , 1, 5. , 0, 
              0, 0, 0}], \n\t\t\t\t\t\t0.1\ IdentityMatrix[
              2]\ ]; \n\[ScriptCapitalL] // MatrixForm]\)], "Input"],

Cell[TextData[{
  "The control law to be used is\n\t",
  Cell[BoxData[
      \(TraditionalForm\`u[t] = \(-\[ScriptCapitalL] . x[t]\)\)]],
  " \nwhere ",
  Cell[BoxData[
      \(TraditionalForm\`x[t]\)]],
  " are measurements of the states. This gives the following closed loop \
system to simulate\n\t",
  Cell[BoxData[
      \(TraditionalForm\`\(x\& . \)[t] = 
        f\ [x[t], \(-\[ScriptCapitalL] . x[t]\)]\)]],
  "."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The Linearized Model (ONLY if Control Systems Professional is  NOT \
installed)\
\>", "Subsection"],

Cell["\<\
This subsection only needs to be evaluated if the application \
package Control System Professional is not available.\
\>", "Text"],

Cell[BoxData[
    RowBox[{
      RowBox[{"\[ScriptCapitalA]", "=", 
        RowBox[{
          RowBox[{"(", GridBox[{
                {"0", "0", "0", "0", "1", "0", "0", "0"},
                {"0", "0", "0", "0", "0", "1", "0", "0"},
                {"0", "0", "0", "0", "0", "0", "1", "0"},
                {"0", "0", "0", "0", "0", "0", "0", "1"},
                {\(-\(\(g\ h\ m\_1\)\/J\)\), \(g - \(c\ g\ h\ m\_s\)\/J\), \
\(-\(\(g\ h\ \((m\_2 + m\_p)\)\)\/J\)\), "0", "0", "0", "0", "0"},
                {\(\(g\ m\_1\)\/J\), \(\(c\ g\ m\_s\)\/J\), \(\(g\ \((m\_2 + 
                            m\_p)\)\)\/J\), "0", "0", "0", "0", "0"},
                {\(-\(\(g\ h\ m\_1\)\/J\)\), \(g - \(c\ g\ h\ m\_s\)\/J\), \
\(-\(\(g\ h\ \((m\_2 + m\_p)\)\)\/J\)\), \(-\(\(g\ m\_p\)\/m\_2\)\), "0", "0",
                   "0", "0"},
                {\(-\(\(g\ m\_1\)\/J\)\), \(-\(\(c\ g\ m\_s\)\/J\)\), \
\(-\(\(g\ \((m\_2 + m\_p)\)\)\/J\)\), \(\(g\ \((m\_2 + 
                            m\_p)\)\)\/\(l\_p\ m\_2\)\), "0", "0", "0", "0"}
                }], ")"}], "/.", "physicalvalues"}]}], ";"}]], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"\[ScriptCapitalB]", "=", 
        RowBox[{
          RowBox[{"(", GridBox[{
                {"0", "0"},
                {"0", "0"},
                {"0", "0"},
                {"0", "0"},
                {\(h\^2\/J + 1\/m\_1\), \(h\^2\/J\)},
                {\(-\(h\/J\)\), \(-\(h\/J\)\)},
                {\(h\^2\/J\), \(h\^2\/J + 1\/m\_2\)},
                {\(h\/J\), \(h\/J - 1\/\(l\_p\ m\_2\)\)}
                }], ")"}], "/.", "physicalvalues"}]}], ";"}]], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"\[ScriptCapitalC]", "=", 
        TagBox[
          RowBox[{"(", GridBox[{
                {"1", "0", "0", "0", "0", "0", "0", "0"},
                {"0", "1", "0", "0", "0", "0", "0", "0"},
                {"0", "0", "1", "0", "0", "0", "0", "0"},
                {"0", "0", "0", "1", "0", "0", "0", "0"}
                }], ")"}],
          (MatrixForm[ #]&)]}], ";"}]], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"\[ScriptCapitalD]", "=", 
        TagBox[
          RowBox[{"(", GridBox[{
                {"0", "0"},
                {"0", "0"},
                {"0", "0"},
                {"0", "0"}
                }], ")"}],
          (MatrixForm[ #]&)]}], ";"}]], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"\[ScriptCapitalL]", "=", 
        RowBox[{"(", GridBox[{
              {"26.7929176340570451`", "52.940017572506921`", 
                "21.3676324363225722`", "11.8260757522378035`", 
                "7.23981708831001657`", "18.3074986680684386`", 
                "6.54027930973907345`", "2.28919536473904816`"},
              {\(-7.65182809411970588`\), \(-26.7908119385335119`\), \
\(-10.0775285345229015`\), \(-27.725924950006533`\), \
\(-1.62889424735209065`\), \(-8.77816698970899089`\), \(-5.38840857578496201`\
\), \(-4.54939001709865742`\)}
              }], ")"}]}], ";"}]], "Input"],

Cell["The following standard package is needed later on.", "Text"],

Cell[BoxData[
    \(Needs["\<LinearAlgebra`MatrixManipulation`\>"]\)], "Input"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Simulation", "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[TextData[{
  "Load ",
  StyleBox["MathCode C++",
    FontSlant->"Italic"],
  " and set Current Directory"
}], "Subsection"],

Cell[TextData[{
  "Load the ",
  StyleBox["MathCode C++",
    FontSlant->"Italic"],
  " package"
}], "Text"],

Cell["Needs[\"MathCode`\"];", "Input"],

Cell[TextData[{
  "Since a number of temporary files will be generated by ",
  StyleBox["MathCode C++",
    FontSlant->"Italic"],
  " we create a temporary directory and set this to the current working \
directory."
}], "Text"],

Cell["SetDirectory[$MCRoot<>\"/Demos/Pendulum\"];", "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Preliminaries", "Subsection"],

Cell["We store the states as a vector", "Text"],

Cell[BoxData[
    \(x\_ss := {x\_1, \[Theta], x\_2, \[Alpha], dx\_1, d\[Theta], dx\_2, 
        d\[Alpha]}\)], "Input"],

Cell["\<\
and close the feedback loop which gives the following right hand \
side of the state space equations\
\>", "Text"],

Cell[BoxData[
    \(f1 = 
      f /. Thread[{F\_1, 
              F\_2} \[Rule] \(-\[ScriptCapitalL] . x\_ss\)]; \)], "Input"],

Cell["Turn off the spell checking", "Text"],

Cell[BoxData[
    \(Off[General::"\<spell1\>"]\)], "Input"],

Cell["\<\
Change names of variables to only ASCII characters (needed for code \
generation)\
\>", "Text"],

Cell[BoxData[
    \(Cvariables\  = \n\t\t{x\_1 \[Rule] \ x1, \[Theta]\  \[Rule] \ 
          theta, \n\t\t\tx\_2\  \[Rule] \ x2, \[Alpha]\  \[Rule] \ 
          alpha, \n\t\t\tdx\_1\  \[Rule] \ dx1, 
        d\[Theta]\  \[Rule] \ dtheta, \n\t\t\tdx\_2\  \[Rule] \ dx2, \ 
        d\[Alpha] \[Rule] \ dalpha}; \)], "Input"],

Cell["\<\
The right hand side of the closed loop state space equations \
(intended for code generation)\
\>", "Text"],

Cell[BoxData[
    \(rhs = \(f1 /. physicalvalues\)\  /. \ Cvariables; \)], "Input"],

Cell[CellGroupData[{

Cell["The right hand side of the state space equations", "Subsubsection"],

Cell["\<\
We define the function ClosedLoopPendulumEquationsRHS which has the \
huge rhs expression from above as its body. Note that the expression also can \
read from the file PendulumExpr.m which is created above (in case of memory \
problems). This file makes it quicker to try out the compilation in different \
sessions without recomputing rhs each time.\
\>", "Text"],

Cell[BoxData[
    \(ClosedLoopPendulumEquationsRHS[\n\tReal\ x1_, Real\ theta_, Real\ x2_, 
          Real\ alpha_, \n\tReal\ dx1_, Real\ dtheta_, Real\ dx2_, 
          Real\ dalpha_] \[Rule] Real[8] = rhs; \)], "Input"],

Cell["\<\
Note that we have added type declarations to be able to generate \
code for this function later on.\
\>", "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
  "Simulation of the Nonlinear Model within ",
  StyleBox["Mathematica",
    FontSlant->"Italic"]
}], "Subsection"],

Cell[TextData[{
  "We define ",
  Cell[BoxData[
      \(TraditionalForm\`f\_cl[t]\)]],
  " to be a short notation for the ClosedLoopPendulumEquationsRHS"
}], "Text"],

Cell[BoxData[
    \(f\_cl[t_] := 
      ClosedLoopPendulumEquationsRHS[x1[t], theta[t], x2[t], alpha[t], 
        dx1[t], dtheta[t], dx2[t], dalpha[t]]\)], "Input"],

Cell["The state vector", "Text"],

Cell[BoxData[
    \(x[t_] := {x1[t], theta[t], x2[t], alpha[t], dx1[t], dtheta[t], dx2[t], 
        dalpha[t]}; \)], "Input"],

Cell["The differential equations for the closed loop system", "Text"],

Cell[BoxData[
    \(deq := Thread[\(x'\)[t] == f\_cl[t]]\)], "Input"],

Cell["\<\
Initial conditions for the simulation. Cart 1 is positioned at -1, \
the seesaw is horizontal, cart 2 is positioned at +1 and the pendulum has a \
small deviation of 5.7\[Degree] from the vertical axis. The system is at rest \
at these positions (all time derivatives are zero).\
\>", "Text"],

Cell[BoxData[
    \(initeq := Thread[x[0] == {\(-1\), 0, 1, 0.1, 0, 0, 0, 0}]\)], "Input"],

Cell["We simulate the system using NDSolve", "Text"],

Cell[TextData[StyleBox["Standard usage",
  FontWeight->"Bold"]], "Text"],

Cell[BoxData[{
    \(\(dsolStandard = 
        NDSolve[\ 
          deq\  \[Union] \ initeq, {x1, theta, x2, alpha, dx1, dtheta, dx2, 
            dalpha}, {t, 0, 10}\ ]; \) // Timing\), "\n", 
    \(NDSolveStandardTime = First[%]; \)}], "Input"],

Cell[TextData[StyleBox["Uncompiled equations",
  FontWeight->"Bold"]], "Text"],

Cell[BoxData[{
    \(\(dsolUncompiled = 
        NDSolve[\ 
          deq\  \[Union] \ initeq, {x1, theta, x2, alpha, dx1, dtheta, dx2, 
            dalpha}, {t, 0, 10}, Compiled \[Rule] False\ ]; \) // 
      Timing\), "\n", 
    \(NDSolveUncompiledTime = First[%]; \)}], "Input"],

Cell[TextData[{
  StyleBox["Force ",
    FontWeight->"Bold"],
  StyleBox["Mathematica",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  StyleBox[" to use 4-5 Runge-Kutta with step size 0.01",
    FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[{
    \(\(dsolRK = 
        NDSolve[\ 
          deq\  \[Union] \ initeq, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t{x1, theta, 
            x2, alpha, dx1, dtheta, dx2, dalpha}, {t, 0. , 
            10}, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tMaxStepSize \[Rule] 
            0.01, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tMethod \[Rule] 
            RungeKutta, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tMaxSteps \[Rule] 
            10000\ ]; \) // Timing\), "\n", 
    \(NDSolveRKTime = First[%]; \)}], "Input"],

Cell[TextData[{
  StyleBox["Force ",
    FontWeight->"Bold"],
  StyleBox["Mathematica",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  StyleBox[" to use 4-5 Runge-Kutta with step size 0.01 (uncompiled \
equations)",
    FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[{
    \(\(dsolRKUncompiled = 
        NDSolve[\ 
          deq\  \[Union] \ initeq, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t{x1, theta, 
            x2, alpha, dx1, dtheta, dx2, dalpha}, {t, 0. , 
            10}, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tMaxStepSize \[Rule] 
            0.01, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tMethod \[Rule] 
            RungeKutta, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tMaxSteps \[Rule] 
            10000, \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tCompiled \[Rule] 
            False\ ]; \) // Timing\), "\n", 
    \(NDSolveRKUncompiledTime = First[%]; \)}], "Input"],

Cell[TextData[StyleBox["The linear system",
  FontWeight->"Bold"]], "Text"],

Cell[BoxData[
    \(dlineq := 
      Thread[\(x'\)[
              t] == \((\[ScriptCapitalA] - \[ScriptCapitalB] . \
\[ScriptCapitalL])\) . x[t]] // Chop\)], "Input"],

Cell["We simulate the linear system using NDSolve", "Text"],

Cell[BoxData[
    \(dsollin = 
      NDSolve[\ dlineq\  \[Union] \ initeq, {x1, theta, x2, alpha, dx1, 
          dtheta, dx2, dalpha}, {t, 0, 8}\ ]; \)], "Input"],

Cell[TextData[{
  "Plot the result for the first 4 state variables (the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " standard usage of NDSolve) together with the result for the linear \
system. The other simulation methods above can be shown to give the same \
result as the standard usage of NDSolve."
}], "Text"],

Cell[BoxData[
    \(Map[\n\t
      Plot[#, {t, 0, 8}, PlotRange \[Rule] All, 
          DisplayFunction \[Rule] Identity, 
          PlotStyle \[Rule] {{RGBColor[0, 0, 0]}, {RGBColor[1, 0, 
                  0]}}] &, {Part[Flatten[x[t] /. dsolStandard], 
            Range[4]], \n\t\tPart[Flatten[x[t] /. dsollin], Range[4]]} // 
        Transpose\ ]; \)], "Input"],

Cell[BoxData[
    \(\(Show[GraphicsArray[%], 
        DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell["\<\
We notice that the linear response (red) differs significantly from \
the nonlinear one.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["External Simulation of the Nonlinear Model", "Subsection"],

Cell[CellGroupData[{

Cell["Code for External Simulation of the Nonlinear Model", "Subsubsection"],

Cell["\<\
We first make an implementation of a state equation solver using \
the Runge-Kutta method\
\>", "Text"],

Cell[BoxData[
    \(RK[Integer\ n_, Real\ h_, Real\ t0_, Real[_]\ startv_, Integer\ dimen_, 
          Integer\ dimPlusOne_] \[Rule] 
        Real[n, dimPlusOne] := \n\t\t\tModule[{\n\t\t\t\t\tReal[dimen]\ {x, 
              k1, k2, k3, 
              k4}, \n\t\t\t\t\tInteger\ i, \n\t\t\t\t\tReal\ t, \n\t\t\t\t\t\
Real[n, dimPlusOne]\ res\ \n\t\t\t\t}, \n\t\t\t\tt = t0; \n\t\t\t\tx = 
          startv; \n\t\t\t\tres\[LeftDoubleBracket]1, 
            1\[RightDoubleBracket] = 
          t; \n\t\t\t\tres\[LeftDoubleBracket]1, 
            2 | dimPlusOne\[RightDoubleBracket] = x; \n\t\t\t\tFor[i = 1, \ 
          i \[LessEqual] n - 1, \ \(i++\), \n\t\t\t\t\tk1 = 
            h\ ClosedLoopPendulumEquationsRHS[
                x\[LeftDoubleBracket]1\[RightDoubleBracket], 
                x\[LeftDoubleBracket]2\[RightDoubleBracket], 
                x\[LeftDoubleBracket]3\[RightDoubleBracket], 
                x\[LeftDoubleBracket]4\[RightDoubleBracket], 
                x\[LeftDoubleBracket]5\[RightDoubleBracket], 
                x\[LeftDoubleBracket]6\[RightDoubleBracket], 
                x\[LeftDoubleBracket]7\[RightDoubleBracket], 
                x\[LeftDoubleBracket]8\[RightDoubleBracket]]; \n\t\t\t\t\tk2 \
= h\ ClosedLoopPendulumEquationsRHS[
                x\[LeftDoubleBracket]1\[RightDoubleBracket] + 
                  k1\[LeftDoubleBracket]1\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]2\[RightDoubleBracket] + 
                  k1\[LeftDoubleBracket]2\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]3\[RightDoubleBracket] + 
                  k1\[LeftDoubleBracket]3\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]4\[RightDoubleBracket] + 
                  k1\[LeftDoubleBracket]4\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]5\[RightDoubleBracket] + 
                  k1\[LeftDoubleBracket]5\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]6\[RightDoubleBracket] + 
                  k1\[LeftDoubleBracket]6\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]7\[RightDoubleBracket] + 
                  k1\[LeftDoubleBracket]7\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]8\[RightDoubleBracket] + 
                  k1\[LeftDoubleBracket]8\[RightDoubleBracket]/
                    2]; \n\t\t\t\t\tk3 = 
            h\ ClosedLoopPendulumEquationsRHS[
                x\[LeftDoubleBracket]1\[RightDoubleBracket] + 
                  k2\[LeftDoubleBracket]1\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]2\[RightDoubleBracket] + 
                  k2\[LeftDoubleBracket]2\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]3\[RightDoubleBracket] + 
                  k2\[LeftDoubleBracket]3\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]4\[RightDoubleBracket] + 
                  k2\[LeftDoubleBracket]4\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]5\[RightDoubleBracket] + 
                  k2\[LeftDoubleBracket]5\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]6\[RightDoubleBracket] + 
                  k2\[LeftDoubleBracket]6\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]7\[RightDoubleBracket] + 
                  k2\[LeftDoubleBracket]7\[RightDoubleBracket]/2, 
                x\[LeftDoubleBracket]8\[RightDoubleBracket] + 
                  k2\[LeftDoubleBracket]8\[RightDoubleBracket]/
                    2]; \n\t\t\t\t\tk4 = 
            h\ ClosedLoopPendulumEquationsRHS[\n\t\t\t\t\t\tx\
\[LeftDoubleBracket]1\[RightDoubleBracket] + 
                  k3\[LeftDoubleBracket]1\[RightDoubleBracket], 
                x\[LeftDoubleBracket]2\[RightDoubleBracket] + 
                  k3\[LeftDoubleBracket]2\[RightDoubleBracket], 
                x\[LeftDoubleBracket]3\[RightDoubleBracket] + 
                  k3\[LeftDoubleBracket]3\[RightDoubleBracket], 
                x\[LeftDoubleBracket]4\[RightDoubleBracket] + 
                  k3\[LeftDoubleBracket]4\[RightDoubleBracket], 
                x\[LeftDoubleBracket]5\[RightDoubleBracket] + 
                  k3\[LeftDoubleBracket]5\[RightDoubleBracket], 
                x\[LeftDoubleBracket]6\[RightDoubleBracket] + 
                  k3\[LeftDoubleBracket]6\[RightDoubleBracket], 
                x\[LeftDoubleBracket]7\[RightDoubleBracket] + 
                  k3\[LeftDoubleBracket]7\[RightDoubleBracket], 
                x\[LeftDoubleBracket]8\[RightDoubleBracket] + 
                  k3\[LeftDoubleBracket]8\[RightDoubleBracket]]; \n\t\t\t\t\t\
x = x + 1/6\ \((k1 + 2\ k2 + 2  k3 + k4)\); \n\t\t\t\t\tt = 
            t + h; \n\t\t\t\t\tres\[LeftDoubleBracket]i + 1, 
              1\[RightDoubleBracket] = 
            t; \n\t\t\t\t\tres\[LeftDoubleBracket]i + 1, 
              2 | dimPlusOne\[RightDoubleBracket] = 
            x; \n\t\t\t\t]; \n\t\t\t\tres\n\t\t\t]\)], "Input"],

Cell["We use RK in the following ODE solver", "Text"],

Cell[BoxData[
    \(ndSolve[initvalue_, maxstepsize_, {x_, xmin_, xmax_}] := \n\t
      Module[{n, dimen}, \n\t\tn = \ 
          IntegerPart[\((xmax - xmin)\)/maxstepsize] + 1; \n\t\tdimen = 
          Length[initvalue]; \n\t\tres = 
          RK[n, maxstepsize, xmin, initvalue, dimen, 
            dimen + 1]; \n\t\tArray[\((Interpolation[
                AppendRows[
                  Transpose[{res\[LeftDoubleBracket]_, 
                        1\[RightDoubleBracket]}], \n\t\t\t\t\t\t\t\t\t\
Transpose[{res\[LeftDoubleBracket]_, # + 
                          1\[RightDoubleBracket]}]]])\) &, \ {8}]\n\t]\)], \
"Input"],

Cell["\<\
To get a correct estimate of the time spent by the external code \
for simulating the system we define the folowing function which repeats the \
external simulation n times.\
\>", "Text"],

Cell[BoxData[
    \(RKRepeated[Integer\ n_, Real\ h_, Real\ t0_, Real[_]\ startv_, 
          Integer\ dimen_, Integer\ dimPlusOne_, Integer\ loops_] \[Rule] 
        Real[n, dimPlusOne] := \n\t\ Module[{Real[n, 
              dimPlusOne]\ res}, \ \n\t\tDo[\t
          res\  = \ 
            RK[n, h, t0, startv, dimen, 
              dimPlusOne], {loops}]; \n\t\tres\n\t]\)], "Input"],

Cell[BoxData[
    \(ndSolveRepeated[\n\tinitvalue_, maxstepsize_, {x_, xmin_, xmax_}, 
        loops_] := \n\t
      Module[{n, dimen}, \n\t\tn = \ 
          IntegerPart[\((xmax - xmin)\)/maxstepsize] + 1; \n\t\tdimen = 
          Length[initvalue]; \n\t\tres = 
          RKRepeated[n, maxstepsize, xmin, initvalue, dimen, dimen + 1, \ 
            loops]; \n\t\tArray[\((Interpolation[
                AppendRows[
                  Transpose[{res\[LeftDoubleBracket]_, 
                        1\[RightDoubleBracket]}], \n\t\t\t\t\t\t\t\t\t\
Transpose[{res\[LeftDoubleBracket]_, # + 
                          1\[RightDoubleBracket]}]]])\) &, \ {8}]\n\t]\)], \
"Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compilation", "Subsubsection"],

Cell["\<\
We are now ready to compile the package. The rhs of the state \
equations are optimized using common subexpression elimination\
\>", "Text"],

Cell[BoxData[
    \(CompilePackage[
      EvaluateFunctions \[Rule] {ClosedLoopPendulumEquationsRHS}]\)], "Input"],

Cell["The corresponding code", "Text"],

Cell[BoxData[
    \(FilePrint["\<Global.cc\>"]\)], "Input"],

Cell[TextData[{
  "All necessary files for building binaries are now generated. The \
MakeBinary command can build either a standalone application using the code \
specified by the option GenerateMainFileAndFunction or a ",
  StyleBox["MathLink",
    FontSlant->"Italic"],
  " version which can be easily installed into ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ". We build the ",
  StyleBox["MathLink",
    FontSlant->"Italic"],
  " version"
}], "Text"],

Cell[BoxData[
    \(MakeBinary[]; \)], "Input"],

Cell["\<\
To use the generated code for the simulation we only have to \
install the code using the Install command.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["External Simulation", "Subsubsection"],

Cell["\<\
This function is a fix to be able to measure the time for external \
processes.\
\>", "Text"],

Cell[BoxData[
    \(SetAttributes[AbsTime, HoldFirst]; 
    AbsTime[x_] := 
      Module[{start, res}, \n\t\tstart = AbsoluteTime[]; 
        res = x; {\((AbsoluteTime[] - start)\)\ Second, res}]; \)], "Input"],

Cell[TextData[{
  StyleBox["Uncompiled code (computations within ",
    FontWeight->"Bold"],
  StyleBox["Mathematica",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  StyleBox[")",
    FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[{
    \(\(dsolUncompiled = 
        ndSolve[{\(-1\), 0, 1, 0.1, 0, 0, 0, 0}, 0.01, {t, 0. , 10. }]; \) // 
      Timing\), "\n", 
    \(ndSolveUncompiledTime = First[%]; \)}], "Input"],

Cell["We plot the result to verify the Runge-Kutta solver", "Text"],

Cell[BoxData[
    \(Map[
      Plot[#[t], {t, 0, 8}, PlotRange \[Rule] All, 
          DisplayFunction \[Rule] Identity] &, 
      Part[dsolUncompiled, Range[4]]]; \)], "Input"],

Cell[BoxData[
    \(Show[GraphicsArray[%], 
      DisplayFunction \[Rule] $DisplayFunction]; \)], "Input"],

Cell[TextData[StyleBox["Compiled code (external computations)",
  FontWeight->"Bold"]], "Text"],

Cell["We install the executable code", "Text"],

Cell[BoxData[
    \(InstallCode[]; \)], "Input"],

Cell["\<\
The external simulation is performed 600 times to get a accurate \
measure of its timing\
\>", "Text"],

Cell[BoxData[{
    \(\(dsolCompiled = 
        ndSolveRepeated[{\(-1\), 0, 1, 0.1, 0, 0, 0, 0}, 
          0.01, {t, 0. , 10. }, 600]; \) // AbsTime\), "\n", 
    \(ndSolveCompiledTime = First[%]/600; \)}], "Input"],

Cell["We plot the result to verify the compiled Runge-Kutta solver", "Text"],

Cell[BoxData[
    \(Map[
      Plot[#[t], {t, 0, 8}, PlotRange \[Rule] All, 
          DisplayFunction \[Rule] Identity] &, 
      Part[dsolCompiled, Range[4]]]; \)], "Input"],

Cell[BoxData[
    \(Show[GraphicsArray[%], 
      DisplayFunction \[Rule] $DisplayFunction]; \)], "Input"],

Cell["Looks good!", "Text"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Comparing Performance", "Section"],

Cell["\<\
We present the comparision of performance in the following \
table\
\>", "Text"],

Cell[BoxData[
    RowBox[{"TableForm", "[", 
      RowBox[{
        RowBox[{"(", GridBox[{
              {
                "NDSolveStandardTime", \(NDSolveStandardTime/
                  ndSolveCompiledTime\)},
              {
                "NDSolveUncompiledTime", \(NDSolveUncompiledTime/
                  ndSolveCompiledTime\)},
              {"NDSolveRKTime", \(NDSolveRKTime/ndSolveCompiledTime\)},
              {
                "NDSolveRKUncompiledTime", \(NDSolveRKUncompiledTime/
                  ndSolveCompiledTime\)},
              {
                "ndSolveUncompiledTime", \(ndSolveUncompiledTime/
                  ndSolveCompiledTime\)},
              {
                "ndSolveCompiledTime", \(ndSolveCompiledTime/
                  ndSolveCompiledTime\)}
              }], ")"}], 
        ",", \(TableHeadings \[Rule] {{"\<NDSolve, Standard Mathematica\>", "\
\<NDSolve, Uncomp eqns, Mathematica\>", "\<NDSolve, Runge-Kutta \
Mathematica\>", "\<NDSolve, Runge-Kutta, Uncomp eqns, Mathematica\>", \
"\<ndSolve, Mathematica\>", "\<ndSolve, MathCode C++\>"}, {"\<Time \
consumed\>", "\<Relative\>"}}\)}], " ", "]"}]], "Input"],

Cell[TextData[StyleBox["Measurements on 1.5 GHz computer, Windows XP,  \
Mathcode 1.2, Mathematica  4.1, VC++ 6.0",
  FontWeight->"Bold"]], "Text"],

Cell[BoxData[
    TagBox[GridBox[{
          {"\<\"\"\>", "\<\"Time consumed\"\>", "\<\"Relative\"\>"},
          {"\<\"NDSolve, Standard Mathematica\"\>", \(0.1399999999999988`\ \
Second\), "10.49999999999991`"},
          {"\<\"NDSolve, Uncomp eqns, Mathematica\"\>", \(1.0210000000000008`\
\ Second\), "76.57500000000006`"},
          {"\<\"NDSolve, Runge-Kutta Mathematica\"\>", \(1.3420000000000005`\ \
Second\), "100.65000000000003`"},
          {"\<\"NDSolve, Runge-Kutta, Uncomp eqns, Mathematica\"\>", \
\(13.429`\ Second\), "1007.1750000000001`"},
          {"\<\"ndSolve, Mathematica\"\>", \(13.889999999999997`\ Second\), 
            "1041.7499999999998`"},
          {"\<\"ndSolve, MathCode C++\"\>", \(0.0133333333333333333`6.6227\ \
Second\), "0.9999999999999999991`6.3216"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      Function[ BoxForm`e$, 
        TableForm[ 
        BoxForm`e$, TableHeadings -> {{"NDSolve, Standard Mathematica", 
          "NDSolve, Uncomp eqns, Mathematica", 
          "NDSolve, Runge-Kutta Mathematica", 
          "NDSolve, Runge-Kutta, Uncomp eqns, Mathematica", 
          "ndSolve, Mathematica", "ndSolve, MathCode C++"}, {"Time consumed", 
          "Relative"}}]]]], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Cleaning Up", "Section"],

Cell["Uninstall the code and delete the temporary files. ", "Text"],

Cell[BoxData[{
    \(UninstallCode[]; \), "\n", 
    \(CleanMathCodeFiles[Confirm \[Rule] False]; \)}], "Input"]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1280}, {0, 949}},
AutoGeneratedPackage->Automatic,
ScreenStyleEnvironment->"Working",
WindowToolbars->{},
CellGrouping->Automatic,
WindowSize->{861, 598},
WindowMargins->{{Automatic, 0}, {Automatic, 32}},
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PrintingMargins"->{{54, 54}, {72, 72}},
"PrintCellBrackets"->False,
"PrintRegistrationMarks"->False,
"PrintMultipleHorizontalPages"->False},
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, 128}},
ShowCellLabel->True,
ShowCellTags->False,
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
Magnification->1,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of \
all cells in a given style.  Make modifications to any definition using \
commands in the Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->WindowWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->8,
    FontSize->11],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the \
Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageHeaderLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->False,
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->20],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    Background->GrayLevel[0]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    ShowCellBracket->False,
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->False,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{6, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{22, Inherited}, {2, 12}},
    FontSize->10],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Open  ]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{8, 10}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    FontSize->7]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 1, 0]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{62, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{40, 10}, {4, 4}}],
  
  Cell[StyleData["Input", "Printout"],
    CellMargins->{{44, 0}, {6, 6}},
    Background->GrayLevel[0.8]]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{62, Inherited}, {12, 5}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{40, Inherited}, {4, 1}}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{44, 0}, {6, 2}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    ShowCellBracket->False,
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->16,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->20],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->12],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    ShowCellBracket->False,
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{17, Inherited}, {2, 12}},
    FontSize->10],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellDingbat->None,
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{62, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{74, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{52, 10}, {2, 2}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{54, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Itemize"],
    CellDingbat->"\[FilledSmallCircle]",
    CellMargins->{{65, Inherited}, {8, 8}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Itemize", "Presentation"],
    CellMargins->{{Inherited, 10}, {8, 12}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Itemize", "Condensed"],
    CellMargins->{{Inherited, Inherited}, {2, 12}},
    FontSize->10],
  
  Cell[StyleData["Itemize", "Printout"],
    CellMargins->{{Inherited, 0}, {4, 20}},
    FontSize->11]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayCode"],
    CellFrame->2,
    CellMargins->{{52, 10}, {8, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->True,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    FontWeight->"Bold",
    Background->RGBColor[1, 1, 0]],
  
  Cell[StyleData["DisplayCode", "Presentation"],
    CellMargins->{{62, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["DisplayCode", "Condensed"],
    CellMargins->{{40, 10}, {4, 4}}],
  
  Cell[StyleData["DisplayCode", "Printout"],
    CellMargins->{{44, 0}, {6, 6}},
    Background->GrayLevel[0.8]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->18],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, \
etc.  They automatically give the cell a FrameLabel containing a reference to \
a particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Open  ]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->6],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[1, 0.4, 0],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line \
help system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[1, 0.4, 0],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->GrayLevel[1],
    Background->RGBColor[1, 0.4, 0],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->GrayLevel[1],
    Background->RGBColor[1, 0.4, 0],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[1, 0.4, 0],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[1, 0.4, 0],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]]
  }, Closed]]
  }, Open  ]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1754, 51, 228, 5, 55, "Title",
  Evaluatable->False],
Cell[1985, 58, 166, 4, 80, "Subtitle"],
Cell[2154, 64, 65, 0, 56, "Copyright"],
Cell[2222, 66, 176, 4, 34, "Text"],
Cell[2401, 72, 148, 3, 50, "Input"],

Cell[CellGroupData[{
Cell[2574, 79, 31, 0, 56, "Section"],
Cell[2608, 81, 150, 3, 29, "Text"],
Cell[2761, 86, 256, 5, 48, "Text"],
Cell[3020, 93, 679, 18, 67, "Text"],
Cell[3702, 113, 12193, 202, 302, 12101, 199, "GraphicsData", "Bitmap", \
"Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15932, 320, 90, 2, 56, "Section",
  Evaluatable->False],
Cell[16025, 324, 569, 17, 215, "Text",
  Evaluatable->False],
Cell[16597, 343, 745, 21, 298, "Text",
  Evaluatable->False],
Cell[17345, 366, 362, 6, 70, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17744, 377, 88, 2, 55, "Section",
  Evaluatable->False],
Cell[17835, 381, 474, 8, 64, "Text"],

Cell[CellGroupData[{
Cell[18334, 393, 99, 2, 46, "Subsection",
  Evaluatable->False],
Cell[18436, 397, 111, 2, 28, "Text",
  Evaluatable->False],
Cell[18550, 401, 106, 2, 63, "Input"],
Cell[18659, 405, 112, 2, 28, "Text",
  Evaluatable->False],
Cell[18774, 409, 142, 3, 82, "Input"],
Cell[18919, 414, 98, 2, 28, "Text",
  Evaluatable->False],
Cell[19020, 418, 106, 2, 63, "Input"],
Cell[19129, 422, 85, 2, 28, "Text",
  Evaluatable->False],
Cell[19217, 426, 134, 2, 63, "Input"],
Cell[19354, 430, 108, 2, 28, "Text",
  Evaluatable->False],
Cell[19465, 434, 182, 4, 82, "Input"],
Cell[19650, 440, 85, 2, 28, "Text",
  Evaluatable->False],
Cell[19738, 444, 134, 2, 63, "Input"],
Cell[19875, 448, 108, 2, 28, "Text",
  Evaluatable->False],
Cell[19986, 452, 182, 4, 82, "Input"],
Cell[20171, 458, 113, 2, 28, "Text",
  Evaluatable->False],
Cell[20287, 462, 156, 2, 66, "Input"],
Cell[20446, 466, 114, 2, 28, "Text",
  Evaluatable->False],
Cell[20563, 470, 177, 4, 83, "Input"],
Cell[20743, 476, 90, 2, 28, "Text",
  Evaluatable->False],
Cell[20836, 480, 70, 1, 48, "Input"],
Cell[20909, 483, 88, 2, 28, "Text",
  Evaluatable->False],
Cell[21000, 487, 70, 1, 48, "Input"],
Cell[21073, 490, 93, 2, 28, "Text",
  Evaluatable->False],
Cell[21169, 494, 53, 1, 47, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21259, 500, 108, 2, 46, "Subsection",
  Evaluatable->False],
Cell[21370, 504, 327, 9, 95, "Text"],
Cell[21700, 515, 600, 12, 106, "Input"],
Cell[22303, 529, 1057, 26, 166, "Text"],
Cell[23363, 557, 98, 2, 28, "Text",
  Evaluatable->False],
Cell[23464, 561, 333, 6, 84, "Input"],
Cell[23800, 569, 151, 3, 28, "Text"],
Cell[23954, 574, 36, 1, 47, "Input"],
Cell[23993, 577, 43, 0, 28, "Text"],
Cell[24039, 579, 59, 1, 47, "Input"],
Cell[24101, 582, 119, 2, 28, "Text",
  Evaluatable->False],
Cell[24223, 586, 530, 10, 95, "Input"],
Cell[24756, 598, 282, 8, 28, "Text",
  Evaluatable->False],
Cell[25041, 608, 756, 17, 177, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[25834, 630, 152, 5, 46, "Subsection",
  Evaluatable->False],
Cell[25989, 637, 234, 5, 46, "Text"],
Cell[26226, 644, 79, 1, 47, "Input"],
Cell[26308, 647, 705, 22, 88, "Text",
  Evaluatable->False],
Cell[27016, 671, 354, 7, 95, "Input"],
Cell[27373, 680, 130, 3, 28, "Text"],
Cell[27506, 685, 187, 3, 47, "Input"],
Cell[27696, 690, 64, 1, 47, "Input"],
Cell[27763, 693, 64, 1, 47, "Input"],
Cell[27830, 696, 64, 1, 47, "Input"],
Cell[27897, 699, 64, 1, 47, "Input"],
Cell[27964, 702, 37, 0, 28, "Text"],
Cell[28004, 704, 76, 1, 47, "Input"],
Cell[28083, 707, 124, 3, 28, "Text"],
Cell[28210, 712, 301, 5, 111, "Input"],
Cell[28514, 719, 433, 13, 118, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[28984, 737, 109, 3, 46, "Subsection"],
Cell[29096, 742, 141, 3, 28, "Text"],
Cell[29240, 747, 1104, 19, 219, "Input"],
Cell[30347, 768, 514, 13, 217, "Input"],
Cell[30864, 783, 423, 10, 101, "Input"],
Cell[31290, 795, 303, 10, 101, "Input"],
Cell[31596, 807, 636, 12, 117, "Input"],
Cell[32235, 821, 66, 0, 28, "Text"],
Cell[32304, 823, 79, 1, 47, "Input"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[32432, 830, 77, 2, 55, "Section",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[32534, 836, 128, 5, 46, "Subsection"],
Cell[32665, 843, 108, 5, 28, "Text"],
Cell[32776, 850, 38, 0, 47, "Input"],
Cell[32817, 852, 227, 6, 28, "Text"],
Cell[33047, 860, 60, 0, 47, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[33144, 865, 35, 0, 46, "Subsection"],
Cell[33182, 867, 47, 0, 28, "Text"],
Cell[33232, 869, 119, 2, 47, "Input"],
Cell[33354, 873, 124, 3, 28, "Text"],
Cell[33481, 878, 126, 3, 47, "Input"],
Cell[33610, 883, 43, 0, 28, "Text"],
Cell[33656, 885, 59, 1, 47, "Input"],
Cell[33718, 888, 105, 3, 28, "Text"],
Cell[33826, 893, 322, 5, 111, "Input"],
Cell[34151, 900, 117, 3, 28, "Text"],
Cell[34271, 905, 83, 1, 47, "Input"],

Cell[CellGroupData[{
Cell[34379, 910, 73, 0, 37, "Subsubsection"],
Cell[34455, 912, 375, 6, 64, "Text"],
Cell[34833, 920, 221, 3, 79, "Input"],
Cell[35057, 925, 123, 3, 28, "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[35229, 934, 131, 4, 46, "Subsection"],
Cell[35363, 940, 165, 5, 28, "Text"],
Cell[35531, 947, 164, 3, 63, "Input"],
Cell[35698, 952, 32, 0, 28, "Text"],
Cell[35733, 954, 125, 2, 47, "Input"],
Cell[35861, 958, 69, 0, 28, "Text"],
Cell[35933, 960, 69, 1, 47, "Input"],
Cell[36005, 963, 302, 5, 46, "Text"],
Cell[36310, 970, 90, 1, 47, "Input"],
Cell[36403, 973, 52, 0, 28, "Text"],
Cell[36458, 975, 72, 1, 28, "Text"],
Cell[36533, 978, 246, 5, 79, "Input"],
Cell[36782, 985, 78, 1, 28, "Text"],
Cell[36863, 988, 281, 6, 79, "Input"],
Cell[37147, 996, 231, 8, 28, "Text"],
Cell[37381, 1006, 491, 9, 143, "Input"],
Cell[37875, 1017, 256, 9, 28, "Text"],
Cell[38134, 1028, 580, 10, 159, "Input"],
Cell[38717, 1040, 75, 1, 28, "Text"],
Cell[38795, 1043, 166, 4, 47, "Input"],
Cell[38964, 1049, 59, 0, 28, "Text"],
Cell[39026, 1051, 163, 3, 47, "Input"],
Cell[39192, 1056, 328, 7, 46, "Text"],
Cell[39523, 1065, 365, 7, 111, "Input"],
Cell[39891, 1074, 111, 2, 47, "Input"],
Cell[40005, 1078, 112, 3, 28, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[40154, 1086, 64, 0, 46, "Subsection"],

Cell[CellGroupData[{
Cell[40243, 1090, 76, 0, 37, "Subsubsection"],
Cell[40322, 1092, 113, 3, 28, "Text"],
Cell[40438, 1097, 4892, 79, 610, "Input"],
Cell[45333, 1178, 53, 0, 28, "Text"],
Cell[45389, 1180, 626, 12, 190, "Input"],
Cell[46018, 1194, 197, 4, 48, "Text"],
Cell[46218, 1200, 386, 7, 150, "Input"],
Cell[46607, 1209, 672, 13, 210, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[47316, 1227, 36, 0, 37, "Subsubsection"],
Cell[47355, 1229, 150, 3, 29, "Text"],
Cell[47508, 1234, 114, 2, 50, "Input"],
Cell[47625, 1238, 38, 0, 29, "Text"],
Cell[47666, 1240, 59, 1, 50, "Input"],
Cell[47728, 1243, 469, 13, 48, "Text"],
Cell[48200, 1258, 47, 1, 50, "Input"],
Cell[48250, 1261, 130, 3, 29, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[48417, 1269, 44, 0, 37, "Subsubsection"],
Cell[48464, 1271, 103, 3, 29, "Text"],
Cell[48570, 1276, 210, 4, 70, "Input"],
Cell[48783, 1282, 220, 8, 29, "Text"],
Cell[49006, 1292, 197, 4, 70, "Input"],
Cell[49206, 1298, 67, 0, 29, "Text"],
Cell[49276, 1300, 177, 4, 50, "Input"],
Cell[49456, 1306, 106, 2, 50, "Input"],
Cell[49565, 1310, 95, 1, 29, "Text"],
Cell[49663, 1313, 46, 0, 29, "Text"],
Cell[49712, 1315, 48, 1, 50, "Input"],
Cell[49763, 1318, 112, 3, 29, "Text"],
Cell[49878, 1323, 215, 4, 70, "Input"],
Cell[50096, 1329, 76, 0, 70, "Text"],
Cell[50175, 1331, 175, 4, 70, "Input"],
Cell[50353, 1337, 106, 2, 70, "Input"],
Cell[50462, 1341, 27, 0, 70, "Text"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[50550, 1348, 40, 0, 70, "Section"],
Cell[50593, 1350, 90, 3, 70, "Text"],
Cell[50686, 1355, 1146, 25, 70, "Input"],
Cell[51835, 1382, 147, 2, 70, "Text"],
Cell[51985, 1386, 1315, 27, 70, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[53337, 1418, 30, 0, 70, "Section"],
Cell[53370, 1420, 67, 0, 70, "Text"],
Cell[53440, 1422, 112, 2, 70, "Input"]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

